@@ -973,71 +973,11 @@ gskill_write_button_macro(struct ratbag_device *device,
 	return 0;
 }
 
-static int
-gskill_probe(struct ratbag_device *device)
-{
-	struct gskill_data *drv_data = NULL;
-	struct ratbag_profile *profile;
-	unsigned int active_idx;
-	int ret;
-
-	ret = ratbag_open_hidraw(device);
-	if (ret)
-		return ret;
-
-	drv_data = zalloc(sizeof(*drv_data));
-	ratbag_set_drv_data(device, drv_data);
-
-	ret = gskill_get_firmware_version(device);
-	if (ret < 0)
-		goto err;
-
-	log_debug(device->ratbag,
-		 "Firmware version: %d\n", ret);
-
-	ret = gskill_get_profile_count(device);
-	if (ret < 0)
-		goto err;
-	drv_data->profile_count = ret;
-
-	ratbag_device_init_profiles(device, GSKILL_PROFILE_MAX, GSKILL_NUM_DPI,
-				    GSKILL_BUTTON_MAX, GSKILL_NUM_LED);
-
-	ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_QUERY_CONFIGURATION);
-	ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_BUTTON_KEY);
-	ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_BUTTON_MACROS);
-	ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_DISABLE_PROFILE);
-
-	ret = gskill_get_active_profile_idx(device);
-	if (ret < 0)
-		goto err;
-
-	active_idx = ret;
-	list_for_each(profile, &device->profiles, link) {
-		if (profile->index == active_idx) {
-			profile->is_active = true;
-			break;
-		}
-	}
-
-	return 0;
-
-err:
-	if (drv_data) {
-		ratbag_set_drv_data(device, NULL);
-		free(drv_data);
-	}
-
-	ratbag_close_hidraw(device);
-	return ret;
-}
-
 static void
 gskill_read_resolutions(struct ratbag_profile *profile,
 			struct gskill_profile_report *report)
 {
 	struct gskill_profile_data *pdata = profile_to_pdata(profile);
-	struct ratbag_resolution *resolution;
 	int dpi_x, dpi_y, hz, i;
 
 	log_debug(profile->device->ratbag,
@@ -1047,16 +987,24 @@ gskill_read_resolutions(struct ratbag_profile *profile,
 	hz = GSKILL_MAX_POLLING_RATE / (report->polling_rate + 1);
 
 	for (i = 0; i < report->dpi_num; i++) {
+		_cleanup_resolution_ struct ratbag_resolution *resolution = NULL;
+		unsigned int rates[] = { 500, 1000 }; /* let's assume that is true */
+
 		dpi_x = report->dpi_levels[i].x * GSKILL_DPI_UNIT;
 		dpi_y = report->dpi_levels[i].y * GSKILL_DPI_UNIT;
 
-		resolution = ratbag_resolution_init(profile, i, dpi_x, dpi_y,
-						    hz);
+		resolution = ratbag_profile_get_resolution(profile, i);
+		ratbag_resolution_set_resolution(resolution, dpi_x, dpi_y, hz);
 		resolution->is_active = (i == report->current_dpi_level);
 		pdata->res_idx_to_dev_idx[i] = i;
 
 		ratbag_resolution_set_cap(resolution,
 					  RATBAG_RESOLUTION_CAP_SEPARATE_XY_RESOLUTION);
+
+		ratbag_resolution_set_dpi_list_from_range(resolution,
+							  GSKILL_MIN_DPI, GSKILL_MAX_DPI);
+		ratbag_resolution_set_report_rate_list(resolution, rates,
+						       ARRAY_LENGTH(rates));
 	}
 }
 
@@ -1082,7 +1030,7 @@ gskill_read_profile_name(struct ratbag_device *device,
 }
 
 static void
-gskill_read_profile(struct ratbag_profile *profile, unsigned int index)
+gskill_read_profile(struct ratbag_profile *profile)
 {
 	struct ratbag_device *device = profile->device;
 	struct gskill_data *drv_data = ratbag_get_drv_data(device);
@@ -1091,7 +1039,7 @@ gskill_read_profile(struct ratbag_profile *profile, unsigned int index)
 	uint8_t checksum;
 	int rc, retries;
 
-	if (index >= drv_data->profile_count) {
+	if (profile->index >= drv_data->profile_count) {
 		profile->is_enabled = false;
 		return;
 	}
@@ -1105,7 +1053,7 @@ gskill_read_profile(struct ratbag_profile *profile, unsigned int index)
 	 * mouse we're doing something wrong.
 	 */
 	for (retries = 0; retries < 3; retries++) {
-		rc = gskill_select_profile(device, index, false);
+		rc = gskill_select_profile(device, profile->index, false);
 		if (rc < 0)
 			return;
 
@@ -1123,12 +1071,11 @@ gskill_read_profile(struct ratbag_profile *profile, unsigned int index)
 			return;
 		}
 
-		if (report->profile_num == index)
+		if (report->profile_num == profile->index)
 			break;
 
 		log_debug(device->ratbag,
-			  "Mouse send wrong profile %d instead of %d, retrying...\n",
-			  profile->index, index);
+			  "Mouse send wrong profile, retrying...\n");
 	}
 
 	checksum = gskill_calculate_checksum((uint8_t*)report, sizeof(*report));
@@ -1161,10 +1108,11 @@ gskill_update_resolutions(struct ratbag_profile *profile)
 	 * lost on exit
 	 */
 	for (i = 0; i < GSKILL_NUM_DPI; i++) {
-		struct ratbag_resolution *res = &profile->resolution.modes[i];
+		_cleanup_resolution_ struct ratbag_resolution *res = NULL;
 		struct gskill_raw_dpi_level *level =
 			&report->dpi_levels[report->dpi_num];
 
+		res = ratbag_profile_get_resolution(profile, i);
 		if (!res->dpi_x || !res->dpi_y)
 			continue;
 
@@ -1172,8 +1120,8 @@ gskill_update_resolutions(struct ratbag_profile *profile)
 		level->y = res->dpi_y / GSKILL_DPI_UNIT;
 		pdata->res_idx_to_dev_idx[i] = report->dpi_num;
 
-		log_debug(device->ratbag, "Profile %d res %ld mapped to %d\n",
-			  profile->index, res - profile->resolution.modes,
+		log_debug(device->ratbag, "Profile %d res %d mapped to %d\n",
+			  profile->index, res->index,
 			  report->dpi_num);
 
 		report->dpi_num++;
@@ -1428,6 +1376,73 @@ gskill_update_profile(struct ratbag_profile *profile)
 	return 0;
 }
 
+static int
+gskill_probe(struct ratbag_device *device)
+{
+	struct gskill_data *drv_data = NULL;
+	struct ratbag_profile *profile;
+	unsigned int active_idx;
+	int ret;
+
+	ret = ratbag_open_hidraw(device);
+	if (ret)
+		return ret;
+
+	drv_data = zalloc(sizeof(*drv_data));
+	ratbag_set_drv_data(device, drv_data);
+
+	ret = gskill_get_firmware_version(device);
+	if (ret < 0)
+		goto err;
+
+	log_debug(device->ratbag,
+		 "Firmware version: %d\n", ret);
+
+	ret = gskill_get_profile_count(device);
+	if (ret < 0)
+		goto err;
+	drv_data->profile_count = ret;
+
+	ratbag_device_init_profiles(device, GSKILL_PROFILE_MAX, GSKILL_NUM_DPI,
+				    GSKILL_BUTTON_MAX, GSKILL_NUM_LED);
+
+	ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_QUERY_CONFIGURATION);
+	ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_BUTTON);
+	ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_BUTTON_KEY);
+	ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_BUTTON_MACROS);
+	ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_DISABLE_PROFILE);
+
+	ret = gskill_get_active_profile_idx(device);
+	if (ret < 0)
+		goto err;
+
+	active_idx = ret;
+	ratbag_device_for_each_profile(device, profile) {
+		struct ratbag_button *button;
+
+		gskill_read_profile(profile);
+
+		ratbag_profile_for_each_button(profile, button)
+			gskill_read_button(button);
+
+		if (profile->index == active_idx) {
+			profile->is_active = true;
+			break;
+		}
+	}
+
+	return 0;
+
+err:
+	if (drv_data) {
+		ratbag_set_drv_data(device, NULL);
+		free(drv_data);
+	}
+
+	ratbag_close_hidraw(device);
+	return ret;
+}
+
 static int
 gskill_commit(struct ratbag_device *device)
 {
@@ -1505,8 +1520,6 @@ struct ratbag_driver gskill_driver = {
 	.id = "gskill",
 	.probe = gskill_probe,
 	.remove = gskill_remove,
-	.read_profile = gskill_read_profile,
 	.commit = gskill_commit,
 	.set_active_profile = gskill_set_active_profile,
-	.read_button = gskill_read_button,
 };
