@@ -47,6 +47,7 @@
 
 #include "libratbag-private.h"
 #include "libratbag-hidraw.h"
+#include "libratbag-data.h"
 
 struct hidpp10drv_data {
 	struct hidpp10_device *dev;
@@ -290,27 +291,22 @@ hidpp10drv_read_button(struct ratbag_button *button)
 }
 
 static int
-hidpp10drv_write_button(struct ratbag_button *button,
+hidpp10drv_write_button(struct hidpp10_profile *profile,
+			struct ratbag_button *button,
 			const struct ratbag_button_action *action)
 {
 	struct ratbag_device *device = button->profile->device;
 	struct hidpp10drv_data *drv_data = ratbag_get_drv_data(device);
 	struct hidpp10_device *hidpp10 = drv_data->dev;
-	struct hidpp10_profile profile;
 	uint8_t code;
-	int ret;
 
 	if (hidpp10->profile_type == HIDPP10_PROFILE_UNKNOWN)
 		return -ENOTSUP;
 
-	ret = hidpp10_get_profile(hidpp10, button->profile->index, &profile);
-	if (ret)
-		return ret;
-
 	switch (action->type) {
 	case RATBAG_BUTTON_ACTION_TYPE_BUTTON:
-		profile.buttons[button->index].button.type = PROFILE_BUTTON_TYPE_BUTTON;
-		profile.buttons[button->index].button.button = action->action.button;
+		profile->buttons[button->index].button.type = PROFILE_BUTTON_TYPE_BUTTON;
+		profile->buttons[button->index].button.button = action->action.button;
 		break;
 	case RATBAG_BUTTON_ACTION_TYPE_KEY:
 		code = ratbag_hidraw_get_keyboard_usage_from_keycode(device, action->action.key.key);
@@ -319,26 +315,63 @@ hidpp10drv_write_button(struct ratbag_button *button,
 			if (code == 0)
 				return -EINVAL;
 
-			profile.buttons[button->index].consumer_control.type = PROFILE_BUTTON_TYPE_CONSUMER_CONTROL;
-			profile.buttons[button->index].consumer_control.consumer_control = code;
+			profile->buttons[button->index].consumer_control.type = PROFILE_BUTTON_TYPE_CONSUMER_CONTROL;
+			profile->buttons[button->index].consumer_control.consumer_control = code;
 		} else {
-			profile.buttons[button->index].keys.type = PROFILE_BUTTON_TYPE_KEYS;
-			profile.buttons[button->index].keys.key = code;
+			profile->buttons[button->index].keys.type = PROFILE_BUTTON_TYPE_KEYS;
+			profile->buttons[button->index].keys.key = code;
 		}
 		break;
 	case RATBAG_BUTTON_ACTION_TYPE_SPECIAL:
 		code = hidpp10_onboard_profiles_get_code_from_special(action->action.special);
 		if (code == 0)
 			return -EINVAL;
-		profile.buttons[button->index].special.type = PROFILE_BUTTON_TYPE_SPECIAL;
-		profile.buttons[button->index].special.special = code;
+		profile->buttons[button->index].special.type = PROFILE_BUTTON_TYPE_SPECIAL;
+		profile->buttons[button->index].special.special = code;
 		break;
 	case RATBAG_BUTTON_ACTION_TYPE_MACRO:
 	default:
 		return -ENOTSUP;
 	}
 
-	return hidpp10_set_profile(drv_data->dev, button->profile->index, &profile);
+	return 0;
+}
+
+static void
+hidpp10drv_read_led(struct ratbag_led *led)
+{
+	struct ratbag_device *device = led->profile->device;
+	struct hidpp10drv_data *drv_data = ratbag_get_drv_data(device);
+	struct hidpp10_device *hidpp10 = drv_data->dev;
+	struct hidpp10_profile profile;
+	int ret;
+
+	ret = hidpp10_get_profile(hidpp10, led->profile->index, &profile);
+	if (ret)
+		return;
+
+	switch (hidpp10->profile_type) {
+	case HIDPP10_PROFILE_G500:
+		led->colordepth = RATBAG_LED_COLORDEPTH_RGB_888;
+		break;
+	default:
+		led->colordepth = RATBAG_LED_COLORDEPTH_MONOCHROME;
+		break;
+	}
+
+	led->mode = RATBAG_LED_ON;
+	led->color.red = profile.red;
+	led->color.green = profile.green;
+	led->color.blue = profile.blue;
+}
+
+static void
+hidpp10drv_write_led(struct hidpp10_profile *profile,
+		     struct ratbag_led *led)
+{
+	profile->red = led->color.red;
+	profile->green = led->color.green;
+	profile->blue = led->color.blue;
 }
 
 static int
@@ -351,24 +384,37 @@ hidpp10drv_set_current_profile(struct ratbag_device *device, unsigned int index)
 }
 
 static void
-hidpp10drv_read_profile(struct ratbag_profile *profile, unsigned int index)
+hidpp10drv_read_profile(struct ratbag_profile *profile)
 {
 	struct ratbag_device *device = profile->device;
+	struct ratbag_button *button;
+	struct ratbag_led *led;
 	struct hidpp10drv_data *drv_data;
 	struct hidpp10_device *hidpp10;
-	struct hidpp10_profile p;
+	struct hidpp10_profile p = {0};
 	struct ratbag_resolution *res;
 	int rc;
-	unsigned int i;
 	uint16_t xres, yres;
-	int8_t idx;
+	uint8_t idx;
+
+	/* 0x64 USB_REFRESH_RATE has time between reports in ms. so let's
+	 * assume the 1000/500/250 rates exist on the devices */
+	unsigned int rates[] = {250, 500, 1000};
 
 	drv_data = ratbag_get_drv_data(device);
 	hidpp10 = drv_data->dev;
-	rc = hidpp10_get_profile(hidpp10, index, &p);
+	rc = hidpp10_get_profile(hidpp10, profile->index, &p);
 	if (rc)
 		return;
 
+	ratbag_profile_set_cap(profile, RATBAG_PROFILE_CAP_WRITABLE_NAME);
+
+	profile->is_enabled = p.enabled;
+
+	if (profile->name)
+		free(profile->name);
+	profile->name = strdup_safe((char*)p.name);
+
 	rc = hidpp10_get_current_profile(hidpp10, &idx);
 	if (rc == 0 && (unsigned int)idx == profile->index)
 		profile->is_active = true;
@@ -377,99 +423,89 @@ hidpp10drv_read_profile(struct ratbag_profile *profile, unsigned int index)
 	if (rc)
 		xres = 0xffff;
 
-	for (i = 0; i < profile->resolution.num_modes; i++) {
-		res = ratbag_resolution_init(profile, i,
-					     p.dpi_modes[i].xres,
-					     p.dpi_modes[i].yres,
-					     p.refresh_rate);
+	ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_RESOLUTION);
+	ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_SWITCHABLE_RESOLUTION);
+
+	ratbag_profile_for_each_resolution(profile, res) {
+		unsigned int dpis[hidpp10->dpi_count];
+
+		ratbag_resolution_set_resolution(res,
+						 p.dpi_modes[res->index].xres,
+						 p.dpi_modes[res->index].yres,
+						 p.refresh_rate);
 		ratbag_resolution_set_cap(res,
 					  RATBAG_RESOLUTION_CAP_SEPARATE_XY_RESOLUTION);
 		if (profile->is_active &&
 		    res->dpi_x == xres &&
 		    res->dpi_y == yres)
 			res->is_active = true;
-		if (i == p.default_dpi_mode)
+		if (res->index == p.default_dpi_mode) {
 			res->is_default = true;
-	}
-}
+			if (!profile->is_active)
+				res->is_active = true;
+		}
 
-static int
-hidpp10drv_write_profile(struct ratbag_profile *profile)
-{
-	return 0;
-}
+		if (hidpp10->dpi_table_is_range) {
+			unsigned int min, max;
 
-static int
-hidpp10drv_write_resolution_dpi(struct ratbag_resolution *resolution,
-				int dpi_x, int dpi_y)
-{
-	struct ratbag_profile *profile = resolution->profile;
-	struct ratbag_device *device = profile->device;
-	struct hidpp10drv_data *drv_data = ratbag_get_drv_data(device);
-	struct hidpp10_device *hidpp10 = drv_data->dev;
-	struct hidpp10_profile p;
-	unsigned int index;
-	uint16_t cur_dpi_x, cur_dpi_y;
-	int rc;
+			min = hidpp10_dpi_table_get_min_dpi(hidpp10);
+			max = hidpp10_dpi_table_get_max_dpi(hidpp10);
+			/* FIXME: this relies on libratbag using the
+			 * same steps that we support */
+			ratbag_resolution_set_dpi_list_from_range(res, min, max);
+		} else {
+			for (int i = 0; i < hidpp10->dpi_count; i++)
+				dpis[i] = hidpp10->dpi_table[i].dpi;
 
-	rc = hidpp10_get_profile(hidpp10, profile->index, &p);
-	if (rc)
-		return rc;
+			ratbag_resolution_set_dpi_list(res, dpis, hidpp10->dpi_count);
+		}
 
-	/* store the current resolution */
-	rc = hidpp10_get_current_resolution(hidpp10, &cur_dpi_x, &cur_dpi_y);
-	if (rc)
-		return rc;
-
-	if (resolution->is_active) {
-		/* we need to switch to the new resolution */
-		cur_dpi_x = dpi_x;
-		cur_dpi_y = dpi_y;
+		ratbag_resolution_set_report_rate_list(res, rates,
+						       ARRAY_LENGTH(rates));
 	}
 
-	/* retrieve which resolution is asked to be changed */
-	index = resolution - profile->resolution.modes;
+	ratbag_profile_for_each_button(profile, button)
+		hidpp10drv_read_button(button);
 
-	p.dpi_modes[index].xres = dpi_x;
-	p.dpi_modes[index].yres = dpi_y;
-
-	/* this effectively switches the resolution to the default in the profile */
-	rc = hidpp10_set_profile(drv_data->dev, profile->index, &p);
-	if (rc)
-		return rc;
-
-	/* restore the current setting */
-	return hidpp10_set_current_resolution(hidpp10, cur_dpi_x, cur_dpi_y);
+	ratbag_profile_for_each_led(profile, led)
+		hidpp10drv_read_led(led);
 }
 
 static int
 hidpp10drv_fill_from_profile(struct ratbag_device *device, struct hidpp10_device *dev)
 {
-	int rc;
-	struct hidpp10_profile profile;
-	struct hidpp10_directory directory[16];
-	int count;
-
-	count = hidpp10_get_profile_directory(dev, directory, ARRAY_LENGTH(directory));
-	if (count < 0)
-		return count;
+	int rc, num_leds;
+	struct hidpp10_profile profile = {0};
+	unsigned int i;
 
 	/* We don't know the HID++1.0 requests to query for buttons, etc.
-	 * Simply get the first profile and fill the device information in
-	 * from that.
+	 * Simply get the first enabled profile and fill in the device
+	 * information from that.
 	 */
-	rc = hidpp10_get_profile(dev, 0, &profile);
-	if (rc)
-		return rc;
+	for (i = 0; i < dev->profile_count; i++) {
+		rc = hidpp10_get_profile(dev, i, &profile);
+		if (rc)
+			return rc;
+		if (profile.enabled)
+			break;
+	}
+
+	/* let the .device file override LED count from the profile */
+	num_leds = ratbag_device_data_hidpp10_get_led_count(device->data);
+	if (num_leds >= 0)
+		profile.num_leds = num_leds;
 
 	ratbag_device_init_profiles(device,
-				    count,
+				    dev->profile_count,
 				    profile.num_dpi_modes,
 				    profile.num_buttons,
 				    profile.num_leds);
 
 	if (dev->profile_type != HIDPP10_PROFILE_UNKNOWN) {
+		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_PROFILE);
 		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_SWITCHABLE_PROFILE);
+		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_DISABLE_PROFILE);
+		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_BUTTON);
 		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_BUTTON_KEY);
 		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_BUTTON_MACROS);
 	}
@@ -491,6 +527,73 @@ hidpp10_log(void *userdata, enum hidpp_log_priority priority, const char *format
 	log_msg_va(device->ratbag, priority, format, args);
 }
 
+static int
+hidpp10drv_commit(struct ratbag_device *device)
+{
+	struct hidpp10drv_data *drv_data = ratbag_get_drv_data(device);
+	struct hidpp10_device *dev = drv_data->dev;
+	struct ratbag_profile *profile;
+	struct ratbag_button *button;
+	struct ratbag_led *led;
+	struct ratbag_resolution *resolution;
+	struct ratbag_resolution *active_resolution = NULL;
+	struct hidpp10_profile p;
+	int rc;
+
+	list_for_each(profile, &device->profiles, link) {
+		if (!profile->dirty)
+			continue;
+
+		rc = hidpp10_get_profile(dev, profile->index, &p);
+		if (rc)
+			return rc;
+
+		p.enabled = profile->is_enabled;
+		strncpy_safe((char*)p.name, profile->name, 24);
+
+		ratbag_profile_for_each_resolution(profile, resolution) {
+			p.dpi_modes[resolution->index].xres = resolution->dpi_x;
+			p.dpi_modes[resolution->index].yres = resolution->dpi_y;
+
+			if (profile->is_active && resolution->is_active)
+				active_resolution = resolution;
+		}
+
+		list_for_each(button, &profile->buttons, link) {
+			struct ratbag_button_action action = button->action;
+
+			if (!button->dirty)
+				continue;
+
+			rc = hidpp10drv_write_button(&p, button, &action);
+			if (rc)
+				return RATBAG_ERROR_DEVICE;
+		}
+
+		ratbag_profile_for_each_led(profile, led)
+			hidpp10drv_write_led(&p, led);
+
+		if (dev->profile_type != HIDPP10_PROFILE_UNKNOWN) {
+			rc = hidpp10_set_profile(dev, profile->index, &p);
+			if (rc)
+				return RATBAG_ERROR_DEVICE;
+		}
+
+		/* Update the current resolution in case it changed */
+		if (active_resolution) {
+			rc = hidpp10_set_current_resolution(dev,
+						       active_resolution->dpi_x,
+						       active_resolution->dpi_y);
+			if (rc)
+				return RATBAG_ERROR_DEVICE;
+
+			active_resolution = NULL;
+		}
+	}
+
+	return RATBAG_SUCCESS;
+}
+
 static int
 hidpp10drv_probe(struct ratbag_device *device)
 {
@@ -499,48 +602,44 @@ hidpp10drv_probe(struct ratbag_device *device)
 	struct hidpp10_device *dev = NULL;
 	struct hidpp_device base;
 	enum hidpp10_profile_type type = HIDPP10_PROFILE_UNKNOWN;
-	const char *prop;
+	const char *typestr;
 	int device_idx = HIDPP_WIRED_DEVICE_IDX;
-	int nread = 0;
+	unsigned int profile_count = 1;
+	struct ratbag_profile *profile;
 
 	rc = ratbag_find_hidraw(device, hidpp10drv_test_hidraw);
 	if (rc)
 		goto err;
 
 	drv_data = zalloc(sizeof(*drv_data));
-	hidpp_device_init(&base, device->hidraw.fd);
+	hidpp_device_init(&base, device->hidraw[0].fd);
 	hidpp_device_set_log_handler(&base, hidpp10_log, HIDPP_LOG_PRIORITY_RAW, device);
 
-	prop = ratbag_device_get_udev_property(device, "RATBAG_HIDPP10_PROFILE");
-	if (prop) {
-		if (strcasecmp("G500", prop) == 0)
+	typestr = ratbag_device_data_hidpp10_get_profile_type(device->data);
+	if (typestr) {
+		if (strcasecmp("G500", typestr) == 0)
 			type = HIDPP10_PROFILE_G500;
-		else if (strcasecmp("G700", prop) == 0)
+		else if (strcasecmp("G700", typestr) == 0)
 			type = HIDPP10_PROFILE_G700;
-		else if (strcasecmp("G9", prop) == 0)
+		else if (strcasecmp("G9", typestr) == 0)
 			type = HIDPP10_PROFILE_G9;
+
+		profile_count = ratbag_device_data_hidpp10_get_profile_count(device->data);
 	}
 
-	prop = ratbag_device_get_udev_property(device, "RATBAG_HIDPP10_INDEX");
-	if (prop) {
-		sscanf(prop, "%d%n", &device_idx, &nread);
-		if (!nread || (prop[nread]) != '\0' || device_idx < 0) {
-			log_error(device->ratbag,
-				  "Error parsing RATBAG_HIDPP10_INDEX: '%s' for %s\n",
-				  prop,
-				  device->name);
-			device_idx = HIDPP_WIRED_DEVICE_IDX;
-		}
-	}
+	device_idx = ratbag_device_data_hidpp10_get_index(device->data);
+	if (device_idx == -1)
+		device_idx = HIDPP_WIRED_DEVICE_IDX;
+
 	/* In the general case, we can treat all devices as wired devices
 	 * here. If we talk to the correct hidraw device the kernel adjusts
 	 * the device index for us, so even for unifying receiver devices
 	 * we can just use 0x00 as device index.
 	 *
-	 * If there is a special need like for G700(s), we can pass a
-	 * udev prop RATBAG_HIDPP10_INDEX.
+	 * If there is a special need like for G700(s), then add a DeviceIndex
+	 * entry to the .device file.
 	 */
-	dev = hidpp10_device_new(&base, device_idx, type);
+	dev = hidpp10_device_new(&base, device_idx, type, profile_count);
 
 	if (!dev) {
 		log_error(device->ratbag,
@@ -550,45 +649,63 @@ hidpp10drv_probe(struct ratbag_device *device)
 	}
 
 	if (type != HIDPP10_PROFILE_UNKNOWN) {
-		prop = ratbag_device_get_udev_property(device, "RATBAG_HIDPP10_DPI");
-		if (prop) {
-			rc = hidpp10_build_dpi_table_from_dpi_info(dev, prop);
+		_cleanup_(dpi_list_freep) struct dpi_list *list = NULL;
+		_cleanup_(freep) struct dpi_range *range = NULL;
+
+		range = ratbag_device_data_hidpp10_get_dpi_range(device->data);
+		if (range) {
+			rc = hidpp10_build_dpi_table_from_dpi_info(dev, range);
 			if (rc)
 				log_error(device->ratbag,
-					  "Error parsing RATBAG_HIDPP10_DPI: '%s' for %s\n",
-					  prop,
+					  "Error parsing DpiRange for %s\n",
 					  device->name);
 		}
 
-		prop = ratbag_device_get_udev_property(device, "RATBAG_HIDPP10_DPI_LIST");
-		if (prop) {
-			rc = hidpp10_build_dpi_table_from_list(dev, prop);
+		list = ratbag_device_data_hidpp10_get_dpi_list(device->data);
+		if (list) {
+			rc = hidpp10_build_dpi_table_from_list(dev, list);
 			if (rc)
 				log_error(device->ratbag,
-					  "Error parsing RATBAG_HIDPP10_DPI_LIST: '%s' for %s\n",
-					  prop,
+					  "Error parsing DpiList for %s\n",
 					  device->name);
 		}
 
 		if (!dev->dpi_count)
 			log_info(device->ratbag,
 				  "Device %s might have wrong dpi settings. "
-				  "Please add RATBAG_HIDPP10_DPI or RATBAG_HIDPP10_DPI_LIST "
-				  "to the udev properties.\n",
+				  "Please adjust the .device file.\n",
 				  device->name);
 	}
 
+	rc = hidpp10_device_read_profiles(dev);
+	if (rc)
+		goto err;
+
 	drv_data->dev = dev;
 	ratbag_set_drv_data(device, drv_data);
 
 	if (hidpp10drv_fill_from_profile(device, dev)) {
 		/* Fall back to something that every mouse has */
-		struct ratbag_profile *profile;
+		_cleanup_profile_ struct ratbag_profile *profile;
 
 		ratbag_device_init_profiles(device, 1, 1, 3, 0);
 		profile = ratbag_device_get_profile(device, 0);
 		profile->is_active = true;
-		ratbag_profile_unref(profile);
+	}
+
+	ratbag_device_for_each_profile(device, profile)
+		hidpp10drv_read_profile(profile);
+
+	if (device->num_profiles == 1) {
+		_cleanup_profile_ struct ratbag_profile *profile;
+
+		profile = ratbag_device_get_profile(device, 0);
+		if (!profile->is_active) {
+			log_debug(device->ratbag,
+				  "%s: forcing profile 0 to active.\n",
+				  device->name);
+			profile->is_active = true;
+		}
 	}
 
 	return 0;
@@ -622,10 +739,6 @@ struct ratbag_driver hidpp10_driver = {
 	.id = "hidpp10",
 	.probe = hidpp10drv_probe,
 	.remove = hidpp10drv_remove,
-	.read_profile = hidpp10drv_read_profile,
-	.write_profile = hidpp10drv_write_profile,
 	.set_active_profile = hidpp10drv_set_current_profile,
-	.read_button = hidpp10drv_read_button,
-	.write_button = hidpp10drv_write_button,
-	.write_resolution_dpi = hidpp10drv_write_resolution_dpi,
+	.commit = hidpp10drv_commit,
 };
