@@ -45,14 +45,17 @@
 
 #include "libratbag-private.h"
 #include "libratbag-hidraw.h"
+#include "libratbag-data.h"
 
 #define HIDPP_CAP_RESOLUTION_2200			(1 << 0)
 #define HIDPP_CAP_SWITCHABLE_RESOLUTION_2201		(1 << 1)
 #define HIDPP_CAP_BUTTON_KEY_1b04			(1 << 2)
 #define HIDPP_CAP_BATTERY_LEVEL_1000			(1 << 3)
 #define HIDPP_CAP_KBD_REPROGRAMMABLE_KEYS_1b00		(1 << 4)
-#define HIDPP_CAP_COLOR_LED_EFFECTS_8070			(1 << 5)
+#define HIDPP_CAP_COLOR_LED_EFFECTS_8070		(1 << 5)
 #define HIDPP_CAP_ONBOARD_PROFILES_8100			(1 << 6)
+#define HIDPP_CAP_LED_SW_CONTROL_1300			(1 << 7)
+#define HIDPP_CAP_ADJUSTIBLE_REPORT_RATE_8060		(1 << 8)
 
 struct hidpp20drv_data {
 	struct hidpp20_device *dev;
@@ -62,7 +65,10 @@ struct hidpp20drv_data {
 	unsigned num_controls;
 	struct hidpp20_control_id *controls;
 	struct hidpp20_profiles *profiles;
-	struct hidpp20_color_led_zone_info *led_infos;
+	union hidpp20_generic_led_zone_info led_infos;
+
+	unsigned int report_rates[4];
+	unsigned int num_report_rates;
 
 	unsigned int num_profiles;
 	unsigned int num_resolutions;
@@ -84,7 +90,7 @@ hidpp20drv_read_button_1b04(struct ratbag_button *button)
 
 	control = &drv_data->controls[button->index];
 	mapping = control->control_id;
-	if (control->reporting.divert || control->reporting.persist)
+	if (control->reporting.remapped)
 		mapping = control->reporting.remapped;
 	log_raw(device->ratbag,
 		  " - button%d: %s (%02x) %s%s:%d\n",
@@ -96,7 +102,7 @@ hidpp20drv_read_button_1b04(struct ratbag_button *button)
 	button->type = hidpp20_1b04_get_physical_mapping(control->task_id);
 	action = hidpp20_1b04_get_logical_mapping(mapping);
 	if (action)
-		button->action = *action;
+		ratbag_button_set_action(button, action);
 
 	ratbag_button_enable_action_type(button, RATBAG_BUTTON_ACTION_TYPE_BUTTON);
 	ratbag_button_enable_action_type(button, RATBAG_BUTTON_ACTION_TYPE_KEY);
@@ -191,6 +197,7 @@ hidpp20drv_read_button_8100(struct ratbag_button *button)
 	struct ratbag_device *device = button->profile->device;
 	struct hidpp20drv_data *drv_data = ratbag_get_drv_data(device);
 	struct hidpp20_profile *profile;
+	unsigned int modifiers = 0;
 	int rc;
 
 	if (!(drv_data->capabilities & HIDPP_CAP_ONBOARD_PROFILES_8100))
@@ -209,6 +216,7 @@ hidpp20drv_read_button_8100(struct ratbag_button *button)
 			button->action.type = RATBAG_BUTTON_ACTION_TYPE_KEY;
 			button->action.action.key.key = ratbag_hidraw_get_keycode_from_keyboard_usage(device,
 								profile->buttons[button->index].keyboard_keys.key);
+			modifiers = profile->buttons[button->index].keyboard_keys.modifier_flags;
 			break;
 		case HIDPP20_BUTTON_HID_TYPE_CONSUMER_CONTROL:
 			button->action.type = RATBAG_BUTTON_ACTION_TYPE_KEY;
@@ -231,6 +239,20 @@ hidpp20drv_read_button_8100(struct ratbag_button *button)
 		break;
 	}
 
+	if (button->action.type == RATBAG_BUTTON_ACTION_TYPE_KEY) {
+		int rc;
+
+		rc = ratbag_button_macro_new_from_keycode(button,
+							  button->action.action.key.key,
+							  modifiers);
+		if (rc < 0) {
+			log_error(device->ratbag,
+				  "Error while reading button %d\n",
+				  button->index);
+			button->action.type = RATBAG_BUTTON_ACTION_TYPE_NONE;
+		}
+	}
+
 	ratbag_button_enable_action_type(button, RATBAG_BUTTON_ACTION_TYPE_BUTTON);
 	ratbag_button_enable_action_type(button, RATBAG_BUTTON_ACTION_TYPE_KEY);
 	ratbag_button_enable_action_type(button, RATBAG_BUTTON_ACTION_TYPE_SPECIAL);
@@ -245,18 +267,75 @@ hidpp20drv_read_button(struct ratbag_button *button)
 }
 
 static void
-hidpp20drv_read_led(struct ratbag_led *led)
+hidpp20drv_read_led_1300(struct ratbag_led *led, struct hidpp20drv_data* data)
 {
-	struct ratbag_device *device = led->profile->device;
-	struct hidpp20drv_data *drv_data = ratbag_get_drv_data(device);
-	struct hidpp20_profile *profile;
-	struct hidpp20_color_led_zone_info *led_info;
-	struct hidpp20_led *h_led;
+	struct hidpp20_led_sw_ctrl_led_state state;
+	struct hidpp20_led_sw_ctrl_led_info *info;
+	int rc;
 
-	if (!(drv_data->capabilities & HIDPP_CAP_COLOR_LED_EFFECTS_8070))
+	info = &data->led_infos.leds[led->index];
+
+	rc = hidpp20_led_sw_control_get_led_state(data->dev, led->index, &state);
+
+	if (rc != 0)
 		return;
 
-	led_info = &drv_data->led_infos[led->index];
+	led->colordepth = RATBAG_LED_COLORDEPTH_MONOCHROME;
+
+	switch (info->type)
+	{
+	case HIDPP20_LED_TYPE_LOGO:
+		led->type = RATBAG_LED_TYPE_LOGO;
+		break;
+	case HIDPP20_LED_TYPE_BATTERY:
+		led->type = RATBAG_LED_TYPE_BATTERY;
+		break;
+	case HIDPP20_LED_TYPE_DPI:
+		led->type = RATBAG_LED_TYPE_DPI;
+		break;
+	case HIDPP20_LED_TYPE_PROFILE:
+	case HIDPP20_LED_TYPE_COSMETIC:
+		led->type = RATBAG_LED_TYPE_SIDE;
+		break;
+	}
+
+	switch (state.mode)
+	{
+	case HIDPP20_LED_MODE_OFF:
+		led->mode = RATBAG_LED_OFF;
+		break;
+	case HIDPP20_LED_MODE_ON:
+		led->mode = RATBAG_LED_ON;
+		break;
+	case HIDPP20_LED_MODE_BREATHING:
+		led->mode = RATBAG_LED_BREATHING;
+		led->ms = state.breathing.period;
+		led->brightness = state.breathing.brightness;
+		break;
+	default:
+		led->mode = RATBAG_LED_ON;
+	}
+
+	if (info->caps & HIDPP20_LED_MODE_ON)
+		ratbag_led_set_mode_capability(led, RATBAG_LED_ON);
+	if (info->caps & HIDPP20_LED_MODE_OFF)
+		ratbag_led_set_mode_capability(led, RATBAG_LED_OFF);
+	if (info->caps & (HIDPP20_LED_MODE_BLINK | HIDPP20_LED_MODE_TRAVEL |
+			  HIDPP20_LED_MODE_RAMP_UP | HIDPP20_LED_MODE_RAMP_DOWN |
+			  HIDPP20_LED_MODE_HEARTBEAT | HIDPP20_LED_MODE_BREATHING))
+		ratbag_led_set_mode_capability(led, RATBAG_LED_BREATHING);
+}
+
+static void
+hidpp20drv_read_led_8070(struct ratbag_led *led, struct hidpp20drv_data* drv_data)
+{
+	struct hidpp20_profile *profile;
+	struct hidpp20_led *h_led;
+	struct hidpp20_color_led_zone_info* led_info;
+	struct hidpp20_color_led_info info;
+	int rc;
+
+	led_info = &drv_data->led_infos.color_leds[led->index];
 	profile = &drv_data->profiles->profiles[led->profile->index];
 	h_led = &profile->leds[led->index];
 
@@ -274,21 +353,74 @@ hidpp20drv_read_led(struct ratbag_led *led)
 		led->mode = RATBAG_LED_OFF;
 		break;
 	}
-	led->type = hidpp20_8070_get_location_mapping(led_info->location);
+
+	/* pre-filled, only override if unknown */
+	if (led->type == RATBAG_LED_TYPE_UNKNOWN)
+		led->type = hidpp20_8070_get_location_mapping(led_info->location);
 	led->color.red = h_led->color.red;
 	led->color.green = h_led->color.green;
 	led->color.blue = h_led->color.blue;
-	led->hz = h_led->period;  /* FIXME: should be ceil(1000.0 / h_led->period), but that would be very small */
+	led->ms = h_led->period;
 	led->brightness = h_led->brightness * 255 / 100;
+
+	rc = hidpp20_color_led_effects_get_info(drv_data->dev, &info);
+	if (rc == 0 &&
+	    info.ext_caps & HIDPP20_COLOR_LED_INFO_EXT_CAP_MONOCHROME_ONLY)
+		led->colordepth = RATBAG_LED_COLORDEPTH_MONOCHROME;
+	else
+		led->colordepth = RATBAG_LED_COLORDEPTH_RGB_888;
+
+	for (int i = 0; i < led_info->num_effects; i++) {
+		struct hidpp20_color_led_zone_effect_info ei;
+		rc = hidpp20_color_led_effect_get_zone_effect_info(drv_data->dev,
+								   led_info->index,
+								   i, &ei);
+		if (rc < 0)
+			break;
+
+		switch (ei.effect_id) {
+		case HIDPP20_COLOR_LED_ZONE_EFFECT_DISABLED:
+			ratbag_led_set_mode_capability(led, RATBAG_LED_OFF);
+			break;
+		case HIDPP20_COLOR_LED_ZONE_EFFECT_FIXED:
+			ratbag_led_set_mode_capability(led, RATBAG_LED_ON);
+			break;
+		case HIDPP20_COLOR_LED_ZONE_EFFECT_CYCLING:
+			ratbag_led_set_mode_capability(led, RATBAG_LED_CYCLE);
+			break;
+		case HIDPP20_COLOR_LED_ZONE_EFFECT_WAVE:
+		case HIDPP20_COLOR_LED_ZONE_EFFECT_BREATHING:
+			ratbag_led_set_mode_capability(led, RATBAG_LED_BREATHING);
+			break;
+		default:
+			log_bug_libratbag(led->profile->device->ratbag,
+					  "%s: Unknown effect id %d\n",
+					  led->profile->device->name,
+					  ei.effect_id);
+			break;
+		}
+	}
+}
+
+static void
+hidpp20drv_read_led(struct ratbag_led *led)
+{
+	struct ratbag_device *device = led->profile->device;
+	struct hidpp20drv_data *drv_data = ratbag_get_drv_data(device);
+
+	if (drv_data->capabilities & HIDPP_CAP_COLOR_LED_EFFECTS_8070)
+		hidpp20drv_read_led_8070(led, drv_data);
+	else if (drv_data->capabilities & HIDPP_CAP_LED_SW_CONTROL_1300)
+		hidpp20drv_read_led_1300(led, drv_data);
 }
 
 static int
-hidpp20drv_write_button_1b04(struct ratbag_button *button,
-			     const struct ratbag_button_action *action)
+hidpp20drv_update_button_1b04(struct ratbag_button *button)
 {
 	struct ratbag_device *device = button->profile->device;
 	struct hidpp20drv_data *drv_data = ratbag_get_drv_data(device);
 	struct hidpp20_control_id *control;
+	struct ratbag_button_action *action = &button->action;
 	uint16_t mapping;
 	int rc;
 
@@ -300,14 +432,11 @@ hidpp20drv_write_button_1b04(struct ratbag_button *button,
 	if (!mapping)
 		return -EINVAL;
 
-	control->reporting.divert = 1;
+	control->reporting.divert = 0;
 	control->reporting.remapped = mapping;
 	control->reporting.updated = 1;
 
 	rc = hidpp20_special_key_mouse_set_control(drv_data->dev, control);
-	if (rc == ERR_INVALID_ADDRESS)
-		return -EINVAL;
-
 	if (rc)
 		log_error(device->ratbag,
 			  "Error while writing profile: '%s' (%d)\n",
@@ -318,12 +447,14 @@ hidpp20drv_write_button_1b04(struct ratbag_button *button,
 }
 
 static int
-hidpp20drv_update_button_8100(struct ratbag_button *button,
-			      const struct ratbag_button_action *action)
+hidpp20drv_update_button_8100(struct ratbag_button *button)
 {
 	struct ratbag_device *device = button->profile->device;
 	struct hidpp20drv_data *drv_data = ratbag_get_drv_data(device);
 	struct hidpp20_profile *profile;
+	struct ratbag_button_action *action = &button->action;
+	unsigned int modifiers, key;
+	int rc;
 	uint8_t code, type, subtype;
 
 	if (!(drv_data->capabilities & HIDPP_CAP_ONBOARD_PROFILES_8100))
@@ -337,6 +468,34 @@ hidpp20drv_update_button_8100(struct ratbag_button *button,
 		profile->buttons[button->index].button.subtype = HIDPP20_BUTTON_HID_TYPE_MOUSE;
 		profile->buttons[button->index].button.buttons = action->action.button;
 		break;
+	case RATBAG_BUTTON_ACTION_TYPE_MACRO:
+		type = HIDPP20_BUTTON_HID_TYPE;
+		subtype = HIDPP20_BUTTON_HID_TYPE_KEYBOARD;
+		rc = ratbag_action_keycode_from_macro(action,
+						      &key,
+						      &modifiers);
+		if (rc < 0) {
+			log_error(device->ratbag,
+				  "Error while writing macro for button %d\n",
+				  button->index);
+		}
+
+		code = ratbag_hidraw_get_keyboard_usage_from_keycode(device, key);
+		if (code == 0) {
+			subtype = HIDPP20_BUTTON_HID_TYPE_CONSUMER_CONTROL;
+			code = ratbag_hidraw_get_consumer_usage_from_keycode(device, action->action.key.key);
+			if (code == 0)
+				return -EINVAL;
+		}
+		profile->buttons[button->index].subany.type = type;
+		profile->buttons[button->index].subany.subtype = subtype;
+		if (subtype == HIDPP20_BUTTON_HID_TYPE_KEYBOARD) {
+			profile->buttons[button->index].keyboard_keys.key = code;
+			profile->buttons[button->index].keyboard_keys.modifier_flags = modifiers;
+		} else {
+			profile->buttons[button->index].consumer_control.consumer_control = code;
+		}
+		break;
 	case RATBAG_BUTTON_ACTION_TYPE_KEY:
 		type = HIDPP20_BUTTON_HID_TYPE;
 		subtype = HIDPP20_BUTTON_HID_TYPE_KEYBOARD;
@@ -361,42 +520,98 @@ hidpp20drv_update_button_8100(struct ratbag_button *button,
 		profile->buttons[button->index].special.type = HIDPP20_BUTTON_SPECIAL;
 		profile->buttons[button->index].special.special = code;
 		break;
-	case RATBAG_BUTTON_ACTION_TYPE_MACRO:
 	default:
 		return -ENOTSUP;
 	}
 
-	return hidpp20_onboard_profiles_write(drv_data->dev, button->profile->index, drv_data->profiles);
+	return 0;
 }
 
 static int
-hidpp20drv_update_button(struct ratbag_button *button,
-			const struct ratbag_button_action *action)
+hidpp20drv_update_button(struct ratbag_button *button)
 {
 	struct ratbag_device *device = button->profile->device;
 	struct hidpp20drv_data *drv_data = ratbag_get_drv_data(device);
 
 	if (drv_data->capabilities & HIDPP_CAP_ONBOARD_PROFILES_8100)
-		return hidpp20drv_update_button_8100(button, action);
+		return hidpp20drv_update_button_8100(button);
 
 	if (drv_data->capabilities & HIDPP_CAP_BUTTON_KEY_1b04)
-		return hidpp20drv_write_button_1b04(button, action);
+		return hidpp20drv_update_button_1b04(button);
 
 	return -ENOTSUP;
 }
 
 static int
-hidpp20drv_update_led(struct ratbag_led *led)
+hidpp20drv_update_led_1300(struct ratbag_led *led, struct hidpp20drv_data *data)
+{
+	const uint16_t led_caps = data->led_infos.leds[led->index].caps;
+	struct hidpp20_led_sw_ctrl_led_state h_led;
+	int rc;
+
+	h_led.index = led->index;
+
+	switch(led->mode)
+	{
+	case RATBAG_LED_BREATHING:
+		h_led.mode = HIDPP20_LED_MODE_BREATHING;
+		h_led.breathing.brightness = led->brightness;
+		h_led.breathing.period = led->ms;
+		h_led.breathing.timeout = 300;
+		break;
+	case RATBAG_LED_OFF:
+		h_led.mode = HIDPP20_LED_MODE_OFF;
+		h_led.on.index = HIDPP20_LED_SW_CONTROL_LED_INDEX_ALL;
+		break;
+	case RATBAG_LED_ON:
+		h_led.mode = HIDPP20_LED_MODE_ON;
+		h_led.on.index = HIDPP20_LED_SW_CONTROL_LED_INDEX_ALL;
+		break;
+	case RATBAG_LED_CYCLE:
+		return -ENOTSUP;
+	}
+
+	if (!(h_led.mode & led_caps)) {
+		hidpp_log_error(&data->dev->base, "LED %d does not support effect %s(%04x), supports %04x\n",
+						led->index, hidpp20_sw_led_control_get_mode_string(h_led.mode),
+						h_led.mode, led_caps);
+		return -ENOTSUP;
+	}
+
+	if (!hidpp20_led_sw_control_get_sw_ctrl(data->dev)) {
+		rc = hidpp20_led_sw_control_set_sw_ctrl(data->dev, true);
+		if (rc)
+			return rc;
+	}
+
+	rc = hidpp20_led_sw_control_set_led_state(data->dev, &h_led);
+	if (rc)
+		return rc;
+
+	/*
+	 * We revert to HW control of the LEDs.
+	 * This effectively disable the capability of changing the DPI/profile
+	 * LEDs, but this allows to make them work with the internal DPI/profile
+	 * switches.
+	 *
+	 * The net result is:
+	 * - DPI/profiles can't be controlled (but who wants this?)
+	 * - The logo can properly controlled
+	 */
+	rc = hidpp20_led_sw_control_set_sw_ctrl(data->dev, false);
+	if (rc)
+		return rc;
+
+	return RATBAG_SUCCESS;
+}
+
+static int
+hidpp20drv_update_led_8070(struct ratbag_led *led, struct ratbag_profile* profile,
+						   struct hidpp20drv_data *drv_data)
 {
-	struct ratbag_profile *profile = led->profile;
-	struct ratbag_device *device = profile->device;
-	struct hidpp20drv_data *drv_data = ratbag_get_drv_data(device);
 	struct hidpp20_profile *h_profile;
 	struct hidpp20_led *h_led;
 
-	if (!(drv_data->capabilities & HIDPP_CAP_COLOR_LED_EFFECTS_8070))
-		return -ENOTSUP;
-
 	h_profile = &drv_data->profiles->profiles[profile->index];
 
 	h_led = &(h_profile->leds[led->index]);
@@ -421,12 +636,28 @@ hidpp20drv_update_led(struct ratbag_led *led)
 	h_led->color.red = led->color.red;
 	h_led->color.green = led->color.green;
 	h_led->color.blue = led->color.blue;
-	h_led->period = led->hz; /* FIXME: should be 1000 / led->hz, but that would be very small */
+	h_led->period = led->ms;
 	h_led->brightness = led->brightness * 100 / 255;
 
 	return RATBAG_SUCCESS;
 }
 
+static int
+hidpp20drv_update_led(struct ratbag_led *led)
+{
+	struct ratbag_profile *profile = led->profile;
+	struct ratbag_device *device = profile->device;
+	struct hidpp20drv_data *drv_data = ratbag_get_drv_data(device);
+
+	if (drv_data->capabilities & HIDPP_CAP_COLOR_LED_EFFECTS_8070)
+		return hidpp20drv_update_led_8070(led, profile, drv_data);
+
+	if (drv_data->capabilities & HIDPP_CAP_LED_SW_CONTROL_1300)
+		return hidpp20drv_update_led_1300(led, drv_data);
+
+	return RATBAG_ERROR_CAPABILITY;
+}
+
 static int
 hidpp20drv_current_profile(struct ratbag_device *device)
 {
@@ -458,7 +689,8 @@ hidpp20drv_set_current_profile(struct ratbag_device *device, unsigned int index)
 
 	h_profile = &drv_data->profiles->profiles[index];
 	if (!h_profile->enabled) {
-		rc = hidpp20_onboard_profiles_write(drv_data->dev, index, drv_data->profiles);
+		h_profile->enabled = 1;
+		rc = hidpp20_onboard_profiles_commit(drv_data->dev, drv_data->profiles);
 		if (rc)
 			return rc;
 	}
@@ -493,7 +725,43 @@ hidpp20drv_read_resolution_dpi_2201(struct ratbag_device *device)
 		  drv_data->sensors[0].dpi_max);
 
 	drv_data->num_sensors = rc;
-	drv_data->num_resolutions = drv_data->num_sensors;
+
+	/* if 0x8100 has already been enumerated we already have the supported
+	 * number of resolutions and shouldn't overwrite it
+	 */
+	if (!(drv_data->capabilities & HIDPP_CAP_ONBOARD_PROFILES_8100))
+		drv_data->num_resolutions = drv_data->num_sensors;
+
+	return 0;
+}
+
+static int
+hidpp20drv_read_report_rate_8060(struct ratbag_device *device)
+{
+	struct hidpp20drv_data *drv_data = ratbag_get_drv_data(device);
+	struct ratbag *ratbag = device->ratbag;
+	uint8_t bitflags_ms;
+	int nrates = 0;
+	int rc;
+
+	rc = hidpp20_adjustable_report_rate_get_report_rate_list(drv_data->dev,
+								 &bitflags_ms);
+	if (rc < 0)
+		return rc;
+
+	/* We only care about 'standard' rates */
+	if (bitflags_ms & 0x80)
+		drv_data->report_rates[nrates++] = 125;
+	if (bitflags_ms & 0x8)
+		drv_data->report_rates[nrates++] = 250;
+	if (bitflags_ms & 0x2)
+		drv_data->report_rates[nrates++] = 500;
+	if (bitflags_ms & 0x1)
+		drv_data->report_rates[nrates++] = 1000;
+
+	drv_data->num_report_rates = nrates;
+
+	log_debug(ratbag, "device has %d report rates\n", nrates);
 
 	return 0;
 }
@@ -505,8 +773,9 @@ hidpp20drv_read_resolution_dpi(struct ratbag_profile *profile)
 	struct ratbag *ratbag = device->ratbag;
 	struct hidpp20drv_data *drv_data = ratbag_get_drv_data(device);
 	struct ratbag_resolution *res;
+	unsigned int default_dpi = 1000;
+	unsigned int default_rate = 500;
 	int rc;
-	unsigned int i;
 
 	if (drv_data->capabilities & HIDPP_CAP_RESOLUTION_2200) {
 		uint16_t resolution;
@@ -527,17 +796,40 @@ hidpp20drv_read_resolution_dpi(struct ratbag_profile *profile)
 		rc = hidpp20drv_read_resolution_dpi_2201(device);
 		if (rc < 0)
 			return rc;
-		for (i = 0; i < profile->resolution.num_modes; i++) {
-			int dpi = drv_data->sensors[i].dpi;
-			/* FIXME: retrieve the refresh rate */
-			res = ratbag_resolution_init(profile, i, dpi, dpi, 0);
 
+		ratbag_profile_for_each_resolution(profile, res) {
+			struct hidpp20_sensor *sensor;
+
+			/* We only look at the first sensor. Multiple
+			 * sensors is too niche to care about right now */
+			sensor = &drv_data->sensors[0];
+
+			/* FIXME: retrieve the refresh rate */
+			ratbag_resolution_set_resolution(res, sensor->dpi, sensor->dpi, 0);
+			ratbag_resolution_set_dpi_list_from_range(res,
+								  sensor->dpi_min,
+								  sensor->dpi_max);
 			/* FIXME: we mark all resolutions as active because
 			 * they are from different sensors */
 			res->is_active = true;
 		}
+	} else {
+		ratbag_profile_for_each_resolution(profile, res)
+			ratbag_resolution_set_dpi_list(res, &default_dpi, 1);
+	}
 
-		return 0;
+	if (drv_data->capabilities & HIDPP_CAP_ADJUSTIBLE_REPORT_RATE_8060) {
+		rc = hidpp20drv_read_report_rate_8060(device);
+		if (rc < 0)
+			return rc;
+
+		ratbag_profile_for_each_resolution(profile, res)
+			ratbag_resolution_set_report_rate_list(res,
+							       drv_data->report_rates,
+							       drv_data->num_report_rates);
+	} else {
+		ratbag_profile_for_each_resolution(profile, res)
+			ratbag_resolution_set_report_rate_list(res, &default_rate, 1);
 	}
 
 	return 0;
@@ -551,14 +843,10 @@ hidpp20drv_update_resolution_dpi_8100(struct ratbag_resolution *resolution,
 	struct ratbag_device *device = profile->device;
 	struct hidpp20drv_data *drv_data = ratbag_get_drv_data(device);
 	struct hidpp20_profile *h_profile;
-	unsigned int index;
 	int dpi = dpi_x; /* dpi_x == dpi_y if we don't have the individual resolution cap */
 
-	/* retrieve which resolution is asked to be changed */
-	index = resolution - profile->resolution.modes;
-
 	h_profile = &drv_data->profiles->profiles[profile->index];
-	h_profile->dpi[index] = dpi;
+	h_profile->dpi[resolution->index] = dpi;
 
 	return RATBAG_SUCCESS;
 }
@@ -662,11 +950,11 @@ hidpp20drv_read_color_leds(struct ratbag_device *device)
 	if (!(drv_data->capabilities & HIDPP_CAP_COLOR_LED_EFFECTS_8070))
 		return 0;
 
-	free(drv_data->led_infos);
-	drv_data->led_infos = NULL;
+	free(drv_data->led_infos.color_leds);
+	drv_data->led_infos.color_leds = NULL;
 	drv_data->num_leds = 0;
 
-	rc = hidpp20_color_led_effects_get_zone_infos(drv_data->dev, &drv_data->led_infos);
+	rc = hidpp20_color_led_effects_get_zone_infos(drv_data->dev, &drv_data->led_infos.color_leds);
 
 	if (rc > 0) {
 		drv_data->num_leds = rc;
@@ -677,71 +965,101 @@ hidpp20drv_read_color_leds(struct ratbag_device *device)
 }
 
 static int
-hidpp20drv_read_onboard_profile(struct ratbag_device *device, unsigned index)
+hidpp20drv_read_leds(struct ratbag_device *device)
 {
 	struct hidpp20drv_data *drv_data = ratbag_get_drv_data(device);
 	int rc;
 
-	if (!(drv_data->capabilities & HIDPP_CAP_ONBOARD_PROFILES_8100))
+	if (!(drv_data->capabilities & HIDPP_CAP_LED_SW_CONTROL_1300))
 		return 0;
 
-	rc = hidpp20_onboard_profiles_read(drv_data->dev, index, drv_data->profiles);
-	if (rc < 0)
-		return rc;
+	free(drv_data->led_infos.leds);
+	drv_data->led_infos.leds = NULL;
 
-	return 0;
+	rc = hidpp20_led_sw_control_read_leds(drv_data->dev, &drv_data->led_infos.leds);
+	if (rc > 0) {
+		drv_data->num_leds = rc;
+		rc = 0;
+	}
+
+	return rc;
 }
 
 static void
-hidpp20drv_read_profile_8100(struct ratbag_profile *profile, unsigned int index)
+hidpp20drv_read_profile_8100(struct ratbag_profile *profile)
 {
 	struct ratbag_device *device = profile->device;
 	struct hidpp20drv_data *drv_data = ratbag_get_drv_data(device);
 	struct ratbag_resolution *res;
 	struct hidpp20_profile *p;
-	unsigned i, dpi = 0;
+	int dpi_index = 0xff;
 
-	hidpp20drv_read_onboard_profile(device, profile->index);
+	profile->is_enabled = drv_data->profiles->profiles[profile->index].enabled;
 
 	profile->is_active = false;
-	if ((int)index == hidpp20drv_current_profile(device))
+	if ((int)profile->index == hidpp20drv_current_profile(device))
 		profile->is_active = true;
 
-	/* retrieve the resolution through 220X as the profile doesn't has it */
 	if (profile->is_active)
-		hidpp20drv_read_resolution_dpi(profile);
+		dpi_index = hidpp20_onboard_profiles_get_current_dpi_index(drv_data->dev);
 
-	dpi = ratbag_resolution_get_dpi(profile->resolution.modes);
+	if (dpi_index < 0)
+		dpi_index = 0xff;
 
-	p = &drv_data->profiles->profiles[index];
+	p = &drv_data->profiles->profiles[profile->index];
 
-	for (i = 0; i < profile->resolution.num_modes; i++) {
-		res = ratbag_resolution_init(profile, i,
-					     p->dpi[i],
-					     p->dpi[i],
-					     p->report_rate);
+	ratbag_profile_for_each_resolution(profile, res) {
+		struct hidpp20_sensor *sensor;
+
+		/* We only look at the first sensor. Multiple
+		 * sensors is too niche to care about right now */
+		sensor = &drv_data->sensors[0];
+
+		ratbag_resolution_set_resolution(res,
+						 p->dpi[res->index],
+						 p->dpi[res->index],
+						 p->report_rate);
 
 		if (profile->is_active &&
-		    res->dpi_x == dpi)
+		    res->index == (unsigned int)dpi_index)
 			res->is_active = true;
-		if (i == p->default_dpi)
+		if (res->index == p->default_dpi) {
 			res->is_default = true;
+			if (!profile->is_active || dpi_index < 0 || dpi_index > 4)
+				res->is_active = true;
+		}
+
+		ratbag_resolution_set_dpi_list_from_range(res,
+							  sensor->dpi_min,
+							  sensor->dpi_max);
+		ratbag_resolution_set_report_rate_list(res,
+						       drv_data->report_rates,
+						       drv_data->num_report_rates);
 	}
 }
 
 static void
-hidpp20drv_read_profile(struct ratbag_profile *profile, unsigned int index)
+hidpp20drv_read_profile(struct ratbag_profile *profile)
 {
 	struct ratbag_device *device = profile->device;
 	struct hidpp20drv_data *drv_data = ratbag_get_drv_data(device);
+	struct ratbag_led *led;
+	struct ratbag_button *button;
 
-	if (drv_data->capabilities & HIDPP_CAP_ONBOARD_PROFILES_8100)
-		return hidpp20drv_read_profile_8100(profile, index);
+	if (drv_data->capabilities & HIDPP_CAP_ONBOARD_PROFILES_8100) {
+		hidpp20drv_read_profile_8100(profile);
+	} else {
+		hidpp20drv_read_resolution_dpi(profile);
+		hidpp20drv_read_special_key_mouse(device);
 
-	hidpp20drv_read_resolution_dpi(profile);
-	hidpp20drv_read_special_key_mouse(device);
+		profile->is_active = (profile->index == 0);
+	}
 
-	profile->is_active = (index == 0);
+	ratbag_profile_for_each_led(profile, led)
+		hidpp20drv_read_led(led);
+
+	ratbag_profile_for_each_button(profile, button)
+		hidpp20drv_read_button(button);
 }
 
 static int
@@ -757,6 +1075,8 @@ hidpp20drv_init_feature(struct ratbag_device *device, uint16_t feature)
 				    &feature_index,
 				    &feature_type,
 				    &feature_version);
+	if (rc < 0)
+		return rc;
 
 	switch (feature) {
 	case HIDPP_PAGE_ROOT:
@@ -765,6 +1085,7 @@ hidpp20drv_init_feature(struct ratbag_device *device, uint16_t feature)
 		break;
 	case HIDPP_PAGE_MOUSE_POINTER_BASIC: {
 		drv_data->capabilities |= HIDPP_CAP_RESOLUTION_2200;
+		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_RESOLUTION);
 		break;
 	}
 	case HIDPP_PAGE_ADJUSTABLE_DPI: {
@@ -774,6 +1095,7 @@ hidpp20drv_init_feature(struct ratbag_device *device, uint16_t feature)
 		rc = hidpp20drv_read_resolution_dpi_2201(device);
 		if (rc < 0)
 			return 0; /* this is not a hard failure */
+		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_RESOLUTION);
 		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_SWITCHABLE_RESOLUTION);
 		drv_data->capabilities |= HIDPP_CAP_SWITCHABLE_RESOLUTION_2201;
 		break;
@@ -781,11 +1103,12 @@ hidpp20drv_init_feature(struct ratbag_device *device, uint16_t feature)
 	case HIDPP_PAGE_SPECIAL_KEYS_BUTTONS: {
 		log_debug(ratbag, "device has programmable keys/buttons\n");
 		drv_data->capabilities |= HIDPP_CAP_BUTTON_KEY_1b04;
+		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_BUTTON);
 		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_BUTTON_KEY);
 		/* we read the profile once to get the correct number of
 		 * supported buttons. */
 		if (!hidpp20drv_read_special_key_mouse(device))
-			device->num_buttons = drv_data->num_controls;
+			drv_data->num_buttons = drv_data->num_controls;
 		break;
 	}
 	case HIDPP_PAGE_BATTERY_LEVEL_STATUS: {
@@ -810,11 +1133,19 @@ hidpp20drv_init_feature(struct ratbag_device *device, uint16_t feature)
 		/* we read the profile once to get the correct number of
 		 * supported buttons. */
 		if (!hidpp20drv_read_kbd_reprogrammable_key(device))
-			device->num_buttons = drv_data->num_controls;
+			drv_data->num_buttons = drv_data->num_controls;
 		break;
 	}
 	case HIDPP_PAGE_ADJUSTABLE_REPORT_RATE: {
 		log_debug(ratbag, "device has adjustable report rate\n");
+
+		/* we read the profile once to get the correct number of
+		 * supported report rates. */
+		rc = hidpp20drv_read_report_rate_8060(device);
+		if (rc < 0)
+			return 0; /* this is not a hard failure */
+
+		drv_data->capabilities |= HIDPP_CAP_ADJUSTIBLE_REPORT_RATE_8060;
 		break;
 	}
 	case HIDPP_PAGE_COLOR_LED_EFFECTS: {
@@ -830,18 +1161,42 @@ hidpp20drv_init_feature(struct ratbag_device *device, uint16_t feature)
 		device->num_leds = drv_data->num_leds;
 		break;
 	}
+	case HIDPP_PAGE_LED_SW_CONTROL: {
+		log_debug(ratbag, "device has non-rgb leds\n");
+		drv_data->capabilities |= HIDPP_CAP_LED_SW_CONTROL_1300;
+
+		if (drv_data->capabilities & HIDPP_PAGE_COLOR_LED_EFFECTS)
+			return 0;
+
+		if (hidpp20drv_read_leds(device))
+			return 0;
+
+		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_LED);
+		break;
+	}
 	case HIDPP_PAGE_ONBOARD_PROFILES: {
 		log_debug(ratbag, "device has onboard profiles\n");
 		drv_data->capabilities |= HIDPP_CAP_ONBOARD_PROFILES_8100;
+		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_PROFILE);
+		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_DISABLE_PROFILE);
+		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_BUTTON);
+		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_BUTTON_KEY);
+		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_BUTTON_MACROS);
 
 		rc = hidpp20_onboard_profiles_allocate(drv_data->dev, &drv_data->profiles);
 		if (rc < 0)
 			return rc;
 
+		rc = hidpp20_onboard_profiles_initialize(drv_data->dev, drv_data->profiles);
+		if (rc < 0)
+			return rc;
+
 		drv_data->num_profiles = drv_data->profiles->num_profiles;
 		drv_data->num_resolutions = drv_data->profiles->num_modes;
 		drv_data->num_buttons = drv_data->profiles->num_buttons;
-		drv_data->num_leds = drv_data->profiles->num_leds;
+		/* We ignore the profile's num_leds and require
+		 * HIDPP_PAGE_COLOR_LED_EFFECTS to succeed instead
+		 */
 
 		break;
 	}
@@ -856,7 +1211,7 @@ hidpp20drv_init_feature(struct ratbag_device *device, uint16_t feature)
 }
 
 static int
-hidpp20_commit(struct ratbag_device *device)
+hidpp20drv_commit(struct ratbag_device *device)
 {
 	struct hidpp20drv_data *drv_data = ratbag_get_drv_data(device);
 	struct ratbag_profile *profile;
@@ -864,14 +1219,15 @@ hidpp20_commit(struct ratbag_device *device)
 	struct ratbag_led *led;
 	struct ratbag_resolution *resolution;
 	int rc;
-	unsigned int i;
+	unsigned int dpi_index = 0;
 
 	list_for_each(profile, &device->profiles, link) {
 		if (!profile->dirty)
 			continue;
 
-		for (i = 0; i < profile->resolution.num_modes; i++) {
-			resolution = &profile->resolution.modes[i];
+		ratbag_profile_for_each_resolution(profile, resolution) {
+			if (resolution->is_active)
+				dpi_index = resolution->index;
 
 			rc = hidpp20drv_update_resolution_dpi(resolution,
 							      resolution->dpi_x,
@@ -881,12 +1237,10 @@ hidpp20_commit(struct ratbag_device *device)
 		}
 
 		list_for_each(button, &profile->buttons, link) {
-			struct ratbag_button_action action = button->action;
-
 			if (!button->dirty)
 				continue;
 
-			rc = hidpp20drv_update_button(button, &action);
+			rc = hidpp20drv_update_button(button);
 			if (rc)
 				return RATBAG_ERROR_DEVICE;
 		}
@@ -900,10 +1254,22 @@ hidpp20_commit(struct ratbag_device *device)
 				return RATBAG_ERROR_DEVICE;
 		}
 
-		if (drv_data->capabilities & HIDPP_CAP_ONBOARD_PROFILES_8100)
-			hidpp20_onboard_profiles_write(drv_data->dev,
-						       profile->index,
-						       drv_data->profiles);
+	}
+
+	if (drv_data->capabilities & HIDPP_CAP_ONBOARD_PROFILES_8100) {
+		list_for_each(profile, &device->profiles, link)
+			drv_data->profiles->profiles[profile->index].enabled = profile->is_enabled;
+
+		rc = hidpp20_onboard_profiles_commit(drv_data->dev,
+						     drv_data->profiles);
+		if (rc)
+			return RATBAG_ERROR_DEVICE;
+
+		list_for_each(profile, &device->profiles, link) {
+			if (profile->is_active)
+				hidpp20_onboard_profiles_set_current_dpi_index(drv_data->dev,
+									       dpi_index);
+		}
 	}
 
 	return RATBAG_SUCCESS;
@@ -954,18 +1320,17 @@ static void
 hidpp20drv_remove(struct ratbag_device *device)
 {
 	struct hidpp20drv_data *drv_data;
-	struct hidpp20_device *dev;
 
 	if (!device)
 		return;
 
 	drv_data = ratbag_get_drv_data(device);
-	dev = drv_data->dev;
 
 	ratbag_close_hidraw(device);
 
 	if (drv_data->profiles)
-		hidpp20_onboard_profiles_destroy(dev, drv_data->profiles);
+		hidpp20_onboard_profiles_destroy(drv_data->profiles);
+	free(drv_data->led_infos.color_leds);
 	free(drv_data->controls);
 	free(drv_data->sensors);
 	if (drv_data->dev)
@@ -973,6 +1338,22 @@ hidpp20drv_remove(struct ratbag_device *device)
 	free(drv_data);
 }
 
+static void
+hidpp20drv_init_device(struct ratbag_device *device,
+		       struct hidpp20drv_data *drv_data)
+{
+	struct ratbag_profile *profile;
+
+	ratbag_device_init_profiles(device,
+				    drv_data->num_profiles,
+				    drv_data->num_resolutions,
+				    drv_data->num_buttons,
+				    drv_data->num_leds);
+
+	ratbag_device_for_each_profile(device, profile)
+		hidpp20drv_read_profile(profile);
+}
+
 static int
 hidpp20drv_probe(struct ratbag_device *device)
 {
@@ -980,9 +1361,7 @@ hidpp20drv_probe(struct ratbag_device *device)
 	struct hidpp20drv_data *drv_data;
 	struct hidpp_device base;
 	struct hidpp20_device *dev;
-	const char *prop;
 	int device_idx = HIDPP_RECEIVER_IDX;
-	int nread = 0;
 
 	rc = ratbag_find_hidraw(device, hidpp20drv_test_hidraw);
 	if (rc)
@@ -990,28 +1369,20 @@ hidpp20drv_probe(struct ratbag_device *device)
 
 	drv_data = zalloc(sizeof(*drv_data));
 	ratbag_set_drv_data(device, drv_data);
-	hidpp_device_init(&base, device->hidraw.fd);
+	hidpp_device_init(&base, device->hidraw[0].fd);
 	hidpp_device_set_log_handler(&base, hidpp20_log, HIDPP_LOG_PRIORITY_RAW, device);
 
-	prop = ratbag_device_get_udev_property(device, "RATBAG_HIDPP20_INDEX");
-	if (prop) {
-		sscanf(prop, "%d%n", &device_idx, &nread);
-		if (!nread || (prop[nread]) != '\0' || device_idx < 0) {
-			log_error(device->ratbag,
-				  "Error parsing RATBAG_HIDPP20_INDEX: '%s' for %s\n",
-				  prop,
-				  device->name);
-			device_idx = HIDPP_RECEIVER_IDX;
-		}
-	}
+	device_idx = ratbag_device_data_hidpp20_get_index(device->data);
+	if (device_idx == -1)
+		device_idx = HIDPP_RECEIVER_IDX;
 
 	/* In the general case, we can treat all devices as wired devices
 	 * here. If we talk to the correct hidraw device the kernel adjusts
 	 * the device index for us, so even for unifying receiver devices
 	 * we can just use 0xff as device index.
 	 *
-	 * If there is a special need like for G900, we can pass a
-	 * udev prop RATBAG_HIDPP20_INDEX.
+	 * If there is a special need like for G900, we can add this in the
+	 * device data file.
 	 */
 	dev = hidpp20_device_new(&base, device_idx);
 	if (!dev) {
@@ -1027,17 +1398,13 @@ hidpp20drv_probe(struct ratbag_device *device)
 	drv_data->num_profiles = 1;
 	drv_data->num_resolutions = 1;
 	drv_data->num_buttons = 8;
-	drv_data->num_leds = 2;
+	drv_data->num_leds = 0;
 
 	rc = hidpp20drv_20_probe(device);
 	if (rc)
 		goto err;
 
-	ratbag_device_init_profiles(device,
-				    drv_data->num_profiles,
-				    drv_data->num_resolutions,
-				    drv_data->num_buttons,
-				    drv_data->num_leds);
+	hidpp20drv_init_device(device, drv_data);
 
 	return rc;
 err:
@@ -1050,9 +1417,6 @@ struct ratbag_driver hidpp20_driver = {
 	.id = "hidpp20",
 	.probe = hidpp20drv_probe,
 	.remove = hidpp20drv_remove,
-	.commit = hidpp20_commit,
-	.read_profile = hidpp20drv_read_profile,
+	.commit = hidpp20drv_commit,
 	.set_active_profile = hidpp20drv_set_current_profile,
-	.read_button = hidpp20drv_read_button,
-	.read_led = hidpp20drv_read_led,
 };
