@@ -29,32 +29,6 @@
 #include "libratbag-private.h"
 #include "libratbag-test.h"
 
-static void
-test_read_profile(struct ratbag_profile *profile, unsigned int index)
-{
-	struct ratbag_test_device *d = ratbag_get_drv_data(profile->device);
-	struct ratbag_test_profile *p;
-	struct ratbag_test_resolution *r;
-	unsigned int i;
-
-	assert(index < d->num_profiles);
-
-	p = &d->profiles[index];
-	for (i = 0; i < d->num_resolutions; i++) {
-		struct ratbag_resolution *res;
-
-		r = &p->resolutions[i];
-		res = ratbag_resolution_init(profile, i, r->xres, r->yres, r->hz);
-		res->is_active = r->active;
-		res->is_default = r->dflt;
-		res->capabilities = r->caps;
-		res->hz = r->hz;
-		assert(res);
-	}
-
-	profile->is_active = p->active;
-}
-
 static int
 test_write_profile(struct ratbag_profile *profile)
 {
@@ -80,11 +54,15 @@ test_read_button(struct ratbag_button *button)
 	struct ratbag_device *device = button->profile->device;
 	struct ratbag_test_device *d = ratbag_get_drv_data(device);
 	struct ratbag_test_profile *p = &d->profiles[button->profile->index];
+	struct ratbag_test_button *b = &p->buttons[button->index];
 	struct ratbag_button_macro *m;
-	const char data[] = "TEST";
-	const char *c;
+	struct ratbag_test_macro_event *e;
+	int idx;
 
-	switch (p->buttons[button->index].type) {
+	/* Only take the button types from the first profile */
+	button->type = d->profiles[0].buttons[button->index].button_type;
+
+	switch (b->action_type) {
 	case RATBAG_BUTTON_ACTION_TYPE_BUTTON:
 		button->action.type = RATBAG_BUTTON_ACTION_TYPE_BUTTON;
 		button->action.action.button = p->buttons[button->index].button;
@@ -97,21 +75,19 @@ test_read_button(struct ratbag_button *button)
 		button->action.type = RATBAG_BUTTON_ACTION_TYPE_MACRO;
 		m = ratbag_button_macro_new("test macro");
 
-		ARRAY_FOR_EACH(data, c) {
-			char str[6];
-			snprintf_safe(str, 6, "KEY_%c", *c);
-			ratbag_button_macro_set_event(m,
-						      0,
-						      RATBAG_MACRO_EVENT_KEY_PRESSED,
-						      libevdev_event_code_from_name(EV_KEY, str));
-			ratbag_button_macro_set_event(m,
-						      0,
-						      RATBAG_MACRO_EVENT_KEY_RELEASED,
-						      libevdev_event_code_from_name(EV_KEY, str));
+		idx = 0;
+		ARRAY_FOR_EACH(b->macro, e) {
+			if (e->type == RATBAG_MACRO_EVENT_NONE)
+				break;
+			ratbag_button_macro_set_event(m, idx++, e->type, e->value);
 		}
 		ratbag_button_copy_macro(button, m);
 		ratbag_button_macro_unref(m);
 		break;
+	case RATBAG_BUTTON_ACTION_TYPE_SPECIAL:
+		button->action.type = RATBAG_BUTTON_ACTION_TYPE_SPECIAL;
+		button->action.action.special = p->buttons[button->index].special;
+		break;
 	default:
 		button->action.type = RATBAG_BUTTON_ACTION_TYPE_UNKNOWN;
 	}
@@ -130,6 +106,11 @@ test_read_led(struct ratbag_led *led)
 	struct ratbag_test_profile *p = &d->profiles[led->profile->index];
 	struct ratbag_test_led t_led = p->leds[led->index];
 
+	ratbag_led_set_mode_capability(led, RATBAG_LED_ON);
+	ratbag_led_set_mode_capability(led, RATBAG_LED_CYCLE);
+	ratbag_led_set_mode_capability(led, RATBAG_LED_BREATHING);
+	ratbag_led_set_mode_capability(led, RATBAG_LED_OFF);
+
 	switch (t_led.mode) {
 	case RATBAG_LED_ON:
 		led->mode = RATBAG_LED_ON;
@@ -146,8 +127,13 @@ test_read_led(struct ratbag_led *led)
 	led->color.red = t_led.color.red;
 	led->color.green = t_led.color.green;
 	led->color.blue = t_led.color.blue;
-	led->hz = t_led.hz;
+	led->ms = t_led.ms;
 	led->brightness = t_led.brightness;
+
+	/* The led type has to be the same anyway so make writing test
+	 * devices easier by always getting the first profile's LED type.
+	 */
+	led->type = d->profiles[0].leds[led->index].type;
 }
 
 static int
@@ -162,7 +148,7 @@ test_write_button(struct ratbag_button *button,
 static int
 test_write_led(struct ratbag_led *led,
 	       enum ratbag_led_mode mode,
-	       struct ratbag_color color, unsigned int hz,
+	       struct ratbag_color color, unsigned int ms,
 	       unsigned int brightness)
 {
 	/* check if the device is still valid */
@@ -170,16 +156,107 @@ test_write_led(struct ratbag_led *led,
 	return 0;
 }
 
+static const char* test_get_svg_name(const struct ratbag_device *device)
+{
+	struct ratbag_test_device *d = ratbag_get_drv_data(device);
+
+	return d->svg;
+}
+
 static int
 test_fake_probe(struct ratbag_device *device)
 {
 	return -ENODEV;
 }
 
-static int
-test_probe(struct ratbag_device *device, void *data)
+static void
+test_read_profile(struct ratbag_profile *profile)
 {
-	struct ratbag_test_device *test_device = data;
+	struct ratbag_test_device *d = ratbag_get_drv_data(profile->device);
+	struct ratbag_test_profile *p;
+	struct ratbag_test_profile *p0;
+	struct ratbag_test_resolution *r;
+	struct ratbag_test_resolution *r0;
+	struct ratbag_button *button;
+	struct ratbag_led *led;
+	unsigned int i;
+	bool active_set = false;
+	bool default_set = false;
+
+	assert(profile->index < d->num_profiles);
+
+	p = &d->profiles[profile->index];
+	p0 = &d->profiles[0];
+	r0 = &p0->resolutions[0];
+
+	for (i = 0; i < d->num_resolutions; i++) {
+		_cleanup_resolution_ struct ratbag_resolution *res = NULL;
+		size_t nrates = 0;
+
+		r = &p->resolutions[i];
+
+		res = ratbag_profile_get_resolution(profile, i);
+		assert(res);
+		ratbag_resolution_set_resolution(res, r->xres, r->yres, r->hz);
+		if (r0->dpi_min != 0 && r0->dpi_max != 0)
+			ratbag_resolution_set_dpi_list_from_range(res,
+								  r0->dpi_min,
+								  r0->dpi_max);
+
+		for (size_t r = 0; r < ARRAY_LENGTH(r0->report_rates); r++) {
+			if (r0->report_rates[r] > 0)
+				nrates++;
+		}
+
+		if (nrates > 0)
+			ratbag_resolution_set_report_rate_list(res,
+							       r0->report_rates,
+							       nrates);
+
+		res->is_active = r->active;
+		if (r->active)
+			active_set = true;
+		res->is_default = r->dflt;
+		if (r->dflt)
+			default_set = true;
+		ratbag_resolution_set_cap(res, r->caps);
+		res->hz = r->hz;
+	}
+
+	/* special case triggered by the test suite when num_resolutions is 0 */
+	if (d->num_resolutions) {
+		_cleanup_resolution_ struct ratbag_resolution *res = NULL;
+
+		res = ratbag_profile_get_resolution(profile, 0);
+		assert(res);
+		if (!active_set)
+			res->is_active = true;
+		if (!default_set)
+			res->is_default = true;
+	}
+
+	ratbag_profile_for_each_button(profile, button)
+		test_read_button(button);
+
+	ratbag_profile_for_each_led(profile, led)
+		test_read_led(led);
+
+	profile->is_active = p->active;
+	profile->is_enabled = !p->disabled;
+
+	for (i = 0; i < ARRAY_LENGTH(p->caps) && p->caps[i]; i++) {
+		ratbag_profile_set_cap(profile, p->caps[i]);
+	}
+}
+
+static int
+test_probe(struct ratbag_device *device, const void *data)
+{
+	struct ratbag_test_device *test_device;
+	struct ratbag_profile *profile;
+
+	test_device = zalloc(sizeof(*test_device));
+	memcpy(test_device, data, sizeof(*test_device));
 
 	ratbag_set_drv_data(device, test_device);
 	ratbag_device_init_profiles(device,
@@ -187,8 +264,14 @@ test_probe(struct ratbag_device *device, void *data)
 				    test_device->num_resolutions,
 				    test_device->num_buttons,
 				    test_device->num_leds);
-	if (test_device->num_profiles > 1)
+
+	if (test_device->num_profiles > 1) {
 		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_BUTTON_MACROS);
+		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_DISABLE_PROFILE);
+	}
+
+	ratbag_device_for_each_profile(device, profile)
+		test_read_profile(profile);
 
 	return 0;
 }
@@ -203,6 +286,7 @@ test_remove(struct ratbag_device *device)
 	if (d->destroyed)
 		d->destroyed(device, d->destroyed_data);
 	ratbag_set_drv_data(device, NULL);
+	free(d);
 }
 
 struct ratbag_driver test_driver = {
@@ -211,11 +295,9 @@ struct ratbag_driver test_driver = {
 	.probe = test_fake_probe,
 	.test_probe = test_probe,
 	.remove = test_remove,
-	.read_profile = test_read_profile,
+	.get_svg_name = test_get_svg_name,
 	.write_profile = test_write_profile,
 	.set_active_profile = test_set_active_profile,
-	.read_button = test_read_button,
-	.read_led = test_read_led,
 	.write_button = test_write_button,
 	.write_resolution_dpi = NULL,
 	.write_led = test_write_led,
