@@ -31,6 +31,7 @@
 
 #include "config.h"
 
+#include <assert.h>
 #include <linux/types.h>
 #include <errno.h>
 #include <stdio.h>
@@ -85,62 +86,51 @@ const char *device_types[0xFF] = {
 	[0x0A ... 0xFE] = NULL,
 };
 
+unsigned int
+hidpp10_dpi_table_get_max_dpi(struct hidpp10_device *dev)
+{
+	struct hidpp10_dpi_mapping *dpi;
+
+	assert(dev->dpi_count > 0);
+
+	/* We assume a sorted list */
+	dpi = &dev->dpi_table[dev->dpi_count - 1];
+
+	return dpi->dpi;
+}
+
+unsigned int
+hidpp10_dpi_table_get_min_dpi(struct hidpp10_device *dev)
+{
+	struct hidpp10_dpi_mapping *dpi;
+
+	assert(dev->dpi_count > 0);
+
+	/* We assume a sorted list, index 0 is always 0 */
+	dpi = &dev->dpi_table[1];
+
+	return dpi->dpi;
+}
+
 int
 hidpp10_build_dpi_table_from_list(struct hidpp10_device *dev,
-				  const char *str_list)
+				  const struct dpi_list *list)
 {
-	unsigned int i, count, index;
-	int nread, dpi = 0;
-	char c;
-	/*
-	 * str_list is in the form:
-	 * "0;200;400;600;800;1000;1200"
-	 */
+	size_t i;
 
-	/* first, count how many elements do we have in the table */
-	count = 1;
-	i = 0;
-	while (str_list[i] != 0) {
-		c = str_list[i++];
-		if (c == ';')
-			count++;
-	}
-
-	index = 0;
-
-	/* check that the max raw value fits in a uint8_t */
-	if (count + 0x80 - 1> 0xff)
-		return -EINVAL;
-
-	dev->dpi_count = count;
-	dev->dpi_table = zalloc(count * sizeof(*dev->dpi_table));
-
-	while (*str_list != 0 && index < count) {
-		if (*str_list == ';') {
-			str_list++;
-			continue;
-		}
-
-		nread = 0;
-		sscanf(str_list, "%d%n", &dpi, &nread);
-		if (!nread || dpi < 0)
-			goto err;
-
-		dev->dpi_table[index].raw_value = index + 0x80;
-		dev->dpi_table[index].dpi = dpi;
-
-		str_list += nread;
-		index++;
-	}
-
-	if (index != count)
+	if (list->nentries + 0x80 - 1> 0xff)
 		goto err;
 
-	/* mark the state as invalid */
-	dev->profile_count = 0;
+	dev->dpi_count = list->nentries;
+	dev->dpi_table = zalloc(list->nentries * sizeof(*dev->dpi_table));
+	dev->dpi_table_is_range = false;
+
+	for (i = 0; i < list->nentries; i++) {
+		dev->dpi_table[i].raw_value = i + 0x80;
+		dev->dpi_table[i].dpi = list->entries[i];
+	}
 
 	return 0;
-
 err:
 	dev->dpi_count = 0;
 	free(dev->dpi_table);
@@ -150,35 +140,24 @@ err:
 
 int
 hidpp10_build_dpi_table_from_dpi_info(struct hidpp10_device *dev,
-				      const char *str_dpi)
+				      const struct dpi_range *range)
 {
-	float min, max, step;
 	unsigned raw_max, i;
-	int rc;
-	/*
-	 * str_list is in the form:
-	 * "MIN:MAX@STEP"
-	 */
 
-	rc = sscanf(str_dpi, "%f:%f@%f", &min, &max, &step);
-	if (rc != 3)
-		return -EINVAL;
 
-	raw_max = (max - min) / step;
+	raw_max = (range->max - range->min) / range->step;
 	if (raw_max > 0xff)
 		return -EINVAL;
 
 	dev->dpi_count = raw_max + 1;
 	dev->dpi_table = zalloc((raw_max + 1) * sizeof(*dev->dpi_table));
+	dev->dpi_table_is_range = true;
 
 	for (i = 1; i <= raw_max; i++) {
 		dev->dpi_table[i].raw_value = i;
-		dev->dpi_table[i].dpi = round((min + step * i) / 25.0f) * 25;
+		dev->dpi_table[i].dpi = round((range->min + range->step * i) / 25.0f) * 25;
 	}
 
-	/* mark the state as invalid */
-	dev->profile_count = 0;
-
 	return 0;
 }
 
@@ -247,7 +226,7 @@ hidpp10_request_command(struct hidpp10_device *dev, union hidpp10_message *msg)
 				"    HID++ error from the %s (%d): %s (%02x)\n",
 				read_buffer.msg.device_idx == HIDPP_RECEIVER_IDX ? "receiver" : "device",
 				read_buffer.msg.device_idx,
-				hidpp_errors[hidpp_err] ? hidpp_errors[hidpp_err] : "Undocumented error code",
+				hidpp10_errors[hidpp_err] ? hidpp10_errors[hidpp_err] : "Undocumented error code",
 				hidpp_err);
 			break;
 		}
@@ -664,7 +643,7 @@ hidpp10_get_dpi_mapping(struct hidpp10_device *dev, unsigned int value)
 	min_delta = INT_MAX;
 
 	for (i = 0; i < dev->dpi_count; i++) {
-		delta = abs(value - m[i].dpi);
+		delta = abs((int)value - (int)m[i].dpi);
 		if (delta < min_delta) {
 			result = m[i].raw_value;
 			min_delta = delta;
@@ -693,29 +672,83 @@ hidpp10_get_dpi_value(struct hidpp10_device *dev, uint8_t raw_value)
 	return 0;
 }
 
-int
-hidpp10_get_profile_directory(struct hidpp10_device *dev,
-			      struct hidpp10_directory *out,
-			      size_t nelems)
+static int
+hidpp10_write_profile_directory(struct hidpp10_device *dev)
 {
-	unsigned int i;
+	unsigned int i, index;
 	int res;
-	uint8_t bytes[HIDPP10_PAGE_SIZE] = { 0 };
+	uint8_t bytes[HIDPP10_PAGE_SIZE];
 	struct hidpp10_directory *directory = (struct hidpp10_directory *)bytes;
-	size_t count;
+	uint16_t crc;
 
 	if (dev->profile_type == HIDPP10_PROFILE_UNKNOWN) {
 		hidpp_log_debug(&dev->base, "no profile type given\n");
 		return 0;
 	}
 
-	if (dev->profile_directory) {
-		count = 0;
-		while (dev->profile_directory[count].page) {
-			count++;
-		}
+	memset(bytes, 0xff, sizeof(bytes));
 
-		goto out;
+	index = 0;
+	for (i = 0; i < dev->profile_count; i++) {
+		if (!dev->profiles[i].enabled)
+			continue;
+
+		directory[index].page = dev->profiles[i].page;
+		directory[index].offset = dev->profiles[i].offset;
+		directory[index].led_mask = ((0b111 << index) >> 2) & 0b111;
+		++index;
+	}
+
+	crc = hidpp_crc_ccitt(bytes, HIDPP10_PAGE_SIZE - 2);
+	hidpp_set_unaligned_be_u16(&bytes[HIDPP10_PAGE_SIZE - 2], crc);
+
+	res = hidpp10_send_hot_payload(dev,
+				       0x00, 0x0000, /* destination: RAM */
+				       bytes,
+				       HIDPP10_PAGE_SIZE / 2);
+	if (res < 0)
+		return res;
+
+	res = hidpp10_erase_memory(dev, 0x01);
+	if (res < 0)
+		return res;
+
+	res = hidpp10_write_flash(dev,
+				  0x00, 0x0000,
+				  0x01, 0x0000,
+				  HIDPP10_PAGE_SIZE / 2);
+	if (res < 0)
+		return res;
+
+	res = hidpp10_send_hot_payload(dev,
+				       0x00, 0x0000, /* destination: RAM */
+				       bytes + HIDPP10_PAGE_SIZE / 2,
+				       HIDPP10_PAGE_SIZE / 2);
+	if (res < 0)
+		return res;
+
+	res = hidpp10_write_flash(dev,
+				  0x00, 0x0000,
+				  0x01, HIDPP10_PAGE_SIZE / 2,
+				  HIDPP10_PAGE_SIZE / 2);
+	if (res < 0)
+		return res;
+
+	return 0;
+}
+
+static int
+hidpp10_read_profile_directory(struct hidpp10_device *dev)
+{
+	unsigned int i;
+	int res;
+	uint8_t bytes[HIDPP10_PAGE_SIZE] = { 0 };
+	struct hidpp10_directory *directory = (struct hidpp10_directory *)bytes;
+	unsigned int count;
+
+	if (dev->profile_type == HIDPP10_PROFILE_UNKNOWN) {
+		hidpp_log_debug(&dev->base, "no profile type given\n");
+		return 0;
 	}
 
 	hidpp_log_raw(&dev->base, "Fetching the profiles' directory\n");
@@ -725,46 +758,29 @@ hidpp10_get_profile_directory(struct hidpp10_device *dev,
 		return res;
 
 	count = 0;
-	/* assume 16 profiles max */
-	for (i = 0; i < 16; i++) {
+	for (i = 0; i < dev->profile_count; i++) {
 		if (directory[i].page == 0xFF)
 			break;
+		dev->profiles[i].page = directory[i].page;
+		dev->profiles[i].offset = directory[i].offset;
+		dev->profiles[i].enabled = true;
 		count++;
 	}
 
-	dev->profile_directory = zalloc((count + 1) * sizeof(struct hidpp10_directory));
-	memcpy(dev->profile_directory, directory, count  * sizeof(struct hidpp10_directory));
-
-out:
-	if (dev->profile_count != count) {
-		if (dev->profiles)
-			free(dev->profiles);
-		dev->profiles = zalloc(count * sizeof(struct hidpp10_profile));
-		dev->profile_count = count;
-	}
-	count = min(count, nelems);
-	memcpy(out, dev->profile_directory, count  * sizeof(out[0]));
+	for (i = count; i < dev->profile_count; i++)
+		dev->profiles[i].enabled = false;
 
 	return count;
 }
 
 int
-hidpp10_get_current_profile(struct hidpp10_device *dev, int8_t *current_profile)
+hidpp10_get_current_profile(struct hidpp10_device *dev, uint8_t *current_profile)
 {
 	unsigned idx = dev->index;
 	union hidpp10_message profile = CMD_PROFILE(idx, GET_REGISTER_REQ);
 	int res;
 	unsigned i;
-	int8_t type, page, offset;
-	struct hidpp10_directory directory[16]; /* completely random profile count */
-	int count = 0;
-
-	hidpp_log_raw(&dev->base, "Fetching the profiles' directory\n");
-
-	count = hidpp10_get_profile_directory(dev, directory,
-					    ARRAY_LENGTH(directory));
-	if (count < 0)
-		return count;
+	uint8_t type, page, offset;
 
 	hidpp_log_raw(&dev->base, "Fetching current profile\n");
 
@@ -779,14 +795,14 @@ hidpp10_get_current_profile(struct hidpp10_device *dev, int8_t *current_profile)
 		*current_profile = page;
 		/* If the profile exceeds the directory length, default to
 		 * the first */
-		if (*current_profile > count)
+		if (*current_profile > dev->profile_count)
 			*current_profile = 0;
 		return 0;
 	case PROFILE_TYPE_ADDRESS:
 		offset = profile.msg.parameters[2];
-		for (i = 0; i < ARRAY_LENGTH(directory) && directory[i].page < 32; i++) {
-			if (page == directory[i].page &&
-			    offset == directory[i].offset) {
+		for (i = 0; i < dev->profile_count; i++) {
+			if (page == dev->profiles[i].page &&
+			    offset == dev->profiles[i].offset) {
 				*current_profile = i;
 				return 0;
 			}
@@ -806,21 +822,12 @@ hidpp10_get_current_profile(struct hidpp10_device *dev, int8_t *current_profile)
 
 static int
 hidpp10_set_internal_current_profile(struct hidpp10_device *dev,
-				     int16_t current_profile,
+				     uint16_t current_profile,
 				     uint8_t profile_type)
 {
 	unsigned idx = dev->index;
 	union hidpp10_message profile = CMD_PROFILE(idx, SET_REGISTER_REQ);
 	int8_t page, offset;
-	struct hidpp10_directory directory[16]; /* completely random profile count */
-	int count = 0;
-
-	hidpp_log_raw(&dev->base, "Fetching the profiles' directory\n");
-
-	count = hidpp10_get_profile_directory(dev, directory,
-					    ARRAY_LENGTH(directory));
-	if (count < 0)
-		return count;
 
 	hidpp_log_raw(&dev->base, "Setting current profile\n");
 
@@ -828,7 +835,7 @@ hidpp10_set_internal_current_profile(struct hidpp10_device *dev,
 
 	switch (profile_type) {
 	case PROFILE_TYPE_INDEX:
-		if (current_profile > count)
+		if (current_profile > dev->profile_count)
 			return -EINVAL;
 		profile.msg.parameters[1] = current_profile & 0xFF;
 		break;
@@ -851,7 +858,7 @@ hidpp10_set_internal_current_profile(struct hidpp10_device *dev,
 }
 
 int
-hidpp10_set_current_profile(struct hidpp10_device *dev, int16_t current_profile)
+hidpp10_set_current_profile(struct hidpp10_device *dev, uint16_t current_profile)
 {
 	return hidpp10_set_internal_current_profile(dev,
 						    current_profile,
@@ -1053,7 +1060,6 @@ hidpp10_onboard_profiles_macro_next(struct hidpp10_device *device,
 		break;
 
 	case HIDPP10_MACRO_END:
-		step = 1;
 		return 0;
 	default:
 		if (macro->any.type >= 0x80 && macro->any.type <= 0xFE) {
@@ -1138,8 +1144,7 @@ hidpp10_onboard_profiles_read_macro(struct hidpp10_device *device,
 	return index;
 
 out_err:
-	if (macro)
-		free(macro);
+	free(macro);
 
 	return rc;
 }
@@ -1405,8 +1410,8 @@ hidpp10_profile_set_names(struct hidpp10_device *dev, struct hidpp10_profile *pr
 	}
 }
 
-int
-hidpp10_get_profile(struct hidpp10_device *dev, int8_t number, struct hidpp10_profile *profile_return)
+static int
+hidpp10_read_profile(struct hidpp10_device *dev, uint8_t number)
 {
 	uint8_t page_data[HIDPP10_PAGE_SIZE];
 	union _hidpp10_profile_data *data = (union _hidpp10_profile_data *)page_data;
@@ -1416,9 +1421,7 @@ hidpp10_get_profile(struct hidpp10_device *dev, int8_t number, struct hidpp10_pr
 	size_t i;
 	int res;
 	struct hidpp10_profile *profile;
-	struct hidpp10_directory directory[16]; /* completely random profile count */
 	union _hidpp10_button_binding *buttons;
-	int count = 0;
 	uint8_t page;
 
 	/* Page 0 is RAM
@@ -1432,19 +1435,34 @@ hidpp10_get_profile(struct hidpp10_device *dev, int8_t number, struct hidpp10_pr
 
 	hidpp_log_raw(&dev->base, "Fetching profile %d\n", number);
 
-	count = hidpp10_get_profile_directory(dev, directory,
-					    ARRAY_LENGTH(directory));
-	if (count < 0)
-		return count;
-
-	if (count == 0 || dev->profile_type == HIDPP10_PROFILE_UNKNOWN)
+	if (dev->profile_type == HIDPP10_PROFILE_UNKNOWN)
 		return -ENOTSUP;
 
-	if (number >= count) {
-		hidpp_log_error(&dev->base, "Profile number %d not in the directory.\n", number);
+	if (number >= dev->profile_count) {
+		hidpp_log_error(&dev->base, "Profile number %d is not supported.\n", number);
 		return -EINVAL;
 	}
 
+	profile = &dev->profiles[number];
+	if (!profile->page) {
+		unsigned long pages = 0xffff;
+		unsigned int i;
+
+		/* pages 0 and 1 are ROM and directory so they are reserved */
+		long_clear_bit(&pages, 0);
+		long_clear_bit(&pages, 1);
+
+		for (i = 0; i < dev->profile_count; i++) {
+			uint8_t page = dev->profiles[i].page;
+
+			assert(page < sizeof(pages) * 8);
+			long_clear_bit(&pages, page);
+		}
+
+		page = ffsl(pages) - 1;
+		profile->page = page;
+	}
+
 	switch (dev->profile_type) {
 	case HIDPP10_PROFILE_G500:
 		buttons = p500->buttons;
@@ -1460,19 +1478,18 @@ hidpp10_get_profile(struct hidpp10_device *dev, int8_t number, struct hidpp10_pr
 		return -EINVAL;
 	}
 
-	profile = &dev->profiles[number];
 	if (!profile->initialized) {
-
-		page = directory[number].page;
+		page = profile->page;
 		res = hidpp10_read_page(dev, page, page_data);
 		if (res == -EILSEQ) {
 			/*
 			 * if the CRC is wrong, the mouse still handles the
 			 * profile. Warn the user.
 			 */
-			hidpp_log_info(&dev->base,
-				      "Profile %d has a wrong CRC, assuming valid.\n",
-				      number);
+			if (profile->enabled)
+				hidpp_log_info(&dev->base,
+					      "Profile %d has a wrong CRC, assuming valid.\n",
+					      number);
 			res = 0;
 		}
 		if (res)
@@ -1500,6 +1517,9 @@ hidpp10_get_profile(struct hidpp10_device *dev, int8_t number, struct hidpp10_pr
 			hidpp10_fill_buttons(dev, profile, buttons, PROFILE_NUM_BUTTONS);
 			break;
 		case HIDPP10_PROFILE_G9:
+			profile->red = p9->red;
+			profile->green = p9->green;
+			profile->blue = p9->blue;
 			profile->default_dpi_mode = p9->default_dpi_mode;
 			profile->refresh_rate = p9->usb_refresh_rate ? 1000/p9->usb_refresh_rate : 0;
 
@@ -1571,8 +1591,21 @@ hidpp10_get_profile(struct hidpp10_device *dev, int8_t number, struct hidpp10_pr
 		}
 	}
 
-	*profile_return = *profile;
+	return 0;
+}
 
+int
+hidpp10_get_profile(struct hidpp10_device *dev, uint8_t number, struct hidpp10_profile *profile_return)
+{
+	if (dev->profile_type == HIDPP10_PROFILE_UNKNOWN)
+		return -ENOTSUP;
+
+	if (number >= dev->profile_count) {
+		hidpp_log_error(&dev->base, "Profile number %d is not supported.\n", number);
+		return -EINVAL;
+	}
+
+	*profile_return = dev->profiles[number];
 	return 0;
 }
 
@@ -1622,7 +1655,7 @@ hidpp10_onboard_profiles_get_code_from_special(enum ratbag_button_action_special
 }
 
 int
-hidpp10_set_profile(struct hidpp10_device *dev, int8_t number, struct hidpp10_profile *profile)
+hidpp10_set_profile(struct hidpp10_device *dev, uint8_t number, struct hidpp10_profile *profile)
 {
 	uint8_t page_data[HIDPP10_PAGE_SIZE];
 	union _hidpp10_profile_data *data = (union _hidpp10_profile_data *)page_data;
@@ -1630,27 +1663,24 @@ hidpp10_set_profile(struct hidpp10_device *dev, int8_t number, struct hidpp10_pr
 	struct _hidpp10_profile_700 *p700 = &data->profile_700;
 	struct _hidpp10_profile_9 *p9 = &data->profile_9;
 	int res;
-	struct hidpp10_directory directory[16]; /* completely random profile count */
 	union _hidpp10_button_binding *buttons;
-	int count = 0;
 	uint16_t crc;
 	uint8_t page;
 
 	hidpp_log_raw(&dev->base, "Fetching profile %d\n", number);
 
-	count = hidpp10_get_profile_directory(dev, directory,
-					    ARRAY_LENGTH(directory));
-	if (count < 0)
-		return count;
-
-	if (count == 0 || dev->profile_type == HIDPP10_PROFILE_UNKNOWN)
+	if (dev->profile_type == HIDPP10_PROFILE_UNKNOWN)
 		return -ENOTSUP;
 
-	if (number >= count) {
-		hidpp_log_error(&dev->base, "Profile number %d not in the directory.\n", number);
+	if (number >= dev->profile_count) {
+		hidpp_log_error(&dev->base, "Profile number %d is incorrect.\n", number);
 		return -EINVAL;
 	}
 
+	/* something went wrong */
+	if (!profile->page)
+		return -ENOTSUP;
+
 	memset(page_data, 0xff, sizeof(page_data));
 
 	switch (dev->profile_type) {
@@ -1665,6 +1695,7 @@ hidpp10_set_profile(struct hidpp10_device *dev, int8_t number, struct hidpp10_pr
 		break;
 	default:
 		hidpp_log_error(&dev->base, "This should never happen, complain to your maintainer.\n");
+		return -ENOTSUP;
 	}
 
 
@@ -1675,7 +1706,7 @@ hidpp10_set_profile(struct hidpp10_device *dev, int8_t number, struct hidpp10_pr
 	case HIDPP10_PROFILE_G9:
 		/* we do not know the actual values of the remaining field right now
 		 * so pre-fill with the current data */
-		res = hidpp10_read_page(dev, directory[number].page, page_data);
+		res = hidpp10_read_page(dev, profile->page, page_data);
 		if (res)
 			return res;
 		break;
@@ -1685,6 +1716,7 @@ hidpp10_set_profile(struct hidpp10_device *dev, int8_t number, struct hidpp10_pr
 		break;
 	default:
 		hidpp_log_error(&dev->base, "This should never happen, complain to your maintainer.\n");
+		return -ENOTSUP;
 	}
 
 	switch (dev->profile_type) {
@@ -1702,7 +1734,7 @@ hidpp10_set_profile(struct hidpp10_device *dev, int8_t number, struct hidpp10_pr
 		break;
 	case HIDPP10_PROFILE_G700:
 		p700->default_dpi_mode = profile->default_dpi_mode;
-		p700->usb_refresh_rate = 1000 / profile->refresh_rate;
+		p700->usb_refresh_rate = profile->refresh_rate ? 1000 / profile->refresh_rate : 0;
 
 		hidpp10_write_dpi_modes_8_dual(dev, profile, p700->dpi_modes, PROFILE_NUM_DPI_MODES);
 		hidpp10_write_buttons(dev, profile, buttons, PROFILE_NUM_BUTTONS);
@@ -1721,6 +1753,7 @@ hidpp10_set_profile(struct hidpp10_device *dev, int8_t number, struct hidpp10_pr
 		break;
 	default:
 		hidpp_log_error(&dev->base, "This should never happen, complain to your maintainer.\n");
+		return -ENOTSUP;
 	}
 
 	crc = hidpp_crc_ccitt(page_data, HIDPP10_PAGE_SIZE - 2);
@@ -1743,6 +1776,13 @@ hidpp10_set_profile(struct hidpp10_device *dev, int8_t number, struct hidpp10_pr
 	if (res < 0)
 		return res;
 
+	if (profile->enabled != dev->profiles[number].enabled) {
+		dev->profiles[number].enabled = profile->enabled;
+		res = hidpp10_write_profile_directory(dev);
+		if (res < 0)
+			return res;
+	}
+
 	res = hidpp10_send_hot_payload(dev,
 				       0x00, 0x0000, /* destination: RAM */
 				       page_data,
@@ -1750,7 +1790,7 @@ hidpp10_set_profile(struct hidpp10_device *dev, int8_t number, struct hidpp10_pr
 	if (res < 0)
 		return res;
 
-	page = directory[number].page;
+	page = profile->page;
 	/* according to the spec, a profile can have an offset.
 	 * For all the devices we know, they all start at 0x0000 */
 	res = hidpp10_erase_memory(dev, page);
@@ -1782,6 +1822,7 @@ hidpp10_set_profile(struct hidpp10_device *dev, int8_t number, struct hidpp10_pr
 	if (res < 0)
 		return res;
 
+	dev->profiles[number] = *profile;
 	return res;
 }
 
@@ -2073,12 +2114,24 @@ hidpp10_set_current_resolution(struct hidpp10_device *dev,
 			       uint16_t yres)
 {
 	unsigned idx = dev->index;
-	union hidpp10_message resolution = CMD_CURRENT_RESOLUTION(REPORT_ID_LONG, idx, SET_LONG_REGISTER_REQ);
+	union hidpp10_message resolution = CMD_CURRENT_RESOLUTION(REPORT_ID_SHORT, idx, SET_REGISTER_REQ);
+	union hidpp10_message resolution_long = CMD_CURRENT_RESOLUTION(REPORT_ID_LONG, idx, SET_LONG_REGISTER_REQ);
+	int res;
 
-	hidpp_set_unaligned_le_u16(&resolution.data[4], hidpp10_get_dpi_mapping(dev, xres));
-	hidpp_set_unaligned_le_u16(&resolution.data[6], hidpp10_get_dpi_mapping(dev, yres));
+	switch (dev->profile_type) {
+	case HIDPP10_PROFILE_G9:
+		resolution.data[4] = hidpp10_get_dpi_mapping(dev, xres);
+		res = hidpp10_request_command(dev, &resolution);
+		break;
+        default:
+		hidpp_set_unaligned_le_u16(&resolution_long.data[4], hidpp10_get_dpi_mapping(dev, xres));
+		hidpp_set_unaligned_le_u16(&resolution_long.data[6], hidpp10_get_dpi_mapping(dev, yres));
 
-	return hidpp10_request_command(dev, &resolution);
+		res = hidpp10_request_command(dev, &resolution_long);
+		break;
+        }
+
+	return res;
 }
 
 /* -------------------------------------------------------------------------- */
@@ -2110,6 +2163,9 @@ hidpp10_get_usb_refresh_rate(struct hidpp10_device *dev,
 	if (res)
 		return res;
 
+	if (!refresh.msg.parameters[0])
+		return -EINVAL;
+
 	*rate = 1000/refresh.msg.parameters[0];
 
 	return 0;
@@ -2626,14 +2682,10 @@ hidpp10_get_device_info(struct hidpp10_device *dev)
 {
 	uint32_t feature_mask, notifications;
 	uint8_t reflect;
-	int i;
 	uint16_t xres, yres;
 	uint16_t refresh_rate;
 	enum hidpp10_led_status led[6];
-	int8_t current_profile;
-	struct hidpp10_profile profiles[HIDPP10_NUM_PROFILES];
-	int count;
-	struct hidpp10_directory directory[16];
+	uint8_t current_profile;
 
 	hidpp10_get_individual_features(dev, &feature_mask);
 	hidpp10_get_hidpp_notifications(dev, &notifications);
@@ -2645,18 +2697,14 @@ hidpp10_get_device_info(struct hidpp10_device *dev)
 	hidpp10_get_optical_sensor_settings(dev, &reflect);
 
 	hidpp10_get_current_profile(dev, &current_profile);
-
-	count = hidpp10_get_profile_directory(dev, directory,
-					      ARRAY_LENGTH(directory));
-	for (i = 0; i < count && i < HIDPP10_NUM_PROFILES; i++)
-		hidpp10_get_profile(dev, i, &profiles[i]);
-
 	return 0;
 }
 
 struct hidpp10_device*
-hidpp10_device_new(const struct hidpp_device *base, int idx,
-		   enum hidpp10_profile_type type)
+hidpp10_device_new(const struct hidpp_device *base,
+		   int idx,
+		   enum hidpp10_profile_type type,
+		   unsigned int profile_count)
 {
 	struct hidpp10_device *dev;
 
@@ -2665,6 +2713,8 @@ hidpp10_device_new(const struct hidpp_device *base, int idx,
 	dev->index = idx;
 	dev->base = *base;
 	dev->profile_type = type;
+	dev->profile_count = profile_count;
+	dev->profiles = zalloc(dev->profile_count * sizeof(struct hidpp10_profile));
 
 	if (hidpp10_get_device_info(dev) != 0) {
 		hidpp10_device_destroy(dev);
@@ -2674,27 +2724,35 @@ hidpp10_device_new(const struct hidpp_device *base, int idx,
 	return dev;
 }
 
+int
+hidpp10_device_read_profiles(struct hidpp10_device *dev)
+{
+	unsigned int i;
+
+	hidpp10_read_profile_directory(dev);
+
+	for (i = 0; i < dev->profile_count && i < HIDPP10_NUM_PROFILES; i++)
+		hidpp10_read_profile(dev, i);
+
+	return 0;
+}
+
 void
 hidpp10_device_destroy(struct hidpp10_device *dev)
 {
 	union hidpp10_macro_data **macro;
 	unsigned i;
 
-	if (dev->dpi_table)
-		free(dev->dpi_table);
-	if (dev->profile_directory)
-		free(dev->profile_directory);
-	if (dev->profiles) {
-		for (i = 0; i < dev->profile_count; i++) {
-			ARRAY_FOR_EACH(dev->profiles[i].macros, macro) {
-				if (*macro) {
-					free(*macro);
-					*macro = NULL;
-				}
+	free(dev->dpi_table);
+	for (i = 0; i < dev->profile_count; i++) {
+		ARRAY_FOR_EACH(dev->profiles[i].macros, macro) {
+			if (*macro) {
+				free(*macro);
+				*macro = NULL;
 			}
 		}
-
-		free(dev->profiles);
 	}
+
+	free(dev->profiles);
 	free(dev);
 }
