@@ -31,6 +31,7 @@
 #include <stdint.h>
 
 #include "hidpp-generic.h"
+#include "libratbag-util.h"
 
 struct _hidpp20_message {
 	uint8_t report_id;
@@ -99,6 +100,18 @@ int hidpp20_root_get_protocol_version(struct hidpp20_device *dev,
 
 #define HIDPP_PAGE_DEVICE_INFO				0x0003
 
+/* -------------------------------------------------------------------------- */
+/* 0x0005: Device Name                                                        */
+/* -------------------------------------------------------------------------- */
+
+#define HIDPP_PAGE_DEVICE_NAME				0x0005
+
+/* -------------------------------------------------------------------------- */
+/* 0x0020: Reset                                                              */
+/* -------------------------------------------------------------------------- */
+
+#define HIDPP_PAGE_RESET					0x0020
+
 /* -------------------------------------------------------------------------- */
 /* 0x1000: Battery level status                                               */
 /* -------------------------------------------------------------------------- */
@@ -125,6 +138,130 @@ int hidpp20_batterylevel_get_battery_level(struct hidpp20_device *device,
 					   uint16_t *level,
 					   uint16_t *next_level);
 
+/* -------------------------------------------------------------------------- */
+/* 0x1300: LED software control                                               */
+/* -------------------------------------------------------------------------- */
+
+#define HIDPP_PAGE_LED_SW_CONTROL              0x1300
+
+enum hidpp20_led_sw_ctrl_led_type {
+	HIDPP20_LED_TYPE_BATTERY = 0x01,
+	HIDPP20_LED_TYPE_DPI,
+	HIDPP20_LED_TYPE_PROFILE,
+	HIDPP20_LED_TYPE_LOGO,
+	HIDPP20_LED_TYPE_COSMETIC
+};
+
+enum hidpp20_led_sw_ctrl_led_mode {
+	HIDPP20_LED_MODE_OFF = 0x01,
+	HIDPP20_LED_MODE_ON = 0x02,
+	HIDPP20_LED_MODE_BLINK = 0x04,
+	HIDPP20_LED_MODE_TRAVEL = 0x08,
+	HIDPP20_LED_MODE_RAMP_UP = 0x10,
+	HIDPP20_LED_MODE_RAMP_DOWN = 0x20,
+	HIDPP20_LED_MODE_HEARTBEAT = 0x40,
+	HIDPP20_LED_MODE_BREATHING = 0x80
+};
+
+#define HIDPP20_LED_SW_CONTROL_LED_INDEX_ALL 0xff
+
+struct hidpp20_led_sw_ctrl_led_info {
+	uint8_t index;
+	uint8_t type;
+	uint8_t physical_count;
+	uint16_t caps;
+	uint8_t nvconfig_caps;
+} __attribute__((packed));
+_Static_assert(sizeof(struct hidpp20_led_sw_ctrl_led_info) == 6, "Invalid size");
+
+struct hidpp20_led_sw_ctrl_led_state {
+	uint8_t index;
+	uint16_t mode;
+	union {
+		struct {
+			uint16_t brightness;
+			uint16_t period;
+			uint16_t timeout;
+		} breathing;
+		struct {
+			uint16_t unused;
+			uint16_t delay;
+		} traveling;
+		struct {
+			uint16_t index;
+			uint16_t on_time;
+			uint16_t off_time;
+		} blink;
+		struct {
+			/* Logical information to display on the LED
+			 * Meaning and value range depend on the LED
+			 */
+			uint16_t index;
+		} on;
+	};
+} __attribute__((packed));
+_Static_assert(sizeof(struct hidpp20_led_sw_ctrl_led_state) == 9, "Invalid size");
+
+int hidpp20_led_sw_control_read_leds(struct hidpp20_device* device,
+		struct hidpp20_led_sw_ctrl_led_info** info_list);
+
+const char*
+hidpp20_sw_led_control_get_mode_string(const enum hidpp20_led_sw_ctrl_led_mode mode);
+
+int hidpp20_led_sw_control_read_leds(struct hidpp20_device* device,
+		struct hidpp20_led_sw_ctrl_led_info** info_list);
+
+/**
+ * Retrieves the number of non-RGB logical leds from the mouse
+ * @return the led count or a negative errno
+ */
+int hidpp20_led_sw_control_get_led_count(struct hidpp20_device* device);
+
+/**
+ * Retrieves the info for a given logical led
+ * @return 0 or a negative errno in case of error
+ */
+int hidpp20_led_sw_control_get_led_info(struct hidpp20_device* device,
+		uint8_t led_idx,
+		struct hidpp20_led_sw_ctrl_led_info* info);
+
+/**
+ * Reads all LEDs on the device for their current state
+ * @return 0 on success or a negative errno
+ */
+int hidpp20_led_sw_control_read_leds(struct hidpp20_device* device,
+	 struct hidpp20_led_sw_ctrl_led_info** info_list);
+
+/**
+ * Get the current sw control value
+ * @return 0 if the leds are sofware-controlled 1 if they are firmware controlled
+ */
+bool hidpp20_led_sw_control_get_sw_ctrl(struct hidpp20_device* device);
+
+/**
+ * Sets who controls the LEDs
+ * ctrl = 0, the software controls the LEDs
+ * ctrl = 1, the firmware controls the LEDs
+ * @return 0 on success or a negative errno
+ */
+int hidpp20_led_sw_control_set_sw_ctrl(struct hidpp20_device* device, bool ctrl);
+
+/**
+ * Gets the state of a LED
+ * @return 0 on success or a negative errno
+ */
+int hidpp20_led_sw_control_get_led_state(struct hidpp20_device* device,
+	 uint8_t led_idx,
+	 struct hidpp20_led_sw_ctrl_led_state* state);
+
+/**
+ * Sets the state of a LED
+ * The LED index should be passed in the state structure
+ * @return 0 on success or a negative errno
+ */
+int hidpp20_led_sw_control_set_led_state(struct hidpp20_device* device,
+	 const struct hidpp20_led_sw_ctrl_led_state* state);
+
 /* -------------------------------------------------------------------------- */
 /* 0x1b00: KBD reprogrammable keys and mouse buttons                          */
 /* -------------------------------------------------------------------------- */
@@ -191,6 +328,12 @@ const char *hidpp20_1b04_get_logical_mapping_name(uint16_t value);
 enum ratbag_button_type hidpp20_1b04_get_physical_mapping(uint16_t value);
 const char *hidpp20_1b04_get_physical_mapping_name(uint16_t value);
 
+/* -------------------------------------------------------------------------- */
+/* 0x1d4b: Wireless Device Status                                             */
+/* -------------------------------------------------------------------------- */
+
+#define HIDPP_PAGE_WIRELESS_DEVICE_STATUS		0x1d4b
+
 /* -------------------------------------------------------------------------- */
 /* 0x2200: Mouse Pointer Basic Optical Sensors                                */
 /* -------------------------------------------------------------------------- */
@@ -251,15 +394,33 @@ int hidpp20_adjustable_dpi_set_sensor_dpi(struct hidpp20_device *device,
 
 #define HIDPP_PAGE_ADJUSTABLE_REPORT_RATE		0x8060
 
+/**
+ * set the bitmap_ms to the supported report rates. Bits enabled reflect a
+ * supported report rate, where bit 0 equals 1ms, bit 1 2ms, bit 2 3ms, etc.
+ */
+int hidpp20_adjustable_report_rate_get_report_rate_list(struct hidpp20_device *device,
+							uint8_t *bitflags_ms);
+
 /* -------------------------------------------------------------------------- */
 /* 0x8070v4 - Color LED effects                                               */
 /* -------------------------------------------------------------------------- */
 
 #define HIDPP_PAGE_COLOR_LED_EFFECTS			0x8070
 
+enum hidpp20_color_led_info_ext_caps {
+	HIDPP20_COLOR_LED_INFO_EXT_CAP_HAS_ZONE_EFFECT = 0x01,
+	HIDPP20_COLOR_LED_INFO_EXT_CAP_NO_GET_EFFECT_SETTINGS = 0x02,
+	HIDPP20_COLOR_LED_INFO_EXT_CAP_HAS_SET_LED_BIN_INFO = 0x04,
+	HIDPP20_COLOR_LED_INFO_EXT_CAP_MONOCHROME_ONLY = 0x08,
+};
 
 struct hidpp20_color_led_info;
 
+int
+hidpp20_color_led_effects_get_info(struct hidpp20_device *device,
+				   struct hidpp20_color_led_info *info);
+
+
 struct hidpp20_color_led_zone_info;
 
 int
@@ -270,6 +431,14 @@ int
 hidpp20_color_led_effects_get_zone_infos(struct hidpp20_device *device,
 					 struct hidpp20_color_led_zone_info **infos_list);
 
+struct hidpp20_color_led_zone_effect_info;
+
+int
+hidpp20_color_led_effect_get_zone_effect_info(struct hidpp20_device *device,
+					      uint8_t zone_index,
+					      uint8_t zone_effect_index,
+					      struct hidpp20_color_led_zone_effect_info *info);
+
 enum hidpp20_color_led_location {
 	HIDPP20_COLOR_LED_LOCATION_UNDEFINED = 0,
 	HIDPP20_COLOR_LED_LOCATION_PRIMARY,
@@ -292,6 +461,19 @@ enum hidpp20_color_led_persistency {
 	HIDPP20_COLOR_LED_PERSISTENCY_ON_OFF,
 };
 
+enum hidpp20_color_led_zone_effect {
+	HIDPP20_COLOR_LED_ZONE_EFFECT_DISABLED = 0,
+	HIDPP20_COLOR_LED_ZONE_EFFECT_FIXED = 1,
+	HIDPP20_COLOR_LED_ZONE_EFFECT_CYCLING = 3,
+	HIDPP20_COLOR_LED_ZONE_EFFECT_WAVE = 4,
+	HIDPP20_COLOR_LED_ZONE_EFFECT_STARLIGHT = 5,
+	HIDPP20_COLOR_LED_ZONE_EFFECT_LIGHT_ON_PRESS = 6,
+	HIDPP20_COLOR_LED_ZONE_EFFECT_AUDIO_VISUALIZER = 7,
+	HIDPP20_COLOR_LED_ZONE_EFFECT_BOOT_UP = 8,
+	HIDPP20_COLOR_LED_ZONE_EFFECT_DEMO_MODE = 8,
+	HIDPP20_COLOR_LED_ZONE_EFFECT_BREATHING = 10,
+};
+
 /* -------------------------------------------------------------------------- */
 /* 0x8100 - Onboard Profiles                                                  */
 /* -------------------------------------------------------------------------- */
@@ -361,6 +543,14 @@ struct hidpp20_color_led_info {
 } __attribute__((packed));
 _Static_assert(sizeof(struct hidpp20_color_led_info) == 5, "Invalid size");
 
+struct hidpp20_non_color_led_zone_info {
+	uint8_t index;
+	uint8_t type;
+	uint16_t capabilities;
+	uint8_t nv_config;
+} __attribute__((packed));
+_Static_assert(sizeof(struct hidpp20_non_color_led_zone_info) == 5, "Invalid size");
+
 struct hidpp20_color_led_zone_info {
 	uint8_t index;
 	uint16_t location;
@@ -369,6 +559,20 @@ struct hidpp20_color_led_zone_info {
 } __attribute__((packed));
 _Static_assert(sizeof(struct hidpp20_color_led_zone_info) == 5, "Invalid size");
 
+union hidpp20_generic_led_zone_info {
+	struct hidpp20_led_sw_ctrl_led_info* leds;
+	struct hidpp20_color_led_zone_info* color_leds;
+};
+
+struct hidpp20_color_led_zone_effect_info {
+       uint8_t zone_index;
+       uint8_t zone_effect_index;
+       uint16_t effect_id;
+       uint16_t effect_caps;
+       uint16_t effect_period;
+} __attribute__((packed));
+_Static_assert(sizeof(struct hidpp20_color_led_zone_effect_info) == 8, "Invalid size");
+
 struct hidpp20_color {
 	uint8_t red;
 	uint8_t green;
@@ -466,18 +670,34 @@ union hidpp20_macro_data {
 _Static_assert(sizeof(union hidpp20_macro_data) == 3, "Invalid size");
 
 struct hidpp20_profile {
-	uint8_t index;
+	uint16_t address;
 	uint8_t enabled;
 	char name[16 * 3];
 	unsigned report_rate;
 	unsigned default_dpi;
 	unsigned switched_dpi;
+	unsigned current_dpi;
 	uint16_t dpi[HIDPP20_DPI_COUNT];
 	union hidpp20_button_binding buttons[32];
 	union hidpp20_macro_data *macros[32];
 	struct hidpp20_led leds[HIDPP20_LED_COUNT];
 };
 
+struct hidpp20_onboard_profiles_info {
+	uint8_t memory_model_id;
+	uint8_t profile_format_id;
+	uint8_t macro_format_id;
+	uint8_t profile_count;
+	uint8_t profile_count_oob;
+	uint8_t button_count;
+	uint8_t sector_count;
+	uint16_t sector_size;
+	uint8_t mechanical_layout;
+	uint8_t various_info;
+	uint8_t reserved[5];
+} __attribute__((packed));
+_Static_assert(sizeof(struct hidpp20_onboard_profiles_info) == 16, "Invalid size");
+
 struct hidpp20_profiles {
 	uint8_t num_profiles;
 	uint8_t num_rom_profiles;
@@ -488,9 +708,20 @@ struct hidpp20_profiles {
 	uint8_t has_dpi_shift;
 	uint8_t corded;
 	uint8_t wireless;
-	struct hidpp20_profile profiles[0];
+	uint8_t sector_count;
+	uint16_t sector_size;
+	struct hidpp20_profile *profiles;
 };
 
+/**
+ * fetches the profiles description as reported by the mouse.
+ *
+ * returns 0 or a negative error.
+ */
+int
+hidpp20_onboard_profiles_get_profiles_desc(struct hidpp20_device *device,
+					   struct hidpp20_onboard_profiles_info *info);
+
 /**
  * allocates a list of profiles that has to be destroyed by the caller.
  * The caller must use hidpp20_onboard_profiles_destroy() to free the memory.
@@ -504,8 +735,15 @@ int hidpp20_onboard_profiles_allocate(struct hidpp20_device *device,
  * free a list of profiles allocated by hidpp20_onboard_profiles_allocate()
  */
 void
-hidpp20_onboard_profiles_destroy(struct hidpp20_device *device,
-				 struct hidpp20_profiles *profiles_list);
+hidpp20_onboard_profiles_destroy(struct hidpp20_profiles *profiles_list);
+
+/**
+ * initialize a struct hidpp20_profiles previous allocated with
+ * hidpp20_onboard_profiles_allocate().
+ */
+int
+hidpp20_onboard_profiles_initialize(struct hidpp20_device *device,
+				    struct hidpp20_profiles *profiles);
 
 /**
  * return the current profile index or a negative error.
@@ -523,18 +761,27 @@ hidpp20_onboard_profiles_set_current_profile(struct hidpp20_device *device,
 					     uint8_t index);
 
 /**
- * parse a given profile from the mouse and fill in the right profile in
- * profiles_list.
+ * return the current dpi index of the current active profile
+ * or a negative error.
+ */
+int hidpp20_onboard_profiles_get_current_dpi_index(struct hidpp20_device *device);
+
+/**
+ * Sets the current dpi index on the current active profile.
+ * Indexes are 0-indexed.
  *
  * return 0 or a negative error.
  */
-int hidpp20_onboard_profiles_read(struct hidpp20_device *device,
-				  unsigned int index,
-				  struct hidpp20_profiles *profiles_list);
+int
+hidpp20_onboard_profiles_set_current_dpi_index(struct hidpp20_device *device,
+					       uint8_t index);
 
-int hidpp20_onboard_profiles_write(struct hidpp20_device *device,
-				   unsigned int index,
-				   struct hidpp20_profiles *profiles_list);
+/**
+ * Write the internal state of the device onto the FLASH.
+ */
+int
+hidpp20_onboard_profiles_commit(struct hidpp20_device *device,
+				struct hidpp20_profiles *profiles_list);
 
 enum ratbag_button_action_special
 hidpp20_onboard_profiles_get_special(uint8_t code);
@@ -543,16 +790,26 @@ uint8_t
 hidpp20_onboard_profiles_get_code_from_special(enum ratbag_button_action_special special);
 
 int
-hidpp20_onboard_profiles_read_memory(struct hidpp20_device *device,
-				     uint8_t read_rom,
-				     uint8_t page,
-				     uint16_t section,
-				     uint8_t result[16]);
+hidpp20_onboard_profiles_read_sector(struct hidpp20_device *device,
+				     uint16_t sector,
+				     uint16_t sector_size,
+				     uint8_t *data);
+
+int
+hidpp20_onboard_profiles_write_sector(struct hidpp20_device *device,
+				      uint16_t sector,
+				      uint16_t sector_size,
+				      uint8_t *data,
+				      bool write_crc);
+
+static inline uint8_t *
+hidpp20_onboard_profiles_allocate_sector(struct hidpp20_profiles *profiles)
+{
+	return zalloc(profiles->sector_size);
+}
 
 /* -------------------------------------------------------------------------- */
 /* 0x8110 - Mouse Button Spy                                                  */
 /* -------------------------------------------------------------------------- */
 
 #define HIDPP_PAGE_MOUSE_BUTTON_SPY			0x8110
-
-
