@@ -32,6 +32,36 @@
 #include "libratbag-util.h"
 #include "libratbag-hidraw.h"
 
+#define RATBAG_LED_TYPE_UNKNOWN 0
+
+void
+log_msg_va(struct ratbag *ratbag,
+	   enum ratbag_log_priority priority,
+	   const char *format,
+	   va_list args)
+	LIBRATBAG_ATTRIBUTE_PRINTF(3, 0);
+void log_msg(struct ratbag *ratbag,
+	enum ratbag_log_priority priority,
+	const char *format, ...)
+	LIBRATBAG_ATTRIBUTE_PRINTF(3, 4);
+void
+log_buffer(struct ratbag *ratbag,
+	enum ratbag_log_priority priority,
+	const char *header,
+	uint8_t *buf, size_t len);
+
+#define log_raw(li_, ...) log_msg((li_), RATBAG_LOG_PRIORITY_RAW, __VA_ARGS__)
+#define log_debug(li_, ...) log_msg((li_), RATBAG_LOG_PRIORITY_DEBUG, __VA_ARGS__)
+#define log_info(li_, ...) log_msg((li_), RATBAG_LOG_PRIORITY_INFO, __VA_ARGS__)
+#define log_error(li_, ...) log_msg((li_), RATBAG_LOG_PRIORITY_ERROR, __VA_ARGS__)
+#define log_bug_kernel(li_, ...) log_msg((li_), RATBAG_LOG_PRIORITY_ERROR, "kernel bug: " __VA_ARGS__)
+#define log_bug_libratbag(li_, ...) log_msg((li_), RATBAG_LOG_PRIORITY_ERROR, "libratbag bug: " __VA_ARGS__)
+#define log_bug_client(li_, ...) log_msg((li_), RATBAG_LOG_PRIORITY_ERROR, "client bug: " __VA_ARGS__)
+#define log_buf_raw(li_, h_, buf_, len_) log_buffer(li_, RATBAG_LOG_PRIORITY_RAW, h_, buf_, len_)
+#define log_buf_debug(li_, h_, buf_, len_) log_buffer(li_, RATBAG_LOG_PRIORITY_DEBUG, h_, buf_, len_)
+#define log_buf_info(li_, h_, buf_, len_) log_buffer(li_, RATBAG_LOG_PRIORITY_INFO, h_, buf_, len_)
+#define log_buf_error(li_, h_, buf_, len_) log_buffer(li_, RATBAG_LOG_PRIORITY_ERROR, h_, buf_, len_)
+
 static inline void
 cleanup_device(struct ratbag_device **d)
 {
@@ -50,9 +80,23 @@ cleanup_resolution(struct ratbag_resolution **r)
 	ratbag_resolution_unref(*r);
 }
 
+static inline void
+cleanup_button(struct ratbag_button **b)
+{
+	ratbag_button_unref(*b);
+}
+
+static inline void
+cleanup_led(struct ratbag_led **l)
+{
+	ratbag_led_unref(*l);
+}
+
 #define _cleanup_device_ _cleanup_(cleanup_device)
 #define _cleanup_profile_ _cleanup_(cleanup_profile)
 #define _cleanup_resolution_ _cleanup_(cleanup_resolution)
+#define _cleanup_button_ _cleanup_(cleanup_button)
+#define _cleanup_led_ _cleanup_(cleanup_led)
 
 #define BUS_ANY					0xffff
 #define VENDOR_ANY				0xffff
@@ -78,17 +122,20 @@ struct ratbag {
 	enum ratbag_log_priority log_priority;
 };
 
+#define MAX_CAP 1000
+
 struct ratbag_device {
 	char *name;
 	void *userdata;
 
 	struct udev_device *udev_device;
-	struct ratbag_hidraw hidraw;
+	struct ratbag_hidraw hidraw[MAX_HIDRAW];
 	int refcount;
 	struct input_id ids;
 	struct ratbag_driver *driver;
 	struct ratbag *ratbag;
-	unsigned long capabilities;
+	struct ratbag_device_data *data;
+	unsigned long capabilities[NLONGS(MAX_CAP)];
 
 	unsigned num_profiles;
 	struct list profiles;
@@ -118,6 +165,10 @@ struct ratbag_driver {
 	 *
 	 * Return -ENODEV to ignore the device and let other drivers
 	 * probe the device. Any other error code will stop the probing.
+	 *
+	 * If a matching driver is found it should initialize itself
+	 * and synchronize all profiles with the current state on the device
+	 * in this callback.
 	 */
 	int (*probe)(struct ratbag_device *device);
 
@@ -143,18 +194,16 @@ struct ratbag_driver {
 	int (*commit)(struct ratbag_device *device);
 
 	/**
-	 * Callback called when a read profile is requested by the
-	 * caller of the library.
+	 * Callback called when the device doesn't have a .device file with
+	 * an SVG name set.
 	 *
-	 * The driver should probe here the device for the requested
-	 * profile and populate the related information.
-	 * There is no need to populate the various struct ratbag_button
-	 * as they are allocated when the user needs it.
+	 * The returned value must not be free by the caller, the driver
+	 * has to free it during .remove().
 	 */
-	void (*read_profile)(struct ratbag_profile *profile, unsigned int index);
+	const char *(*get_svg_name)(const struct ratbag_device *device);
 
 	/*
-	 * FIXME: This function is deprecated and should not be removed. Once
+	 * FIXME: This function is deprecated and should be removed. Once
 	 * we've updated all the device drivers to stop using it we'll remove
 	 * it. Look at commit() instead.
 	 */
@@ -168,21 +217,8 @@ struct ratbag_driver {
 	 */
 	int (*set_active_profile)(struct ratbag_device *device, unsigned int index);
 
-	/**
-	 * For the given button, fill in the struct ratbag_button
-	 * with the available information.
-	 *
-	 * For devices with profiles, profile will not be NULL. For
-	 * device without, profile will be set to NULL.
-	 *
-	 * For devices with profile, there should not be the need to
-	 * actually read the profile from the device. The caller
-	 * should make sure that the profile is up to date.
-	 */
-	void (*read_button)(struct ratbag_button *button);
-
 	/*
-	 * FIXME: This function is deprecated and should not be removed. Once
+	 * FIXME: This function is deprecated and should be removed. Once
 	 * we've updated all the device drivers to stop using it we'll remove
 	 * it. Look at commit() instead.
 	 */
@@ -190,30 +226,24 @@ struct ratbag_driver {
 			    const struct ratbag_button_action *action);
 
 	/*
-	 * FIXME: This function is deprecated and should not be removed. Once
+	 * FIXME: This function is deprecated and should be removed. Once
 	 * we've updated all the device drivers to stop using it we'll remove
 	 * it. Look at commit() instead.
 	 */
 	int (*write_resolution_dpi)(struct ratbag_resolution *resolution,
 				    int dpi_x, int dpi_y);
 
-	/**
-	 * For the given led, fill in the struct ratbag_led
-	 * with the available information.
-	 */
-	void (*read_led)(struct ratbag_led *led);
-
 	/*
 	 * FIXME: This function is deprecated and should be removed. Once
 	 * we've updated all the device drivers to stop using it we'll remove
 	 * it. Look at commit() instead.
 	 */
 	int (*write_led)(struct ratbag_led *led, enum ratbag_led_mode mode,
-			 struct ratbag_color color, unsigned int hz,
+			 struct ratbag_color color, unsigned int ms,
 			 unsigned int brightness);
 
 	/* private */
-	int (*test_probe)(struct ratbag_device *device, void *data);
+	int (*test_probe)(struct ratbag_device *device, const void *data);
 
 	struct list link;
 };
@@ -222,11 +252,22 @@ struct ratbag_resolution {
 	struct ratbag_profile *profile;
 	int refcount;
 	void *userdata;
+	struct list link;
+	unsigned index;
+
+	unsigned int dpis[300];
+	size_t ndpis;
+
 	unsigned int dpi_x;	/**< x resolution in dpi */
 	unsigned int dpi_y;	/**< y resolution in dpi */
 	unsigned int hz;	/**< report rate in Hz */
+
+	unsigned int rates[8];
+	size_t nrates;
+
 	bool is_active;
 	bool is_default;
+	bool dirty;
 	uint32_t capabilities;
 };
 
@@ -238,8 +279,10 @@ struct ratbag_led {
 	unsigned index;
 	enum ratbag_led_type type;
 	enum ratbag_led_mode mode;
+	uint32_t modes;		      /**< supported modes */
 	struct ratbag_color color;
-	unsigned int hz;              /**< rate of action in hz */
+	enum ratbag_led_colordepth colordepth;
+	unsigned int ms;              /**< duration of action in ms */
 	unsigned int brightness;      /**< brightness of the LED */
 	bool dirty;
 };
@@ -247,6 +290,7 @@ struct ratbag_led {
 struct ratbag_profile {
 	int refcount;
 	void *userdata;
+	char *name;
 
 	struct list link;
 	unsigned index;
@@ -254,17 +298,29 @@ struct ratbag_profile {
 	struct list buttons;
 	void *drv_data;
 	void *user_data;
-	struct {
-		struct ratbag_resolution *modes;
-		unsigned int num_modes;
-	} resolution;
+	struct list resolutions;
 	struct list leds;
 
+	unsigned int num_resolutions;
+
 	bool is_active;		/**< profile is the currently active one */
 	bool is_enabled;
 	bool dirty;       /**< profile changed since last commit */
+	unsigned long capabilities[NLONGS(MAX_CAP)];
 };
 
+#define ratbag_device_for_each_profile(device_, profile_) \
+	list_for_each(profile_, &(device_)->profiles, link)
+
+#define ratbag_profile_for_each_button(profile_, button_) \
+	list_for_each(button_, &(profile_)->buttons, link) \
+
+#define ratbag_profile_for_each_led(profile_, led_) \
+	list_for_each(led_, &(profile_)->leds, link)
+
+#define ratbag_profile_for_each_resolution(profile_, resolution_) \
+	list_for_each(resolution_, &(profile_)->resolutions, link)
+
 #define BUTTON_ACTION_NONE \
  { .type = RATBAG_BUTTON_ACTION_TYPE_NONE }
 #define BUTTON_ACTION_BUTTON(num_) \
@@ -300,6 +356,15 @@ struct ratbag_button_macro {
 	struct ratbag_macro macro;
 };
 
+#define MODIFIER_LEFTCTRL (1 << 0)
+#define MODIFIER_LEFTSHIFT (1 << 1)
+#define MODIFIER_LEFTALT (1 << 2)
+#define MODIFIER_LEFTMETA (1 << 3)
+#define MODIFIER_RIGHTCTRL (1 << 4)
+#define MODIFIER_RIGHTSHIFT (1 << 5)
+#define MODIFIER_RIGHTALT (1 << 6)
+#define MODIFIER_RIGHTMETA (1 << 7)
+
 struct ratbag_button_action {
 	enum ratbag_button_action_type type;
 	union ratbag_btn_action {
@@ -325,6 +390,10 @@ struct ratbag_button {
 	bool dirty; /* changed since last commit to device */
 };
 
+void
+ratbag_button_set_action(struct ratbag_button *button,
+			 const struct ratbag_button_action *action);
+
 static inline void
 ratbag_button_enable_action_type(struct ratbag_button *button,
 				 enum ratbag_button_action_type type)
@@ -355,7 +424,7 @@ ratbag_set_drv_data(struct ratbag_device *device, void *drv_data)
 }
 
 static inline void *
-ratbag_get_drv_data(struct ratbag_device *device)
+ratbag_get_drv_data(const struct ratbag_device *device)
 {
 	return device->drv_data;
 }
@@ -371,6 +440,10 @@ void
 ratbag_device_set_capability(struct ratbag_device *device,
 			     enum ratbag_device_capability cap);
 
+void
+ratbag_device_unset_capability(struct ratbag_device *device,
+			     enum ratbag_device_capability cap);
+
 static inline void
 ratbag_profile_set_drv_data(struct ratbag_profile *profile, void *drv_data)
 {
@@ -383,6 +456,15 @@ ratbag_profile_get_drv_data(struct ratbag_profile *profile)
 	return profile->drv_data;
 }
 
+static inline void
+ratbag_profile_set_cap(struct ratbag_profile *profile,
+		       enum ratbag_profile_capability cap)
+{
+	assert(cap <= MAX_CAP);
+
+	long_set_bit(profile->capabilities, cap);
+}
+
 static inline int
 ratbag_button_action_match(const struct ratbag_button_action *action,
 			   const struct ratbag_button_action *match)
@@ -406,21 +488,95 @@ ratbag_button_action_match(const struct ratbag_button_action *action,
 	return 0;
 }
 
-static inline struct ratbag_resolution *
-ratbag_resolution_init(struct ratbag_profile *profile, int index,
-		       int dpi_x, int dpi_y, int hz)
-{
-	struct ratbag_resolution *res = &profile->resolution.modes[index];
+int
+ratbag_action_macro_num_keys(struct ratbag_button_action *action);
 
-	res->profile = profile;
+int
+ratbag_button_macro_new_from_keycode(struct ratbag_button *button,
+				     unsigned int key,
+				     unsigned int modifiers);
+
+int
+ratbag_action_keycode_from_macro(struct ratbag_button_action *action,
+				     unsigned int *key_out,
+				     unsigned int *modifiers_out);
+
+static inline void
+ratbag_resolution_set_resolution(struct ratbag_resolution *res,
+				 int dpi_x, int dpi_y, int hz)
+{
 	res->dpi_x = dpi_x;
 	res->dpi_y = dpi_y;
 	res->hz = hz;
-	res->is_active = false;
-	res->is_default = false;
-	res->capabilities = 0;
+}
 
-	return res;
+static inline void
+ratbag_resolution_set_dpi_list_from_range(struct ratbag_resolution *res,
+					  unsigned int min, unsigned int max)
+{
+	unsigned int stepsize = 50;
+	unsigned int dpi = min;
+	bool maxed_out = false;
+
+	res->ndpis = 0;
+
+	while (res->ndpis < ARRAY_LENGTH(res->dpis)) {
+		if (dpi > (unsigned)max) {
+			maxed_out = true;
+			break;
+		}
+
+		res->dpis[res->ndpis] = dpi;
+		res->ndpis++;
+
+		if (dpi < 1000)
+			stepsize = 50;
+		else if (dpi < 2600)
+			stepsize = 100;
+		else if (dpi < 5000)
+			stepsize = 200;
+		else
+			stepsize = 500;
+
+		dpi += stepsize;
+	}
+
+	if (!maxed_out)
+		log_bug_libratbag(res->profile->device->ratbag,
+				  "%s: resolution range exceeds available space.\n",
+				  res->profile->device->name);
+}
+
+static inline void
+ratbag_resolution_set_dpi_list(struct ratbag_resolution *res,
+			       unsigned int *dpis,
+			       size_t ndpis)
+{
+	assert(ndpis <= ARRAY_LENGTH(res->dpis));
+	_Static_assert(sizeof(*dpis) == sizeof(*res->dpis), "Mismatching size");
+
+	for (size_t i = 0; i < ndpis; i++) {
+		res->dpis[i] = dpis[i];
+		if (i > 0)
+			assert(dpis[i] > dpis[i - 1]);
+	}
+	res->ndpis = ndpis;
+}
+
+static inline void
+ratbag_resolution_set_report_rate_list(struct ratbag_resolution *res,
+				       unsigned int *rates,
+				       size_t nrates)
+{
+	assert(nrates <= ARRAY_LENGTH(res->rates));
+	_Static_assert(sizeof(*rates) == sizeof(*res->rates), "Mismatching size");
+
+	for (size_t i = 0; i < nrates; i++) {
+		res->rates[i] = rates[i];
+		if (i > 0)
+			assert(rates[i] > rates[i - 1]);
+	}
+	res->nrates = nrates;
 }
 
 static inline void
@@ -432,41 +588,25 @@ ratbag_resolution_set_cap(struct ratbag_resolution *res,
 	res->capabilities = (1 << cap);
 }
 
-void
-log_msg_va(struct ratbag *ratbag,
-	   enum ratbag_log_priority priority,
-	   const char *format,
-	   va_list args)
-	LIBRATBAG_ATTRIBUTE_PRINTF(3, 0);
-void log_msg(struct ratbag *ratbag,
-	enum ratbag_log_priority priority,
-	const char *format, ...)
-	LIBRATBAG_ATTRIBUTE_PRINTF(3, 4);
-void
-log_buffer(struct ratbag *ratbag,
-	enum ratbag_log_priority priority,
-	const char *header,
-	uint8_t *buf, size_t len);
+static inline void
+ratbag_led_set_mode_capability(struct ratbag_led *led,
+			       enum ratbag_led_mode mode)
+{
+	assert(mode <= RATBAG_LED_BREATHING);
+	assert(mode < sizeof(led->modes) * 8);
 
-#define log_raw(li_, ...) log_msg((li_), RATBAG_LOG_PRIORITY_RAW, __VA_ARGS__)
-#define log_debug(li_, ...) log_msg((li_), RATBAG_LOG_PRIORITY_DEBUG, __VA_ARGS__)
-#define log_info(li_, ...) log_msg((li_), RATBAG_LOG_PRIORITY_INFO, __VA_ARGS__)
-#define log_error(li_, ...) log_msg((li_), RATBAG_LOG_PRIORITY_ERROR, __VA_ARGS__)
-#define log_bug_kernel(li_, ...) log_msg((li_), RATBAG_LOG_PRIORITY_ERROR, "kernel bug: " __VA_ARGS__)
-#define log_bug_libratbag(li_, ...) log_msg((li_), RATBAG_LOG_PRIORITY_ERROR, "libratbag bug: " __VA_ARGS__)
-#define log_bug_client(li_, ...) log_msg((li_), RATBAG_LOG_PRIORITY_ERROR, "client bug: " __VA_ARGS__)
-#define log_buf_raw(li_, h_, buf_, len_) log_buffer(li_, RATBAG_LOG_PRIORITY_RAW, h_, buf_, len_)
-#define log_buf_debug(li_, h_, buf_, len_) log_buffer(li_, RATBAG_LOG_PRIORITY_DEBUG, h_, buf_, len_)
-#define log_buf_info(li_, h_, buf_, len_) log_buffer(li_, RATBAG_LOG_PRIORITY_INFO, h_, buf_, len_)
-#define log_buf_error(li_, h_, buf_, len_) log_buffer(li_, RATBAG_LOG_PRIORITY_ERROR, h_, buf_, len_)
+	led->modes |= (1 << mode);
+}
 
 /* list of all supported drivers */
 struct ratbag_driver etekcity_driver;
 struct ratbag_driver hidpp20_driver;
 struct ratbag_driver hidpp10_driver;
 struct ratbag_driver logitech_g300_driver;
+struct ratbag_driver logitech_g600_driver;
 struct ratbag_driver roccat_driver;
 struct ratbag_driver gskill_driver;
+struct ratbag_driver steelseries_driver;
 
 struct ratbag_device*
 ratbag_device_new(struct ratbag *ratbag, struct udev_device *udev_device,
@@ -481,7 +621,7 @@ ratbag_device_get_udev_property(const struct ratbag_device* device,
 bool
 ratbag_assign_driver(struct ratbag_device *device,
 		     const struct input_id *dev_id,
-		     struct ratbag_test_device *test_device);
+		     const struct ratbag_test_device *test_device);
 
 void
 ratbag_register_driver(struct ratbag *ratbag, struct ratbag_driver *driver);
