@@ -24,6 +24,8 @@
 
 #pragma once
 
+#include <stdint.h>
+
 #include "libratbag.h"
 
 #define RATBAG_TEST_MAX_PROFILES 12
@@ -31,12 +33,19 @@
 #define RATBAG_TEST_MAX_RESOLUTIONS 8
 #define RATBAG_TEST_MAX_LEDS 8
 
+struct ratbag_test_macro_event {
+	enum ratbag_macro_event_type type;
+	unsigned int value;
+};
+
 struct ratbag_test_button {
-	enum ratbag_button_action_type type;
+	enum ratbag_button_type button_type;
+	enum ratbag_button_action_type action_type;
 	union {
 		int button;
 		int key;
 		enum ratbag_button_action_special special;
+		struct ratbag_test_macro_event macro[24];
 	};
 };
 
@@ -46,6 +55,9 @@ struct ratbag_test_resolution {
 	bool active;
 	bool dflt;
 	uint32_t caps;
+
+	int dpi_min, dpi_max;
+	unsigned int report_rates[5];
 };
 
 struct ratbag_test_color {
@@ -58,7 +70,7 @@ struct ratbag_test_led {
 	enum ratbag_led_type type;
 	enum ratbag_led_mode mode;
 	struct ratbag_test_color color;
-	unsigned int hz;
+	unsigned int ms;
 	unsigned int brightness;
 };
 
@@ -68,6 +80,8 @@ struct ratbag_test_profile {
 	struct ratbag_test_led leds[RATBAG_TEST_MAX_LEDS];
 	bool active;
 	bool dflt;
+	bool disabled;
+	uint32_t caps[10];
 };
 
 struct ratbag_test_device {
@@ -75,11 +89,12 @@ struct ratbag_test_device {
 	unsigned int num_resolutions;
 	unsigned int num_buttons;
 	unsigned int num_leds;
+	const char *svg;
 	struct ratbag_test_profile profiles[RATBAG_TEST_MAX_PROFILES];
 	void (*destroyed)(struct ratbag_device *device, void *data);
 	void *destroyed_data;
 };
 
 struct ratbag_device* ratbag_device_new_test_device(struct ratbag *ratbag,
-						    struct ratbag_test_device *test_device);
+						    const struct ratbag_test_device *test_device);
 
