@@ -23,12 +23,16 @@
 
 #pragma once
 
+#include "config.h"
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
 #include <unistd.h>
 #include <libudev.h>
 
+#include "shared-macro.h"
+
 /*
  * This list data structure is a verbatim copy from wayland-util.h from the
  * Wayland project; except that wl_ prefix has been removed.
@@ -41,19 +45,10 @@ struct list {
 
 void list_init(struct list *list);
 void list_insert(struct list *list, struct list *elm);
+void list_append(struct list *list, struct list *elm);
 void list_remove(struct list *elm);
 int list_empty(const struct list *list);
 
-#ifdef __GNUC__
-#define container_of(ptr, sample, member)				\
-	(__typeof__(sample))((char *)(ptr)	-			\
-		 ((char *)&(sample)->member - (char *)(sample)))
-#else
-#define container_of(ptr, sample, member)				\
-	(void *)((char *)(ptr)	-				        \
-		 ((char *)&(sample)->member - (char *)(sample)))
-#endif
-
 #define list_for_each(pos, head, member)				\
 	for (pos = 0, pos = container_of((head)->next, pos, member);	\
 	     &pos->member != (head);					\
@@ -67,45 +62,6 @@ int list_empty(const struct list *list);
 	     pos = tmp,							\
 	     tmp = container_of(pos->member.next, tmp, member))
 
-#define LONG_BITS (sizeof(long) * 8)
-#define NLONGS(x) (((x) + LONG_BITS - 1) / LONG_BITS)
-#define ARRAY_LENGTH(a) (sizeof (a) / sizeof (a)[0])
-#define ARRAY_FOR_EACH(_arr, _elem) \
-	for (size_t _i = 0; _i < ARRAY_LENGTH(_arr) && (_elem = &_arr[_i]); _i++)
-#define AS_MASK(v) (1 << (v))
-
-#define min(a, b) (((a) < (b)) ? (a) : (b))
-#define max(a, b) (((a) > (b)) ? (a) : (b))
-#define streq(s1, s2) (strcmp((s1), (s2)) == 0)
-#define strneq(s1, s2, n) (strncmp((s1), (s2), (n)) == 0)
-
-static inline void
-cleanup_free(void *p) {
-	free(*(void**)p);
-}
-
-static inline void
-cleanup_close(int *p) {
-	close(*p);
-}
-static inline void
-cleanup_udev_unref(struct udev **udev) {
-	if (*udev)
-		udev_unref(*udev);
-}
-
-static inline void
-cleanup_udev_device_unref(struct udev_device **udev_device) {
-	if (*udev_device)
-		udev_device_unref(*udev_device);
-}
-
-#define _cleanup_(x) __attribute__((cleanup(x)))
-#define _cleanup_free_ _cleanup_(cleanup_free)
-#define _cleanup_close_ _cleanup_(cleanup_close)
-#define _cleanup_udev_unref_ _cleanup_(cleanup_udev_unref)
-#define _cleanup_udev_device_unref_ _cleanup_(cleanup_udev_device_unref)
-
 static inline char*
 strncpy_safe(char *dest, const char *src, size_t n)
 {
@@ -163,6 +119,24 @@ snprintf_safe(char *buf, size_t n, const char *fmt, ...)
 #define sprintf_safe(buf, fmt, ...) \
 	snprintf_safe(buf, ARRAY_LENGTH(buf), fmt, __VA_ARGS__)
 
+__attribute__((format(printf, 1, 2)))
+static inline void *
+asprintf_safe(const char *fmt, ...)
+{
+	va_list args;
+	int rc;
+	char *result;
+
+	va_start(args, fmt);
+	rc = vasprintf(&result, fmt, args);
+	va_end(args);
+
+	if (rc < 0)
+		abort();
+
+	return result;
+}
+
 static inline void
 msleep(unsigned int ms)
 {
@@ -216,3 +190,120 @@ ratbag_utf8_to_enc(char *buf, size_t buf_len, const char *to_enc,
 ssize_t
 ratbag_utf8_from_enc(char *in_buf, size_t in_len, const char *from_enc,
 		     char **out);
+
+__attribute__((format(printf, 2, 3)))
+static inline int
+xasprintf(char **strp, const char *fmt, ...)
+{
+	int rc = 0;
+	va_list args;
+
+	va_start(args, fmt);
+	rc = vasprintf(strp, fmt, args);
+	va_end(args);
+	if ((rc == -1) && strp)
+		*strp = NULL;
+
+	return rc;
+}
+
+struct dpi_range {
+	unsigned int min;
+	unsigned int max;
+	float step;
+};
+
+/* Parse a string in the form min:max@step to a dpi_range */
+static inline struct dpi_range *
+dpi_range_from_string(const char *str)
+{
+	float min, max, step;
+	int rc;
+	struct dpi_range *range;
+
+	rc = sscanf(str, "%f:%f@%f", &min, &max, &step);
+	if (rc != 3 || min < 0 || max <= min || step < 20)
+		return NULL;
+
+	range = zalloc(sizeof(*range));
+	range->min = min;
+	range->max = max;
+	range->step = step;
+
+	return range;
+}
+
+struct dpi_list {
+	int *entries;
+	size_t nentries;
+};
+
+static inline void
+dpi_list_free(struct dpi_list *list)
+{
+	free(list->entries);
+	free(list);
+}
+
+DEFINE_TRIVIAL_CLEANUP_FUNC(struct dpi_list *, dpi_list_free);
+
+/* Parse a string in the form "100;200;400"to a dpi_list */
+static inline struct dpi_list *
+dpi_list_from_string(const char *str)
+{
+	struct dpi_list *list;
+	unsigned int i, count, index;
+	int nread, dpi = 0;
+	char c;
+	void *tmp;
+
+	if (!str || str[0] == '\0')
+		return NULL;
+	printf("%s: ", str);
+
+	/* first, count how many elements do we have in the table */
+	count = 1;
+	i = 0;
+	while (str[i] != 0) {
+		c = str[i++];
+		if (c == ';')
+			count++;
+	}
+
+	list = zalloc(sizeof *list);
+	list->nentries = count;
+	list->entries = zalloc(count * sizeof(list->entries[0]));
+
+	index = 0;
+
+	while (*str != 0 && index < count) {
+		if (*str == ';') {
+			str++;
+			continue;
+		}
+
+		nread = 0;
+		sscanf(str, "%d%n", &dpi, &nread);
+		if (!nread || dpi < 0)
+			goto err;
+
+		list->entries[index] = dpi;
+		str += nread;
+		index++;
+	}
+
+	if (index == 0)
+		goto err;
+
+	/* Drop empty entries for trailing semicolons */
+	list->nentries = index;
+	tmp = realloc(list->entries, index * sizeof(list->entries[0]));
+	if (tmp)
+		list->entries = tmp;
+
+	return list;
+
+err:
+	dpi_list_free(list);
+	return NULL;
+}
