@@ -37,6 +37,7 @@
 #include "usb-ids.h"
 #include "libratbag-private.h"
 #include "libratbag-util.h"
+#include "libratbag-data.h"
 
 static enum ratbag_error_code
 error_code(enum ratbag_error_code code)
@@ -61,6 +62,8 @@ static void
 ratbag_button_destroy(struct ratbag_button *button);
 static void
 ratbag_led_destroy(struct ratbag_led *led);
+static void
+ratbag_resolution_destroy(struct ratbag_resolution *resolution);
 
 static void
 ratbag_default_log_func(struct ratbag *ratbag,
@@ -180,6 +183,7 @@ ratbag_device_new(struct ratbag *ratbag, struct udev_device *udev_device,
 	device->refcount = 1;
 	device->udev_device = udev_device_ref(udev_device);
 	device->ids = *id;
+	device->data = ratbag_device_data_new_for_id(ratbag, id);
 	list_init(&device->profiles);
 
 	list_insert(&ratbag->devices, &device->link);
@@ -217,6 +221,7 @@ ratbag_device_destroy(struct ratbag_device *device)
 	list_remove(&device->link);
 
 	ratbag_unref(device->ratbag);
+	ratbag_device_data_unref(device->data);
 	free(device->name);
 	free(device);
 }
@@ -225,11 +230,10 @@ static inline bool
 ratbag_sanity_check_device(struct ratbag_device *device)
 {
 	struct ratbag *ratbag = device->ratbag;
-	_cleanup_profile_ struct ratbag_profile *profile = NULL;
+	struct ratbag_profile *profile = NULL;
 	bool has_active = false;
-	unsigned int nres, nprofiles;
+	unsigned int nres;
 	bool rc = false;
-	unsigned int i;
 
 	/* arbitrary number: max 16 profiles, does any mouse have more? but
 	 * since we have num_profiles unsigned, it also checks for
@@ -242,11 +246,8 @@ ratbag_sanity_check_device(struct ratbag_device *device)
 		goto out;
 	}
 
-	nprofiles = ratbag_device_get_num_profiles(device);
-	for (i = 0; i < nprofiles; i++) {
-		profile = ratbag_device_get_profile(device, i);
-		if (!profile)
-			goto out;
+	ratbag_device_for_each_profile(device, profile) {
+		struct ratbag_resolution *resolution;
 
 		/* Allow max 1 active profile */
 		if (profile->is_active) {
@@ -268,8 +269,29 @@ ratbag_sanity_check_device(struct ratbag_device *device)
 				goto out;
 		}
 
-		ratbag_profile_unref(profile);
-		profile = NULL;
+		ratbag_profile_for_each_resolution(profile, resolution) {
+			unsigned int vals[300];
+			unsigned int nvals = ARRAY_LENGTH(vals);
+
+			if (!ratbag_device_has_capability(device, RATBAG_DEVICE_CAP_RESOLUTION))
+				break;
+
+			nvals = ratbag_resolution_get_dpi_list(resolution, vals, nvals);
+			if (nvals == 0) {
+				log_bug_libratbag(ratbag,
+						  "%s: invalid dpi list\n",
+						  device->name);
+				goto out;
+			}
+
+			nvals = ratbag_resolution_get_report_rate_list(resolution, vals, nvals);
+			if (nvals == 0) {
+				log_bug_libratbag(ratbag,
+						  "%s: invalid report rate list\n",
+						  device->name);
+				goto out;
+			}
+		}
 	}
 
 	/* Require 1 active profile */
@@ -280,6 +302,29 @@ ratbag_sanity_check_device(struct ratbag_device *device)
 		goto out;
 	}
 
+	/* Require LED to be the same type in each profile */
+	ratbag_device_for_each_profile(device, profile) {
+		_cleanup_profile_ struct ratbag_profile *p0 = ratbag_device_get_profile(device, 0);
+		struct ratbag_led *led;
+
+
+		ratbag_profile_for_each_led(profile, led) {
+			_cleanup_led_ struct ratbag_led *p0_led = ratbag_profile_get_led(p0, led->index);
+
+			assert(led->type != RATBAG_LED_TYPE_UNKNOWN);
+
+			if (p0_led->type != led->type) {
+				log_bug_libratbag(ratbag,
+						  "%s: mismatching LED types (profile %d led %d)\n",
+						  device->name,
+						  profile->index,
+						  led->type);
+				goto out;
+			}
+		}
+
+	}
+
 	rc = true;
 
 out:
@@ -290,7 +335,7 @@ static inline bool
 ratbag_test_driver(struct ratbag_device *device,
 		   const struct input_id *dev_id,
 		   const char *driver_name,
-		   struct ratbag_test_device *test_device)
+		   const struct ratbag_test_device *test_device)
 {
 	struct ratbag *ratbag = device->ratbag;
 	struct ratbag_driver *driver;
@@ -328,9 +373,8 @@ ratbag_test_driver(struct ratbag_device *device,
 		log_error(ratbag, "%s: error opening hidraw node (%s)\n",
 			  device->name, strerror(-rc));
 
-error:
 	device->driver = NULL;
-
+error:
 	return false;
 }
 
@@ -353,13 +397,13 @@ ratbag_driver_fallback_logitech(struct ratbag_device *device,
 bool
 ratbag_assign_driver(struct ratbag_device *device,
 		     const struct input_id *dev_id,
-		     struct ratbag_test_device *test_device)
+		     const struct ratbag_test_device *test_device)
 {
 	bool rc;
 	const char *driver_name;
 
 	if (!test_device) {
-		driver_name = udev_prop_value(device->udev_device, "RATBAG_DRIVER");
+		driver_name = ratbag_device_data_get_driver(device->data);
 	} else {
 		driver_name = "test_driver";
 	}
@@ -373,17 +417,21 @@ ratbag_assign_driver(struct ratbag_device *device,
 	return rc;
 }
 
-static inline char*
-get_device_name(struct udev_device *device)
+static int
+get_device_name(struct udev_device *device,
+		char *name, size_t sz)
 {
 	const char *prop;
 
 	prop = udev_prop_value(device, "NAME");
 	if (!prop)
-		return NULL;
+		return -EINVAL;
 
 	/* udev name is enclosed by " */
-	return strndup(&prop[1], strlen(prop) - 2);
+	sz = min(sz, strlen(prop) - 1);
+	snprintf(name, sz, "%s", &prop[1]);
+
+	return 0;
 }
 
 static inline int
@@ -413,7 +461,7 @@ ratbag_device_new_from_udev_device(struct ratbag *ratbag,
 {
 	struct ratbag_device *device = NULL;
 	enum ratbag_error_code error = RATBAG_ERROR_DEVICE;
-	_cleanup_free_ char *name = NULL;
+	char name[128];
 	struct input_id id;
 
 	assert(ratbag != NULL);
@@ -426,12 +474,11 @@ ratbag_device_new_from_udev_device(struct ratbag *ratbag,
 	if (get_product_id(udev_device, &id) != 0)
 		goto out_err;
 
-	name = get_device_name(udev_device);
-	if (!name)
+	if (get_device_name(udev_device, name, sizeof(name)) != 0)
 		goto out_err;
 
 	device = ratbag_device_new(ratbag, udev_device, name, &id);
-	if (!device)
+	if (!device || !device->data)
 		goto out_err;
 
 	if (!ratbag_assign_driver(device, &device->ids, NULL))
@@ -481,14 +528,14 @@ ratbag_device_get_name(const struct ratbag_device* device)
 LIBRATBAG_EXPORT const char *
 ratbag_device_get_svg_name(const struct ratbag_device* device)
 {
-	return udev_prop_value(device->udev_device, "RATBAG_SVG");
-}
+	const char *name = NULL;
 
-const char *
-ratbag_device_get_udev_property(const struct ratbag_device* device,
-				const char *name)
-{
-	return udev_prop_value(device->udev_device, name);
+	if (device->data)
+		name = ratbag_device_data_get_svg(device->data);
+	if (!name && device->driver->get_svg_name)
+		name = device->driver->get_svg_name(device);
+
+	return name;
 }
 
 void
@@ -536,8 +583,10 @@ ratbag_create_context(const struct ratbag_interface *interface,
 	ratbag_register_driver(ratbag, &hidpp20_driver);
 	ratbag_register_driver(ratbag, &hidpp10_driver);
 	ratbag_register_driver(ratbag, &logitech_g300_driver);
+	ratbag_register_driver(ratbag, &logitech_g600_driver);
 	ratbag_register_driver(ratbag, &roccat_driver);
 	ratbag_register_driver(ratbag, &gskill_driver);
+	ratbag_register_driver(ratbag, &steelseries_driver);
 
 	return ratbag;
 }
@@ -568,19 +617,16 @@ ratbag_unref(struct ratbag *ratbag)
 static struct ratbag_button *
 ratbag_create_button(struct ratbag_profile *profile, unsigned int index)
 {
-	struct ratbag_device *device = profile->device;
 	struct ratbag_button *button;
 
 	button = zalloc(sizeof(*button));
 	button->refcount = 0;
 	button->profile = profile;
 	button->index = index;
+	button->type = RATBAG_BUTTON_TYPE_UNKNOWN;
 
 	list_insert(&profile->buttons, &button->link);
 
-	if (device->driver->read_button)
-		device->driver->read_button(button);
-
 	return button;
 }
 
@@ -594,41 +640,42 @@ ratbag_create_led(struct ratbag_profile *profile, unsigned int index)
 	led->refcount = 0;
 	led->profile = profile;
 	led->index = index;
+	led->colordepth = RATBAG_LED_COLORDEPTH_RGB_888;
+	led->type = RATBAG_LED_TYPE_UNKNOWN;
+	if (device->data)
+		led->type = ratbag_device_data_get_led_type(device->data, led->index);
 
 	list_insert(&profile->leds, &led->link);
 
-	if (device->driver->read_led)
-		device->driver->read_led(led);
-
 	return led;
 }
 
-static int
-ratbag_profile_init_buttons(struct ratbag_profile *profile, unsigned int count)
+LIBRATBAG_EXPORT bool
+ratbag_profile_has_capability(const struct ratbag_profile *profile,
+			      enum ratbag_profile_capability cap)
 {
-	unsigned int i;
+	if (cap == RATBAG_PROFILE_CAP_NONE || cap >= MAX_CAP)
+		abort();
 
-	for (i = 0; i < count; i++)
-		ratbag_create_button(profile, i);
-
-	profile->device->num_buttons = count;
-
-	return 0;
+	return long_bit_is_set(profile->capabilities, cap);
 }
 
-static int
-ratbag_profile_init_leds(struct ratbag_profile *profile, unsigned int count)
+static inline void
+ratbag_create_resolution(struct ratbag_profile *profile, int index)
 {
-	unsigned int i;
+	struct ratbag_resolution *res;
 
-	for (i = 0; i < count; i++)
-		ratbag_create_led(profile, i);
+	res = zalloc(sizeof(*res));
+	res->refcount = 0;
+	res->profile = profile;
+	res->index = index;
 
-	profile->device->num_leds = count;
+	list_insert(&profile->resolutions, &res->link);
 
-	return 0;
+	profile->num_resolutions++;
 }
 
+
 static struct ratbag_profile *
 ratbag_create_profile(struct ratbag_device *device,
 		      unsigned int index,
@@ -643,23 +690,26 @@ ratbag_create_profile(struct ratbag_device *device,
 	profile->refcount = 0;
 	profile->device = device;
 	profile->index = index;
-	profile->resolution.modes = zalloc(num_resolutions *
-					   sizeof(*profile->resolution.modes));
-	profile->resolution.num_modes = num_resolutions;
+	profile->num_resolutions = 0;
 	profile->is_enabled = true;
+	profile->name = asprintf_safe("Profile %d", index + 1);
 
 	list_insert(&device->profiles, &profile->link);
 	list_init(&profile->buttons);
 	list_init(&profile->leds);
+	list_init(&profile->resolutions);
+
+	profile->device->num_buttons = num_buttons;
+	profile->device->num_leds = num_leds;
 
 	for (i = 0; i < num_resolutions; i++)
-		ratbag_resolution_init(profile, i, 0, 0, 0);
+		ratbag_create_resolution(profile, i);
 
-	assert(device->driver->read_profile);
-	device->driver->read_profile(profile, index);
+	for (i = 0; i < num_buttons; i++)
+		ratbag_create_button(profile, i);
 
-	ratbag_profile_init_buttons(profile, num_buttons);
-	ratbag_profile_init_leds(profile, num_leds);
+	for (i = 0; i < num_leds; i++)
+		ratbag_create_led(profile, i);
 
 	return profile;
 }
@@ -680,17 +730,21 @@ ratbag_device_init_profiles(struct ratbag_device *device,
 	device->num_profiles = num_profiles;
 
 	if (num_profiles > 1) {
+		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_PROFILE);
 		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_SWITCHABLE_PROFILE);
 
 		/* having more than one profile means we can remap the buttons
 		 * at least */
+		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_BUTTON);
 		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_BUTTON_KEY);
 	}
 
-	if (num_resolutions > 1)
+	if (num_resolutions > 1) {
+		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_RESOLUTION);
 		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_SWITCHABLE_RESOLUTION);
+	}
 
-	if (num_leds > 1)
+	if (num_leds > 0)
 		ratbag_device_set_capability(device, RATBAG_DEVICE_CAP_LED);
 
 	return 0;
@@ -711,6 +765,7 @@ ratbag_profile_destroy(struct ratbag_profile *profile)
 {
 	struct ratbag_button *button, *b_next;
 	struct ratbag_led *led, *l_next;
+	struct ratbag_resolution *res, *r_next;
 
 	/* if we get to the point where the profile is destroyed, buttons,
 	 * resolutions , etc. are at a refcount of 0, so we can destroy
@@ -721,7 +776,10 @@ ratbag_profile_destroy(struct ratbag_profile *profile)
 	list_for_each_safe(led, l_next, &profile->leds, link)
 		ratbag_led_destroy(led);
 
-	free(profile->resolution.modes);
+	list_for_each_safe(res, r_next, &profile->resolutions, link)
+		ratbag_resolution_destroy(res);
+
+	free(profile->name);
 
 	list_remove(&profile->link);
 	free(profile);
@@ -774,16 +832,16 @@ ratbag_profile_set_enabled(struct ratbag_profile *profile, bool enabled)
 	return RATBAG_SUCCESS;
 }
 
-LIBRATBAG_EXPORT int
+LIBRATBAG_EXPORT bool
 ratbag_profile_is_active(struct ratbag_profile *profile)
 {
-	return profile->is_active;
+	return !!profile->is_active;
 }
 
 LIBRATBAG_EXPORT bool
 ratbag_profile_is_enabled(const struct ratbag_profile *profile)
 {
-	return profile->is_enabled;
+	return !!profile->is_enabled;
 }
 
 LIBRATBAG_EXPORT unsigned int
@@ -808,23 +866,47 @@ void
 ratbag_device_set_capability(struct ratbag_device *device,
 			     enum ratbag_device_capability cap)
 {
-	device->capabilities |= (1UL << cap);
+	if (cap == RATBAG_DEVICE_CAP_NONE || cap >= MAX_CAP)
+		abort();
+
+	if (cap != RATBAG_DEVICE_CAP_QUERY_CONFIGURATION && cap % 100) {
+		enum ratbag_device_capability parent_cap = (cap/100) * 100;
+
+		if (!long_bit_is_set(device->capabilities, parent_cap)) {
+			log_bug_libratbag(device->ratbag,
+					  "Cap %d requires setting %d\n",
+					  cap, parent_cap);
+			abort();
+		}
+	}
+
+	long_set_bit(device->capabilities, cap);
 }
 
-LIBRATBAG_EXPORT int
+void
+ratbag_device_unset_capability(struct ratbag_device *device,
+			     enum ratbag_device_capability cap)
+{
+	if (cap == RATBAG_DEVICE_CAP_NONE || cap >= MAX_CAP)
+		abort();
+
+	long_clear_bit(device->capabilities, cap);
+}
+
+LIBRATBAG_EXPORT bool
 ratbag_device_has_capability(const struct ratbag_device *device,
 			     enum ratbag_device_capability cap)
 {
-	if (cap == RATBAG_DEVICE_CAP_NONE)
+	if (cap == RATBAG_DEVICE_CAP_NONE || cap >= MAX_CAP)
 		abort();
 
-	return !!(device->capabilities & (1UL << cap));
+	return long_bit_is_set(device->capabilities, cap);
 }
 
 static inline enum ratbag_error_code
 write_led_helper(struct ratbag_device *device, struct ratbag_led *led)
 {
-	return device->driver->write_led(led, led->mode, led->color, led->hz, led->brightness);
+	return device->driver->write_led(led, led->mode, led->color, led->ms, led->brightness);
 }
 
 /* FIXME: This is a temporary fix for all of the drivers that have yet to be
@@ -839,7 +921,6 @@ ratbag_old_write_profile(struct ratbag_device *device)
 	struct ratbag_led *led;
 	struct ratbag_resolution *resolution;
 	int rc;
-	unsigned int i;
 
 	assert(device->driver->write_profile);
 
@@ -852,9 +933,7 @@ ratbag_old_write_profile(struct ratbag_device *device)
 			return RATBAG_ERROR_DEVICE;
 
 		if (device->driver->write_resolution_dpi) {
-			for (i = 0; i < profile->resolution.num_modes; i++) {
-				resolution = &profile->resolution.modes[i];
-
+			ratbag_profile_for_each_resolution(profile, resolution) {
 				rc = device->driver->write_resolution_dpi(
 				    resolution, resolution->dpi_x,
 				    resolution->dpi_y);
@@ -898,6 +977,8 @@ ratbag_device_commit(struct ratbag_device *device)
 {
 	struct ratbag_profile *profile;
 	struct ratbag_button *button;
+	struct ratbag_led *led;
+	struct ratbag_resolution *resolution;
 	int rc;
 
 	if (!device->driver->commit) {
@@ -915,6 +996,13 @@ ratbag_device_commit(struct ratbag_device *device)
 
 		list_for_each(button, &profile->buttons, link)
 			button->dirty = false;
+
+		list_for_each(led, &profile->leds, link)
+			led->dirty = false;
+
+		list_for_each(resolution, &profile->resolutions, link)
+			resolution->dirty = false;
+
 	}
 
 	return RATBAG_SUCCESS;
@@ -945,12 +1033,13 @@ ratbag_profile_set_active(struct ratbag_profile *profile)
 LIBRATBAG_EXPORT unsigned int
 ratbag_profile_get_num_resolutions(struct ratbag_profile *profile)
 {
-	return profile->resolution.num_modes;
+	return profile->num_resolutions;
 }
 
 LIBRATBAG_EXPORT struct ratbag_resolution *
 ratbag_profile_get_resolution(struct ratbag_profile *profile, unsigned int idx)
 {
+	struct ratbag_device *device = profile->device;
 	struct ratbag_resolution *res;
 	unsigned max = ratbag_profile_get_num_resolutions(profile);
 
@@ -960,9 +1049,15 @@ ratbag_profile_get_resolution(struct ratbag_profile *profile, unsigned int idx)
 		return NULL;
 	}
 
-	res = &profile->resolution.modes[idx];
+	ratbag_profile_for_each_resolution(profile, res) {
+		if (res->index == idx)
+			return ratbag_resolution_ref(res);
+	}
 
-	return ratbag_resolution_ref(res);
+	log_bug_libratbag(device->ratbag, "Resolution %d, profile %d not found\n",
+			  idx, profile->index);
+
+	return NULL;
 }
 
 LIBRATBAG_EXPORT struct ratbag_resolution *
@@ -989,7 +1084,7 @@ ratbag_resolution_unref(struct ratbag_resolution *resolution)
 	return NULL;
 }
 
-LIBRATBAG_EXPORT int
+LIBRATBAG_EXPORT bool
 ratbag_resolution_has_capability(struct ratbag_resolution *resolution,
 				 enum ratbag_resolution_capability cap)
 {
@@ -1013,6 +1108,7 @@ ratbag_resolution_set_dpi(struct ratbag_resolution *resolution,
 	if (resolution->dpi_x != dpi || resolution->dpi_y != dpi) {
 		resolution->dpi_x = dpi;
 		resolution->dpi_y = dpi;
+		resolution->dirty = true;
 		profile->dirty = true;
 	}
 
@@ -1035,6 +1131,7 @@ ratbag_resolution_set_dpi_xy(struct ratbag_resolution *resolution,
 	if (resolution->dpi_x != x || resolution->dpi_y != y) {
 		resolution->dpi_x = x;
 		resolution->dpi_y = y;
+		resolution->dirty = true;
 		profile->dirty = true;
 	}
 
@@ -1049,18 +1146,19 @@ ratbag_resolution_set_report_rate(struct ratbag_resolution *resolution,
 					      RATBAG_RESOLUTION_CAP_INDIVIDUAL_REPORT_RATE)) {
 		if (resolution->hz != hz) {
 			resolution->hz = hz;
+			resolution->dirty = true;
 			resolution->profile->dirty = true;
 		}
 	} else {
 		struct ratbag_profile *profile = resolution->profile;
-		unsigned int i;
+		struct ratbag_resolution *res;
 
 		/* No indvidual report rate per resolution. Loop through all of
 		 * them and update. */
-		for (i = 0; i < profile->resolution.num_modes; i++) {
-			struct ratbag_resolution *res = &profile->resolution.modes[i];
+		ratbag_profile_for_each_resolution(profile, res) {
 			if (res->hz != hz) {
 				res->hz = hz;
+				res->dirty = true;
 				profile->dirty = true;
 			}
 		}
@@ -1075,6 +1173,24 @@ ratbag_resolution_get_dpi(struct ratbag_resolution *resolution)
 	return resolution->dpi_x;
 }
 
+LIBRATBAG_EXPORT size_t
+ratbag_resolution_get_dpi_list(struct ratbag_resolution *resolution,
+			       unsigned int *resolutions,
+			       size_t nres)
+{
+	_Static_assert(sizeof(*resolutions) == sizeof(*resolution->dpis), "type mismatch");
+
+	assert(nres > 0);
+
+	if (resolution->ndpis == 0)
+		return 0;
+
+	memcpy(resolutions, resolution->dpis,
+	       sizeof(unsigned int) * min(nres, resolution->ndpis));
+
+	return resolution->ndpis;
+}
+
 LIBRATBAG_EXPORT int
 ratbag_resolution_get_dpi_x(struct ratbag_resolution *resolution)
 {
@@ -1093,46 +1209,71 @@ ratbag_resolution_get_report_rate(struct ratbag_resolution *resolution)
 	return resolution->hz;
 }
 
-LIBRATBAG_EXPORT int
+LIBRATBAG_EXPORT size_t
+ratbag_resolution_get_report_rate_list(struct ratbag_resolution *resolution,
+				       unsigned int *rates,
+				       size_t nrates)
+{
+	_Static_assert(sizeof(*rates) == sizeof(*resolution->rates), "type mismatch");
+
+	assert(nrates > 0);
+
+	if (resolution->nrates == 0)
+		return 0;
+
+	memcpy(rates, resolution->rates,
+	       sizeof(unsigned int) * min(nrates, resolution->nrates));
+
+	return resolution->nrates;
+}
+
+LIBRATBAG_EXPORT bool
 ratbag_resolution_is_active(const struct ratbag_resolution *resolution)
 {
-	return resolution->is_active;
+	return !!resolution->is_active;
 }
 
 LIBRATBAG_EXPORT enum ratbag_error_code
 ratbag_resolution_set_active(struct ratbag_resolution *resolution)
 {
+	struct ratbag_profile *profile = resolution->profile;
+	struct ratbag_resolution *res;
+
+	ratbag_profile_for_each_resolution(profile, res)
+		res->is_active = false;
+
 	resolution->is_active = true;
-	/* FIXME: call into the driver */
-	return RATBAG_ERROR_IMPLEMENTATION;
+	resolution->dirty = true;
+	profile->dirty = true;
+	return RATBAG_SUCCESS;
 }
 
 
-LIBRATBAG_EXPORT int
+LIBRATBAG_EXPORT bool
 ratbag_resolution_is_default(const struct ratbag_resolution *resolution)
 {
-	return resolution->is_default;
+	return !!resolution->is_default;
 }
 
 LIBRATBAG_EXPORT enum ratbag_error_code
 ratbag_resolution_set_default(struct ratbag_resolution *resolution)
 {
 	struct ratbag_profile *profile = resolution->profile;
-	unsigned int i;
+	struct ratbag_resolution *other;
 
 	/* Unset the default on the other resolutions */
-	for (i = 0; i < profile->resolution.num_modes; i++) {
-		struct ratbag_resolution *other = &profile->resolution.modes[i];
-
+	ratbag_profile_for_each_resolution(profile, other) {
 		if (other == resolution || !other->is_default)
 			continue;
 
 		other->is_default = false;
+		resolution->dirty = true;
 		profile->dirty = true;
 	}
 
 	if (!resolution->is_default) {
 		resolution->is_default = true;
+		resolution->dirty = true;
 		profile->dirty = true;
 	}
 
@@ -1174,7 +1315,7 @@ ratbag_button_get_action_type(struct ratbag_button *button)
 	return button->action.type;
 }
 
-LIBRATBAG_EXPORT int
+LIBRATBAG_EXPORT bool
 ratbag_button_has_action_type(struct ratbag_button *button,
 			      enum ratbag_button_action_type action_type)
 {
@@ -1200,6 +1341,17 @@ ratbag_button_get_button(struct ratbag_button *button)
 	return button->action.action.button;
 }
 
+void
+ratbag_button_set_action(struct ratbag_button *button,
+			 const struct ratbag_button_action *action)
+{
+	struct ratbag_macro *macro = button->action.macro;
+
+	button->action = *action;
+	if (action->type != RATBAG_BUTTON_ACTION_TYPE_MACRO)
+		button->action.macro = macro;
+}
+
 LIBRATBAG_EXPORT enum ratbag_error_code
 ratbag_button_set_button(struct ratbag_button *button, unsigned int btn)
 {
@@ -1212,7 +1364,7 @@ ratbag_button_set_button(struct ratbag_button *button, unsigned int btn)
 	action.type = RATBAG_BUTTON_ACTION_TYPE_BUTTON;
 	action.action.button = btn;
 
-	button->action = action;
+	ratbag_button_set_action(button, &action);
 	button->dirty = true;
 	button->profile->dirty = true;
 
@@ -1243,7 +1395,7 @@ ratbag_button_set_special(struct ratbag_button *button,
 	action.type = RATBAG_BUTTON_ACTION_TYPE_SPECIAL;
 	action.action.special = act;
 
-	button->action = action;
+	ratbag_button_set_action(button, &action);
 	button->dirty = true;
 	button->profile->dirty = true;
 
@@ -1282,7 +1434,7 @@ ratbag_button_set_key(struct ratbag_button *button,
 	action.type = RATBAG_BUTTON_ACTION_TYPE_KEY;
 	action.action.key.key = key;
 
-	button->action = action;
+	ratbag_button_set_action(button, &action);
 	button->dirty = true;
 	button->profile->dirty = true;
 
@@ -1300,7 +1452,7 @@ ratbag_button_disable(struct ratbag_button *button)
 
 	action.type = RATBAG_BUTTON_ACTION_TYPE_NONE;
 
-	button->action = action;
+	ratbag_button_set_action(button, &action);
 	button->dirty = true;
 	button->profile->dirty = true;
 
@@ -1332,8 +1484,8 @@ ratbag_button_destroy(struct ratbag_button *button)
 {
 	list_remove(&button->link);
 	if (button->action.macro) {
-		if (button->action.macro->name)
-			free(button->action.macro->name);
+		free(button->action.macro->name);
+		free(button->action.macro->group);
 		free(button->action.macro);
 	}
 	free(button);
@@ -1346,6 +1498,13 @@ ratbag_led_destroy(struct ratbag_led *led)
 	free(led);
 }
 
+static void
+ratbag_resolution_destroy(struct ratbag_resolution *res)
+{
+	list_remove(&res->link);
+	free(res);
+}
+
 LIBRATBAG_EXPORT struct ratbag_button *
 ratbag_button_unref(struct ratbag_button *button)
 {
@@ -1386,6 +1545,18 @@ ratbag_led_get_type(struct ratbag_led *led)
 	return led->type;
 }
 
+LIBRATBAG_EXPORT bool
+ratbag_led_has_mode(struct ratbag_led *led,
+		    enum ratbag_led_mode mode)
+{
+	assert(mode <= RATBAG_LED_BREATHING);
+
+	if (mode == RATBAG_LED_OFF)
+		return 1;
+
+	return !!(led->modes & (1 << mode));
+}
+
 LIBRATBAG_EXPORT struct ratbag_color
 ratbag_led_get_color(struct ratbag_led *led)
 {
@@ -1393,9 +1564,9 @@ ratbag_led_get_color(struct ratbag_led *led)
 }
 
 LIBRATBAG_EXPORT int
-ratbag_led_get_effect_rate(struct ratbag_led *led)
+ratbag_led_get_effect_duration(struct ratbag_led *led)
 {
-	return led->hz;
+	return led->ms;
 }
 
 LIBRATBAG_EXPORT unsigned int
@@ -1431,14 +1602,24 @@ ratbag_led_set_color(struct ratbag_led *led, struct ratbag_color color)
 	return RATBAG_SUCCESS;
 }
 
-LIBRATBAG_EXPORT enum ratbag_error_code
-ratbag_led_set_effect_rate(struct ratbag_led *led, unsigned int hz)
+LIBRATBAG_EXPORT enum ratbag_led_colordepth
+ratbag_led_get_colordepth(struct ratbag_led *led)
 {
 	if (!ratbag_device_has_capability(led->profile->device,
 					  RATBAG_DEVICE_CAP_LED))
 		return RATBAG_ERROR_CAPABILITY;
 
-	led->hz = hz;
+	return led->colordepth;
+}
+
+LIBRATBAG_EXPORT enum ratbag_error_code
+ratbag_led_set_effect_duration(struct ratbag_led *led, unsigned int ms)
+{
+	if (!ratbag_device_has_capability(led->profile->device,
+					  RATBAG_DEVICE_CAP_LED))
+		return RATBAG_ERROR_CAPABILITY;
+
+	led->ms = ms;
 	led->dirty = true;
 	led->profile->dirty = true;
 	return RATBAG_SUCCESS;
@@ -1480,6 +1661,32 @@ ratbag_profile_get_led(struct ratbag_profile *profile,
 	return NULL;
 }
 
+LIBRATBAG_EXPORT const char *
+ratbag_profile_get_name(struct ratbag_profile *profile)
+{
+	return profile->name;
+}
+
+LIBRATBAG_EXPORT int
+ratbag_profile_set_name(struct ratbag_profile *profile,
+			const char *name)
+{
+	char *name_copy;
+
+	if (!ratbag_profile_has_capability(profile,
+					   RATBAG_PROFILE_CAP_WRITABLE_NAME))
+		return RATBAG_ERROR_CAPABILITY;
+
+	name_copy = strdup_safe(name);
+	if (profile->name)
+		free(profile->name);
+
+	profile->name = name_copy;
+	profile->dirty = true;
+
+	return 0;
+}
+
 LIBRATBAG_EXPORT void
 ratbag_set_user_data(struct ratbag *ratbag, void *userdata)
 {
@@ -1563,8 +1770,8 @@ ratbag_button_copy_macro(struct ratbag_button *button,
 	if (!button->action.macro)
 		button->action.macro = zalloc(sizeof(struct ratbag_macro));
 	else {
-		if (button->action.macro->name)
-			free(button->action.macro->name);
+		free(button->action.macro->name);
+		free(button->action.macro->group);
 		memset(button->action.macro, 0, sizeof(struct ratbag_macro));
 	}
 
@@ -1716,3 +1923,167 @@ ratbag_button_macro_new(const char *name)
 
 	return macro;
 }
+
+struct ratbag_modifier_mapping {
+	unsigned int modifier_mask;
+	unsigned int key;
+};
+
+struct ratbag_modifier_mapping ratbag_modifier_mapping[] = {
+	{ MODIFIER_LEFTCTRL, KEY_LEFTCTRL },
+	{ MODIFIER_LEFTSHIFT, KEY_LEFTSHIFT },
+	{ MODIFIER_LEFTALT, KEY_LEFTALT },
+	{ MODIFIER_LEFTMETA, KEY_LEFTMETA },
+	{ MODIFIER_RIGHTCTRL, KEY_RIGHTCTRL },
+	{ MODIFIER_RIGHTSHIFT, KEY_RIGHTSHIFT },
+	{ MODIFIER_RIGHTALT, KEY_RIGHTALT },
+	{ MODIFIER_RIGHTMETA, KEY_RIGHTMETA },
+};
+
+int
+ratbag_button_macro_new_from_keycode(struct ratbag_button *button,
+				     unsigned int key,
+				     unsigned int modifiers)
+{
+	struct ratbag_button_macro *macro;
+	struct ratbag_modifier_mapping *mapping;
+	int i;
+
+	macro = ratbag_button_macro_new("key");
+	i = 0;
+
+	ARRAY_FOR_EACH(ratbag_modifier_mapping, mapping) {
+		if (modifiers & mapping->modifier_mask)
+			ratbag_button_macro_set_event(macro,
+						      i++,
+						      RATBAG_MACRO_EVENT_KEY_PRESSED,
+						      mapping->key);
+	}
+
+	ratbag_button_macro_set_event(macro,
+				      i++,
+				      RATBAG_MACRO_EVENT_KEY_PRESSED,
+				      key);
+	ratbag_button_macro_set_event(macro,
+				      i++,
+				      RATBAG_MACRO_EVENT_KEY_RELEASED,
+				      key);
+
+	ARRAY_FOR_EACH(ratbag_modifier_mapping, mapping) {
+		if (modifiers & mapping->modifier_mask)
+			ratbag_button_macro_set_event(macro,
+						      i++,
+						      RATBAG_MACRO_EVENT_KEY_RELEASED,
+						      mapping->key);
+	}
+
+	ratbag_button_copy_macro(button, macro);
+	ratbag_button_macro_unref(macro);
+
+	return 0;
+}
+
+int
+ratbag_action_macro_num_keys(struct ratbag_button_action *action)
+{
+	struct ratbag_macro *macro = action->macro;
+	int i, count;
+
+	count = 0;
+	for (i = 0; i < MAX_MACRO_EVENTS; i++) {
+		struct ratbag_macro_event event;
+
+		event = macro->events[i];
+		if (event.type == RATBAG_MACRO_EVENT_KEY_PRESSED)
+		{
+			switch(event.event.key) {
+			case KEY_LEFTCTRL:
+			case KEY_LEFTSHIFT:
+			case KEY_LEFTALT:
+			case KEY_LEFTMETA:
+			case KEY_RIGHTCTRL:
+			case KEY_RIGHTSHIFT:
+			case KEY_RIGHTALT:
+			case KEY_RIGHTMETA:
+				break;
+			default:
+				count++;
+			}
+		}
+	}
+
+	return count;
+}
+
+int
+ratbag_action_keycode_from_macro(struct ratbag_button_action *action,
+				 unsigned int *key_out,
+				 unsigned int *modifiers_out)
+{
+	struct ratbag_macro *macro = action->macro;
+	unsigned int key = KEY_RESERVED;
+	unsigned int modifiers = 0;
+	unsigned int i;
+
+	if (!macro || action->type != RATBAG_BUTTON_ACTION_TYPE_MACRO)
+		return -EINVAL;
+
+	if (macro->events[0].type == RATBAG_MACRO_EVENT_NONE)
+		return -EINVAL;
+
+	if (ratbag_action_macro_num_keys(action) != 1)
+		return -EINVAL;
+
+	for (i = 0; i < MAX_MACRO_EVENTS; i++) {
+		struct ratbag_macro_event event;
+
+		event = macro->events[i];
+		switch (event.type) {
+		case RATBAG_MACRO_EVENT_INVALID:
+			return -EINVAL;
+		case RATBAG_MACRO_EVENT_NONE:
+			return 0;
+		case RATBAG_MACRO_EVENT_KEY_PRESSED:
+			switch(event.event.key) {
+			case KEY_LEFTCTRL: modifiers |= MODIFIER_LEFTCTRL; break;
+			case KEY_LEFTSHIFT: modifiers |= MODIFIER_LEFTSHIFT; break;
+			case KEY_LEFTALT: modifiers |= MODIFIER_LEFTALT; break;
+			case KEY_LEFTMETA: modifiers |= MODIFIER_LEFTMETA; break;
+			case KEY_RIGHTCTRL: modifiers |= MODIFIER_RIGHTCTRL; break;
+			case KEY_RIGHTSHIFT: modifiers |= MODIFIER_RIGHTSHIFT; break;
+			case KEY_RIGHTALT: modifiers |= MODIFIER_RIGHTALT; break;
+			case KEY_RIGHTMETA: modifiers |= MODIFIER_RIGHTMETA; break;
+			default:
+				if (key != KEY_RESERVED)
+					return -EINVAL;
+
+				key = event.event.key;
+			}
+			break;
+		case RATBAG_MACRO_EVENT_KEY_RELEASED:
+			switch(event.event.key) {
+			case KEY_LEFTCTRL: modifiers &= ~MODIFIER_LEFTCTRL; break;
+			case KEY_LEFTSHIFT: modifiers &= ~MODIFIER_LEFTSHIFT; break;
+			case KEY_LEFTALT: modifiers &= ~MODIFIER_LEFTALT; break;
+			case KEY_LEFTMETA: modifiers &= ~MODIFIER_LEFTMETA; break;
+			case KEY_RIGHTCTRL: modifiers &= ~MODIFIER_RIGHTCTRL; break;
+			case KEY_RIGHTSHIFT: modifiers &= ~MODIFIER_RIGHTSHIFT; break;
+			case KEY_RIGHTALT: modifiers &= ~MODIFIER_RIGHTALT; break;
+			case KEY_RIGHTMETA: modifiers &= ~MODIFIER_RIGHTMETA; break;
+			default:
+				if (event.event.key != key)
+					return -EINVAL;
+
+				*key_out = key;
+				*modifiers_out = modifiers;
+				return 1;
+			}
+		case RATBAG_MACRO_EVENT_WAIT:
+			break;
+		default:
+			return -EINVAL;
+		}
+	}
+
+	return -EINVAL;
+}
