@@ -23,15 +23,13 @@
 
 #pragma once
 
-#ifdef __cplusplus
-extern "C" {
-#endif
-
 #include <stdlib.h>
 #include <stdarg.h>
 #include <stdbool.h>
 #include <libudev.h>
 
+#include "libratbag-enums.h"
+
 #define LIBRATBAG_ATTRIBUTE_PRINTF(_format, _args) \
 	__attribute__ ((format (printf, _format, _args)))
 #define LIBRATBAG_ATTRIBUTE_DEPRECATED __attribute__ ((deprecated))
@@ -160,49 +158,6 @@ struct ratbag_led;
  */
 struct ratbag_button_macro;
 
-/**
- * @ingroup base
- *
- * Error codes used by libratbag.
- */
-enum ratbag_error_code {
-	RATBAG_SUCCESS = 0,
-
-	/**
-	 * An error occured on the device. Either the device is not a
-	 * libratbag device or communication with the device failed.
-	 */
-	RATBAG_ERROR_DEVICE = -1000,
-
-	/**
-	 * Insufficient capabilities. This error occurs when a requested change is
-	 * beyond the device's capabilities.
-	 */
-	RATBAG_ERROR_CAPABILITY = -1001,
-
-	/**
-	 * Invalid value or value range. The provided value or value range
-	 * is outside of the legal or supported range.
-	 */
-	RATBAG_ERROR_VALUE = -1002,
-
-	/**
-	 * A low-level system error has occured, e.g. a failure to access
-	 * files that should be there. This error is usually unrecoverable
-	 * and libratbag will print a log message with details about the
-	 * error.
-	 */
-	RATBAG_ERROR_SYSTEM = -1003,
-
-	/**
-	 * Implementation bug, either in libratbag or in the caller. This
-	 * error is usually unrecoverable and libratbag will print a log
-	 * message with details about the
-	 * error.
-	 */
-	RATBAG_ERROR_IMPLEMENTATION = -1004,
-};
-
 /**
  * @ingroup base
  *
@@ -485,80 +440,6 @@ ratbag_device_get_name(const struct ratbag_device* device);
 const char *
 ratbag_device_get_svg_name(const struct ratbag_device* device);
 
-/**
- * @ingroup device
- */
-enum ratbag_device_capability {
-	RATBAG_DEVICE_CAP_NONE = 0,
-	/**
-	 * The device can change resolution, either software-controlled or
-	 * by a hardware button.
-	 *
-	 * FIXME: what about devices that only have hw buttons? can we
-	 * query that, even if we can't switch it ourselves? Maybe better to
-	 * have a separate cap for that then.
-	 */
-	RATBAG_DEVICE_CAP_SWITCHABLE_RESOLUTION,
-	/**
-	 * The device can switch between hardware profiles.
-	 * A device with this capability can store multiple profiles in the
-	 * hardware and provides the ability to switch between the profiles,
-	 * possibly with a button.
-	 * Devices without this capability will only have a single profile.
-	 */
-	RATBAG_DEVICE_CAP_SWITCHABLE_PROFILE,
-
-	/**
-	 * The device supports assigning button numbers, key events or key +
-	 * modifier combinations.
-	 */
-	RATBAG_DEVICE_CAP_BUTTON_KEY,
-
-	/**
-	 * The device supports assigning LED colors and effects
-	 */
-	RATBAG_DEVICE_CAP_LED,
-
-	/**
-	 * The device supports user-defined key or button sequences.
-	 */
-	RATBAG_DEVICE_CAP_BUTTON_MACROS,
-
-	/**
-	 * The device can have one profile assigned as a default profile.
-	 * A default profile is the profile that is selected when the device
-	 * is plugged in. Devices without this capability may select the
-	 * last-used profile or a specific profile (usually the first).
-	 */
-	RATBAG_DEVICE_CAP_DEFAULT_PROFILE,
-
-	/**
-	 * The device has the capability to query the current hardware
-	 * configuration. If this capability is missing, libratbag cannot
-	 * query the device for its current configuration and the
-	 * configured resolutions and button mappings are unknown.
-	 * libratbag will still provide information about the structure of
-	 * the device such as the number of buttons and resolutions.
-	 * Clients that encounter a device without this resolution are
-	 * encouraged to upload a configuration stored on-disk to the
-	 * device to reset the device to a known state.
-	 *
-	 * Any changes uploaded to the device will be cached in libratbag,
-	 * once a client has sent a full configuration to the device
-	 * libratbag can be used to query the device as normal.
-	 */
-	RATBAG_DEVICE_CAP_QUERY_CONFIGURATION,
-
-	/**
-	 * The device has the capability to disable and enable profiles.  While
-	 * profiles are not immediately deleted after being disabled, it is not
-	 * guaranteed that the device will remember any disabled profiles the
-	 * next time ratbag runs. Furthermore, the order of profiles may get
-	 * changed the next time ratbag runs if profiles are disabled.
-	 */
-	RATBAG_DEVICE_CAP_DISABLE_PROFILE,
-};
-
 /**
  * @ingroup device
  *
@@ -571,7 +452,7 @@ enum ratbag_device_capability {
  * @retval 1 The device has the capability
  * @retval 0 The device does not have the capability
  */
-int
+bool
 ratbag_device_has_capability(const struct ratbag_device *device,
 			     enum ratbag_device_capability cap);
 
@@ -652,6 +533,42 @@ ratbag_profile_ref(struct ratbag_profile *profile);
 struct ratbag_profile *
 ratbag_profile_unref(struct ratbag_profile *profile);
 
+/**
+ * @ingroup profile
+ *
+ * Check if a profile has a specific capability.
+ *
+ * @return non-zero if the capability is available, zero otherwise.
+ */
+bool
+ratbag_profile_has_capability(const struct ratbag_profile *profile,
+			      enum ratbag_profile_capability cap);
+
+/**
+ * @ingroup profile
+ *
+ * Return the ratbag profile name.
+ *
+ * @param profile A previously initialized ratbag profile
+ * @return the profile name
+ */
+const char *
+ratbag_profile_get_name(struct ratbag_profile *profile);
+
+/**
+ * @ingroup profile
+ *
+ * Set the name of a ratbag profile.
+ *
+ * @param profile A previously initialized ratbag profile
+ * @param name the profile name
+ *
+ * @return 0 on success or an error code otherwise
+ */
+int
+ratbag_profile_set_name(struct ratbag_profile *profile,
+			const char *name);
+
 /**
  * @ingroup profile
  *
@@ -738,7 +655,7 @@ ratbag_device_get_profile(struct ratbag_device *device, unsigned int index);
  *
  * @return non-zero if the profile is currently active, zero otherwise
  */
-int
+bool
 ratbag_profile_is_active(struct ratbag_profile *profile);
 
 /**
@@ -758,7 +675,7 @@ ratbag_profile_set_active(struct ratbag_profile *profile);
  *
  * Get the number of @ref ratbag_resolution available in this profile. A
  * resolution mode is a tuple of (resolution, report rate), each mode can be
- * fetched with ratbag_profile_get_resolution_by_idx().
+ * fetched with ratbag_profile_get_resolution().
  *
  * The returned value is the maximum number of modes available and thus
  * identical for all profiles. However, some of the modes may not be
@@ -847,20 +764,6 @@ ratbag_resolution_set_user_data(struct ratbag_resolution *resolution, void *user
 void*
 ratbag_resolution_get_user_data(const struct ratbag_resolution *resolution);
 
-enum ratbag_resolution_capability {
-	/**
-	 * The report rate can be set per resolution mode. If this property
-	 * is not available, all resolutions within the same profile have
-	 * the same report rate and changing one changes the others.
-	 */
-	RATBAG_RESOLUTION_CAP_INDIVIDUAL_REPORT_RATE = 1,
-
-	/**
-	 * The resolution can be set for x and y separately.
-	 */
-	RATBAG_RESOLUTION_CAP_SEPARATE_XY_RESOLUTION,
-};
-
 /**
  * @ingroup resolution
  *
@@ -868,7 +771,7 @@ enum ratbag_resolution_capability {
  *
  * @return non-zero if the capability is available, zero otherwise.
  */
-int
+bool
 ratbag_resolution_has_capability(struct ratbag_resolution *resolution,
 				 enum ratbag_resolution_capability cap);
 
@@ -937,6 +840,30 @@ ratbag_resolution_set_dpi_xy(struct ratbag_resolution *resolution,
 int
 ratbag_resolution_get_dpi(struct ratbag_resolution *resolution);
 
+/**
+ * @ingroup resolution
+ *
+ * Get the number of resolutions in DPI available for this resolution.
+ * The list of DPI values is sorted in ascending order but may be filtered
+ * by libratbag and does not necessarily reflect all resolutions supported
+ * by the physical device.
+ *
+ * This function writes at most nres values but returns the number of
+ * DPI values available on this resolution. In other words, if it returns a
+ * number larger than nres, call it again with an array the size of the
+ * return value.
+ *
+ * @param[out] resolutions Set to the supported DPI values in ascending order
+ * @param[in] nres The number of elements in resolutions
+ *
+ * @return The number of valid items in resolutions. If the returned value
+ * is larger than nres, the list was truncated.
+ */
+size_t
+ratbag_resolution_get_dpi_list(struct ratbag_resolution *resolution,
+			       unsigned int *resolutions,
+			       size_t nres);
+
 /**
  * @ingroup resolution
  *
@@ -1009,6 +936,29 @@ ratbag_resolution_set_report_rate(struct ratbag_resolution *resolution,
 int
 ratbag_resolution_get_report_rate(struct ratbag_resolution *resolution);
 
+/**
+ * @ingroup resolution
+ *
+ * Get the number of report rates in Hz available for this resolution.
+ * The list of report rates is sorted in ascending order but may be filtered
+ * by libratbag and does not necessarily reflect all report rates supported by
+ * the physical device.
+ *
+ * This function writes at most nrates values but returns the number of
+ * report rates available on this resolution. In other words, if it returns a
+ * number larger than nrates, call it again with an array the size of the
+ * return value.
+ *
+ * @param[out] rates Set to the supported report rates in ascending order
+ * @param[in] nrates The number of elements in resolutions
+ *
+ * @return The number of valid items in rates. If the returned value
+ * is larger than nrates, the list was truncated.
+ */
+size_t
+ratbag_resolution_get_report_rate_list(struct ratbag_resolution *resolution,
+				       unsigned int *rates,
+				       size_t nrates);
 /**
  * @ingroup resolution
  *
@@ -1038,7 +988,7 @@ ratbag_resolution_set_active(struct ratbag_resolution *resolution);
  * @return Non-zero if the resolution mode is the active one, zero
  * otherwise.
  */
-int
+bool
 ratbag_resolution_is_active(const struct ratbag_resolution *resolution);
 
 /**
@@ -1073,7 +1023,7 @@ ratbag_resolution_set_default(struct ratbag_resolution *resolution);
  * @return Non-zero if the resolution mode is the default one, zero
  * otherwise.
  */
-int
+bool
 ratbag_resolution_is_default(const struct ratbag_resolution *resolution);
 
 /**
@@ -1123,47 +1073,6 @@ ratbag_button_set_user_data(struct ratbag_button *button, void *userdata);
 void*
 ratbag_button_get_user_data(const struct ratbag_button *button);
 
-/**
- * @ingroup button
- *
- * Button types describing the physical button.
- */
-enum ratbag_button_type {
-	RATBAG_BUTTON_TYPE_UNKNOWN = 0,
-
-	/* mouse buttons */
-	RATBAG_BUTTON_TYPE_LEFT,
-	RATBAG_BUTTON_TYPE_MIDDLE,
-	RATBAG_BUTTON_TYPE_RIGHT,
-	RATBAG_BUTTON_TYPE_THUMB,
-	RATBAG_BUTTON_TYPE_THUMB2,
-	RATBAG_BUTTON_TYPE_THUMB3,
-	RATBAG_BUTTON_TYPE_THUMB4,
-	RATBAG_BUTTON_TYPE_WHEEL_LEFT,
-	RATBAG_BUTTON_TYPE_WHEEL_RIGHT,
-	RATBAG_BUTTON_TYPE_WHEEL_CLICK,
-	RATBAG_BUTTON_TYPE_WHEEL_UP,
-	RATBAG_BUTTON_TYPE_WHEEL_DOWN,
-	/**
-	 * A button to toggle the wheel from free-spinning to click-based.
-	 */
-	RATBAG_BUTTON_TYPE_WHEEL_RATCHET_MODE_SHIFT,
-	RATBAG_BUTTON_TYPE_EXTRA,
-	RATBAG_BUTTON_TYPE_SIDE,
-	RATBAG_BUTTON_TYPE_PINKIE,
-	RATBAG_BUTTON_TYPE_PINKIE2,
-
-	/* DPI switch */
-	RATBAG_BUTTON_TYPE_RESOLUTION_CYCLE_UP,
-	RATBAG_BUTTON_TYPE_RESOLUTION_UP,
-	RATBAG_BUTTON_TYPE_RESOLUTION_DOWN,
-
-	/* Profile */
-	RATBAG_BUTTON_TYPE_PROFILE_CYCLE_UP,
-	RATBAG_BUTTON_TYPE_PROFILE_UP,
-	RATBAG_BUTTON_TYPE_PROFILE_DOWN,
-};
-
 /**
  * @ingroup button
  *
@@ -1183,39 +1092,6 @@ enum ratbag_button_type {
 enum ratbag_button_type
 ratbag_button_get_type(struct ratbag_button *button);
 
-/**
- * @ingroup button
- *
- * The type assigned to a button.
- */
-enum ratbag_button_action_type {
-	/**
-	 * Button action is unknown
-	 */
-	RATBAG_BUTTON_ACTION_TYPE_UNKNOWN = -1,
-	/**
-	 * Button is disabled
-	 */
-	RATBAG_BUTTON_ACTION_TYPE_NONE = 0,
-	/**
-	 * Button sends numeric button events
-	 */
-	RATBAG_BUTTON_ACTION_TYPE_BUTTON,
-	/**
-	 * Button triggers a mouse-specific special function. This includes
-	 * resolution changes and profile changes.
-	 */
-	RATBAG_BUTTON_ACTION_TYPE_SPECIAL,
-	/**
-	 * Button sends a key or key + modifier combination
-	 */
-	RATBAG_BUTTON_ACTION_TYPE_KEY,
-	/**
-	 * Button sends a user-defined key or button sequence
-	 */
-	RATBAG_BUTTON_ACTION_TYPE_MACRO,
-};
-
 /**
  * @ingroup button
  *
@@ -1241,51 +1117,10 @@ ratbag_button_get_action_type(struct ratbag_button *button);
  *
  * @return non-zero if the action type is supported, zero otherwise.
  */
-int
+bool
 ratbag_button_has_action_type(struct ratbag_button *button,
 			      enum ratbag_button_action_type action_type);
 
-/**
- * @ingroup button
- */
-enum ratbag_button_action_special {
-	/**
-	 * This button is not set up for a special action
-	 */
-	RATBAG_BUTTON_ACTION_SPECIAL_INVALID = -1,
-	RATBAG_BUTTON_ACTION_SPECIAL_UNKNOWN = (1 << 30),
-
-	RATBAG_BUTTON_ACTION_SPECIAL_DOUBLECLICK,
-
-	/* Wheel mappings */
-	RATBAG_BUTTON_ACTION_SPECIAL_WHEEL_LEFT,
-	RATBAG_BUTTON_ACTION_SPECIAL_WHEEL_RIGHT,
-	RATBAG_BUTTON_ACTION_SPECIAL_WHEEL_UP,
-	RATBAG_BUTTON_ACTION_SPECIAL_WHEEL_DOWN,
-	RATBAG_BUTTON_ACTION_SPECIAL_RATCHET_MODE_SWITCH,
-
-	/* DPI switch */
-	RATBAG_BUTTON_ACTION_SPECIAL_RESOLUTION_CYCLE_UP,
-	RATBAG_BUTTON_ACTION_SPECIAL_RESOLUTION_CYCLE_DOWN,
-	RATBAG_BUTTON_ACTION_SPECIAL_RESOLUTION_UP,
-	RATBAG_BUTTON_ACTION_SPECIAL_RESOLUTION_DOWN,
-	RATBAG_BUTTON_ACTION_SPECIAL_RESOLUTION_ALTERNATE,
-	RATBAG_BUTTON_ACTION_SPECIAL_RESOLUTION_DEFAULT,
-
-	/* Profile */
-	RATBAG_BUTTON_ACTION_SPECIAL_PROFILE_CYCLE_UP,
-	RATBAG_BUTTON_ACTION_SPECIAL_PROFILE_CYCLE_DOWN,
-	RATBAG_BUTTON_ACTION_SPECIAL_PROFILE_UP,
-	RATBAG_BUTTON_ACTION_SPECIAL_PROFILE_DOWN,
-
-	/* second mode for buttons */
-	RATBAG_BUTTON_ACTION_SPECIAL_SECOND_MODE,
-
-	/* battery level */
-	RATBAG_BUTTON_ACTION_SPECIAL_BATTERY_LEVEL,
-
-};
-
 /**
  * @ingroup button
  *
@@ -1340,34 +1175,6 @@ ratbag_button_set_button(struct ratbag_button *button,
 enum ratbag_button_action_special
 ratbag_button_get_special(struct ratbag_button *button);
 
-/**
- * @ingroup led
- *
- * RATBAG_LED_OFF - led is now off,
- * RATBAG_LED_ON - led is on with static color,
- * RATBAG_LED_CYCLE - led is cycling between all colors.
- * RATBAG_LED_BREATHING - led is pulsating with static color
- *
- * Each LED mode has different properties, e.g. the brightness and rate are only
- * available in modes @ref RATBAG_LED_CYCLE and @ref RATBAG_LED_BREATHING modes
- */
-enum ratbag_led_mode {
-	RATBAG_LED_OFF = 0,
-	RATBAG_LED_ON,
-	RATBAG_LED_CYCLE,
-	RATBAG_LED_BREATHING,
-};
-/**
- * @ingroup led
- *
- * LED types, usually based on their physical location
- */
-enum ratbag_led_type {
-	RATBAG_LED_TYPE_UNKNOWN = -1,
-	RATBAG_LED_TYPE_LOGO = 0,
-	RATBAG_LED_TYPE_SIDE
-};
-
 /**
  * @ingroup led
  *
@@ -1398,13 +1205,30 @@ ratbag_profile_get_led(struct ratbag_profile *profile, unsigned int index);
  */
 enum ratbag_led_type
 ratbag_led_get_type(struct ratbag_led *led);
+
+/**
+ * @ingroup led
+ *
+ * This function returns true if the given mode is supported by the LED, or
+ * false otherwise.
+ *
+ * @param led A previously initialized ratbag LED
+ * @param mode The LED mode @ref ratbag_led_mode to check for
+ * @return True if supported, false otherwise
+ *
+ * @see ratbag_led_set_mode
+ */
+bool
+ratbag_led_has_mode(struct ratbag_led *led,
+		    enum ratbag_led_mode mode);
+
 /**
  * @ingroup led
  *
  * This function returns the mode for ratbag_led.
  *
  * @param led A previously initialized ratbag LED
- * @return The LED mod @ref ratbag_led_mode
+ * @return The LED mode @ref ratbag_led_mode
  *
  * @see ratbag_led_set_mode
  */
@@ -1415,6 +1239,11 @@ ratbag_led_get_mode(struct ratbag_led *led);
  *
  * This function returns the led color.
  *
+ * If any color scaling applies because of the device's color depth
+ * this is not reflected in the returned value. In other words,
+ * the returned value always matches the most recent value provided
+ * to ratbag_led_set_color().
+ *
  * @param led A previously initialized ratbag LED
  * @return The LED color in @ref ratbag_led_mode
  *
@@ -1422,18 +1251,32 @@ ratbag_led_get_mode(struct ratbag_led *led);
  */
 struct ratbag_color
 ratbag_led_get_color(struct ratbag_led *led);
+
 /**
  * @ingroup led
  *
- * This function returns the LED effect rate.
+ * This function returns the color depth of this LED.
  *
  * @param led A previously initialized ratbag LED
- * @return The LED rate in Hz, can be 100 - 20000
+ * @return The bit depth of this LED
  *
- * @see ratbag_led_set_effect_rate
+ * @see ratbag_led_set_color
+ */
+enum ratbag_led_colordepth
+ratbag_led_get_colordepth(struct ratbag_led *led);
+
+/**
+ * @ingroup led
+ *
+ * This function returns the LED effect duration.
+ *
+ * @param led A previously initialized ratbag LED
+ * @return The LED duration in ms, can be 0 - 10000
+ *
+ * @see ratbag_led_set_effect_duration
  */
 int
-ratbag_led_get_effect_rate(struct ratbag_led *led);
+ratbag_led_get_effect_duration(struct ratbag_led *led);
 /**
  * @ingroup led
  *
@@ -1467,6 +1310,12 @@ ratbag_led_set_mode(struct ratbag_led *led, enum ratbag_led_mode mode);
  * If the LED's mode is @ref RATBAG_LED_ON or @ref RATBAG_LED_BREATHING
  * then this function sets the LED color, otherwise it has no effect.
  *
+ * The color provided has to be within the allowed color range (see @ref
+ * ratbag_color). libratbag silently scales and/or clamps this range into
+ * the device's color depth. It is the caller's responsibility to set the
+ * colors in a non-ambiguous way for the device's bit depth. See @ref
+ * ratbag_led_colordepth for more details.
+ *
  * @param led A previously initialized ratbag LED
  * @param color A LED color.
  * @return 0 on success or an error code otherwise.
@@ -1480,16 +1329,16 @@ ratbag_led_set_color(struct ratbag_led *led, struct ratbag_color color);
  * @ingroup led
  *
  * If the LED's mode is @ref RATBAG_LED_CYCLE or @ref RATBAG_LED_BREATHING
- * then this function sets the LED rate in Hz
+ * then this function sets the LED duration in ms
  *
  * @param led A previously initialized ratbag LED
- * @param rate Effect rate in hz, 100 - 20000
+ * @param rate Effect duration in ms, 0 - 10000
  * @return 0 on success or an error code otherwise.
  *
- * @see ratbag_led_get_effect_rate
+ * @see ratbag_led_get_effect_duration
  */
 enum ratbag_error_code
-ratbag_led_set_effect_rate(struct ratbag_led *led, unsigned int rate);
+ratbag_led_set_effect_duration(struct ratbag_led *led, unsigned int rate);
 
 /**
  * @ingroup led
@@ -1579,19 +1428,6 @@ ratbag_button_set_key(struct ratbag_button *button,
 enum ratbag_error_code
 ratbag_button_disable(struct ratbag_button *button);
 
-/**
- * @ingroup button
- *
- * Macro event types describing the event.
- */
-enum ratbag_macro_event_type {
-	RATBAG_MACRO_EVENT_INVALID = -1,
-	RATBAG_MACRO_EVENT_NONE = 0,
-	RATBAG_MACRO_EVENT_KEY_PRESSED,
-	RATBAG_MACRO_EVENT_KEY_RELEASED,
-	RATBAG_MACRO_EVENT_WAIT,
-};
-
 /**
  * @ingroup button
  *
@@ -1814,6 +1650,3 @@ ratbag_led_ref(struct ratbag_led *led);
 struct ratbag_led *
 ratbag_led_unref(struct ratbag_led *led);
 
-#ifdef __cplusplus
-}
-#endif
