@@ -31,8 +31,10 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <limits.h>
+#include <sys/resource.h>
 
 #include "libratbag.h"
+#include "libratbag-util.h"
 #include "libratbag-test.h"
 
 static void
@@ -57,10 +59,16 @@ const struct ratbag_test_device sane_device = {
 	.profiles = {
 		{
 		.resolutions = {
-			{ .xres = 100, .yres = 200, .hz = 1000 },
+			{ .xres = 100, .yres = 200, .hz = 1000,
+				.dpi_min = 100, .dpi_max = 5000,
+				.report_rates = { 500, 1000 } },
 			{ .xres = 200, .yres = 300, .hz = 1000 },
 			{ .xres = 300, .yres = 400, .hz = 1000 },
 		},
+		.leds = {
+			{ .type = RATBAG_LED_TYPE_SIDE, },
+			{ .type = RATBAG_LED_TYPE_LOGO, },
+		},
 		.active = true,
 		.dflt = false,
 		},
@@ -83,14 +91,14 @@ const struct ratbag_test_device sane_device = {
 			{
 				.mode = RATBAG_LED_ON,
 				.color = { .red = 255, .green = 0,	.blue = 0 },
-				.hz = 1,
-				.brightness = 20
+				.ms = 1000,
+				.brightness = 20,
 			},
 			{
 				.mode = RATBAG_LED_CYCLE,
 				.color = { .red = 255, .green = 255, .blue = 0 },
-				.hz = 3,
-				.brightness = 40
+				.ms = 333,
+				.brightness = 40,
 			}
 		},
 		.active = false,
@@ -348,7 +356,9 @@ START_TEST(device_resolutions)
 		.profiles = {
 			{
 			.resolutions = {
-				{ .xres = 100, .yres = 200, .hz = 1000 },
+				{ .xres = 100, .yres = 200, .hz = 1000,
+					.dpi_min = 50, .dpi_max = 5000,
+					.report_rates = { 500, 1000 } },
 				{ .xres = 200, .yres = 300, .hz = 1000, .active = true },
 				{ .xres = 300, .yres = 400, .hz = 1000 },
 			},
@@ -383,6 +393,9 @@ START_TEST(device_resolutions)
 		ck_assert_int_eq(nresolutions, 3);
 
 		for (j = 0; j < nresolutions; j++) {
+			unsigned int dpis[200];
+			int ndpis = ARRAY_LENGTH(dpis);
+
 			res = ratbag_profile_get_resolution(p, j);
 
 			xres = ratbag_resolution_get_dpi_x(res);
@@ -390,11 +403,18 @@ START_TEST(device_resolutions)
 			rate = ratbag_resolution_get_report_rate(res);
 			is_active = ratbag_resolution_is_active(res);
 
+			ndpis = ratbag_resolution_get_dpi_list(res, dpis, ndpis);
+			ck_assert_int_lt(ndpis, ARRAY_LENGTH(dpis));
+			ck_assert_int_gt(ndpis, 20);
+
 			ck_assert_int_eq(xres, i * 1000 + (j + 1) * 100);
 			ck_assert_int_eq(yres, i * 1000 + (j + 1) * 100 + 100);
 			ck_assert_int_eq(xres, ratbag_resolution_get_dpi(res));
 			ck_assert_int_eq(is_active, (j == 1));
 
+			ck_assert_int_eq(dpis[0], 50);
+			ck_assert_int_eq(dpis[ndpis - 1], 5000);
+
 			ck_assert_int_eq(rate, (i + 1) * 1000);
 
 			ratbag_resolution_unref(res);
@@ -488,7 +508,7 @@ START_TEST(device_freed_before_profile)
 	is_active = ratbag_profile_is_active(p);
 	ck_assert_int_eq(is_active, 1);
 
-	p = ratbag_profile_unref(p);
+	ratbag_profile_unref(p);
 	ratbag_unref(r);
 	ck_assert_int_eq(device_freed_count, 1);
 }
@@ -680,7 +700,7 @@ START_TEST(device_buttons)
 	struct ratbag_test_device td = sane_device;
 	td.num_buttons = 10;
 
-	td.profiles[0].buttons[8].type = RATBAG_BUTTON_ACTION_TYPE_MACRO;
+	td.profiles[0].buttons[8].action_type = RATBAG_BUTTON_ACTION_TYPE_MACRO;
 
 	td.destroyed_data = &device_freed_count;
 
@@ -773,18 +793,18 @@ assert_led_equals(struct ratbag_led *l, struct ratbag_test_led e_l)
 {
 	enum ratbag_led_mode mode;
 	struct ratbag_color color;
-	int brightness, hz;
+	int brightness, ms;
 
 	ck_assert(l != NULL);
 	mode = ratbag_led_get_mode(l);
 	color = ratbag_led_get_color(l);
-	hz = ratbag_led_get_effect_rate(l);
+	ms = ratbag_led_get_effect_duration(l);
 	brightness = ratbag_led_get_brightness(l);
 	ck_assert_int_eq(mode, e_l.mode);
 	ck_assert_int_eq(color.red, e_l.color.red);
 	ck_assert_int_eq(color.green, e_l.color.green);
 	ck_assert_int_eq(color.blue, e_l.color.blue);
-	ck_assert_int_eq(hz, e_l.hz);
+	ck_assert_int_eq(ms, e_l.ms);
 	ck_assert_int_eq(brightness, e_l.brightness);
 }
 
@@ -810,10 +830,10 @@ START_TEST(device_leds)
 	p = ratbag_device_get_profile(d, 2);
 	ck_assert(p != NULL);
 
-	led_logo = ratbag_profile_get_led(p, RATBAG_LED_TYPE_LOGO);
-	assert_led_equals(led_logo, td.profiles[2].leds[RATBAG_LED_TYPE_LOGO]);
-	led_side = ratbag_profile_get_led(p, RATBAG_LED_TYPE_SIDE);
-	assert_led_equals(led_side, td.profiles[2].leds[RATBAG_LED_TYPE_SIDE]);
+	led_logo = ratbag_profile_get_led(p, 0);
+	assert_led_equals(led_logo, td.profiles[2].leds[0]);
+	led_side = ratbag_profile_get_led(p, 1);
+	assert_led_equals(led_side, td.profiles[2].leds[1]);
 
 	ratbag_led_unref(led_logo);
 	ratbag_led_unref(led_side);
@@ -852,14 +872,14 @@ START_TEST(device_leds_set)
 	p = ratbag_device_get_profile(d, 0);
 	ck_assert(p != NULL);
 
-	l = ratbag_profile_get_led(p, RATBAG_LED_TYPE_LOGO);
+	l = ratbag_profile_get_led(p, 0);
 
 	ratbag_led_set_mode(l, RATBAG_LED_BREATHING);
 	ratbag_led_set_color(l, c);
-	ratbag_led_set_effect_rate(l, 11);
+	ratbag_led_set_effect_duration(l, 90);
 	ratbag_led_set_brightness(l, 22);
 
-	l = ratbag_profile_get_led(p, RATBAG_LED_TYPE_LOGO);
+	l = ratbag_profile_get_led(p, 0);
 	struct ratbag_test_led e_l = {
 		.mode = RATBAG_LED_BREATHING,
 		.color = {
@@ -867,7 +887,7 @@ START_TEST(device_leds_set)
 			.green = c.green,
 			.blue = c.blue
 		},
-		.hz = 11,
+		.ms = 90,
 		.brightness = 22
 	};
 	assert_led_equals(l, e_l);
@@ -932,6 +952,11 @@ int main(void)
 	int nfailed;
 	Suite *s;
 	SRunner *sr;
+	const struct rlimit corelimit = { 0, 0 };
+
+	setenv("RATBAG_TEST", "1", 0);
+
+	setrlimit(RLIMIT_CORE, &corelimit);
 
 	s = test_context_suite();
         sr = srunner_create(s);
