@@ -28,17 +28,22 @@ udev_device_from_path(struct udev *udev, const char *path)
 {
 	struct udev_device *udev_device;
 	const char *event_node_prefix = "/dev/input/event";
+	_cleanup_(freep) char *path_canonical = NULL;
 
-	if (strneq(path, event_node_prefix, strlen(event_node_prefix))) {
+	if ((path_canonical = canonicalize_file_name(path)) == NULL) {
+		error("Failed to canonicalize path '%s': %s\n", path, strerror(errno));
+		return NULL;
+	}
+	if (strneq(path_canonical, event_node_prefix, strlen(event_node_prefix))) {
 		struct stat st;
-		if (stat(path, &st) == -1) {
+		if (stat(path_canonical, &st) == -1) {
 			error("Failed to stat '%s': %s\n", path, strerror(errno));
 			return NULL;
 		}
 		udev_device = udev_device_new_from_devnum(udev, 'c', st.st_rdev);
 
 	} else {
-		udev_device = udev_device_new_from_syspath(udev, path);
+		udev_device = udev_device_new_from_syspath(udev, path_canonical);
 	}
 	if (!udev_device) {
 		error("Can't open '%s': %s\n", path, strerror(errno));
@@ -51,18 +56,27 @@ udev_device_from_path(struct udev *udev, const char *path)
 const char*
 led_type_to_str(enum ratbag_led_type type)
 {
-	const char *str = "UNKNOWN";
+	const char *str = NULL;
 
 	switch(type) {
-	case RATBAG_LED_TYPE_UNKNOWN:
-		str = "unknown";
-		break;
 	case RATBAG_LED_TYPE_LOGO:
 		str = "logo";
 		break;
 	case RATBAG_LED_TYPE_SIDE:
 		str = "side";
 		break;
+	case RATBAG_LED_TYPE_BATTERY:
+		str = "battery";
+		break;
+	case RATBAG_LED_TYPE_DPI:
+		str = "dpi";
+		break;
+	case RATBAG_LED_TYPE_WHEEL:
+		str = "wheel";
+		break;
+	default:
+		assert(!"Invalid LED type");
+		break;
 	}
 
 	return str;
@@ -283,8 +297,8 @@ struct ratbag_device *
 ratbag_cmd_open_device(struct ratbag *ratbag, const char *path)
 {
 	struct ratbag_device *device;
-	_cleanup_udev_unref_ struct udev *udev = NULL;
-	_cleanup_udev_device_unref_ struct udev_device *udev_device = NULL;
+	_cleanup_(udev_unrefp) struct udev *udev = NULL;
+	_cleanup_(udev_device_unrefp) struct udev_device *udev_device = NULL;
 	enum ratbag_error_code error;
 
 	udev = udev_new();
